#Nearest Singular Matrix Polynomial via Determinant method
#A symbolic-numeric iterative post-refinement technique
#to find the nearest singular matrix polynomial via the optimization formulation
#min ||Delta A|| s.t. det(A+\Delta A)=0.

read("../../../lib/nearest_singular_poly_functions.mpl"):
read("../../../lib/alpha_optimization.mpl"): #needed for Levenberg-Marquardt
read("../../../lib/nearest_smith_functions.mpl"): #needed for approximate GCD

with(NearestPoly):
with(AlphaOptimization):
with(NearestSmith):

read("../../../lib/numeric_poly_linalg.mpl"): #needed for some kernel stuff

read("../../../lib/SQP_solvers.mpl"): #Needed for SLN-newton
#Input: matrix poly in t, determinant of something
#output: Vectorized determinant, padded to have the right dimension (n*d+1)
#Note:
#needed for helping in derivative computation.

vectorize_determinant := proc(A,t,my_det)
    local n,d,size_bound,det_vec;
    
    n:= LinearAlgebra:-RowDimension(A);
    d := matrix_degree(A,t);
    size_bound := n*d+1;
    
 
    det_vec:= Vector(PolynomialTools:-CoefficientVector(my_det,t),size_bound);


    return det_vec; 
    
end proc:

#proc: fast_gradient_determinant
#Input: Matrix polynomial A w/ variable t, list of vars x, list of evaluation pont X_evl
#Output: the gradient of $\det(\cdot)$ evaluated at X_init
#Notes:
#implement these with Maples Determinant function for now. We'll make use 
#of the symbolic implementation for ease of use. 
fast_gradient_determinant := proc(A,t,DA,X,X_init)
    local G, #final answer
           i,
           n,
           d,
           my_exp, #expression for differentiation using our trickery.
           size_bound, #nd+1
           J, #the final jacobian?
           my_grad, #intermediate computation
            active_vars #vars we will be differentiating w.r.t. 
            ;
            
            n := LinearAlgebra:-RowDimension(A); #assume #A# is square
            d := matrix_degree(A,t);
            size_bound := n*d+1; #for the determinant
            J:=Matrix(0,0); #hack for initilization
        
         for i from 1 to LinearAlgebra:-RowDimension(DA) do 
		for j from 1 to LinearAlgebra:-ColumnDimension(DA) do
		    active_vars := convert(PolynomialTools:-CoefficientVector(DA[i,j],t),set) minus {0};
		    #we will use a block-variant of the algorithm. It should be faster. 
		    
		    
		    #this is basically the variables we are ignoring for differentiation.
		    
		    inactive_pt := convert(convert(DeltaA_vars,set) minus active_vars,Vector);   
		    inactive_pt := Equate(inactive_pt, subs(X_init,inactive_pt));
		    
		    #these are the variables we need for differentiation.
		    active_pt := convert(active_vars,Vector);
		    active_pt :=  Equate(active_pt, subs(X_init,active_pt));
		    
		    
		    active_vars := convert(active_vars,list);
		    
		 
		    A_copy := A+subs(inactive_pt,DA); #substitute all points not used in diff
		 
		    #convert into a vectorized form.
		    #note that this routine uses only one Adjoint call.
		    detA_vec := vectorize_determinant(A,t,LinearAlgebra:-Determinant(A_copy));
		    
		    
		    #this can be optimized by differentiatiing wrt to active vars and chaining
		    #rows together.
		    if ArrayNumElems(J) =0 then 
			J:= VectorCalculus:-Jacobian(detA_vec,active_vars);
		    else 
			J:= `<|>`(J,VectorCalculus:-Jacobian(detA_vec,active_vars));
			#J:= J+VectorCalculus:-Jacobian(detA_vec,DeltaA_vars);
		    fi:
		od:
	   od:
            
            return(subs(X_init,J));
end proc:


fast_hessianXX_determinant := proc(A,t,DA,X,Z_init,Lambda)
local i, j,k,ell,
	    n,
	    det,
	    J,J_temp, #the jacobian 
	    active_vars, active_vars_ij,#needed for book keeping
	    active_pt, #needed for book keeping
	    A_copy,
	    inactive_pt,
	    d, #degree bound
	    H, #hessian
	    Lambda_vec,
	    detA_vec, #this is used for some black magic. 
	    size_bound
	    ;


#go parse this cluster fuck some how

            d:=matrix_degree(A,t);
	    n := min(LinearAlgebra:-Dimension(A));
            size_bound := n*d+1;
J:=Matrix(0,0); #hack for initilization
            
            
            
	    #Let's try the block method here.
	    for i from 1 to LinearAlgebra:-RowDimension(DA) do 
		for j from 1 to LinearAlgebra:-ColumnDimension(DA) do
		    active_vars_ij := convert(PolynomialTools:-CoefficientVector(DA[i,j],t),set) minus {0};
		    detA_vec:=Vector(size_bound); # allocate space
	   
		    for k from i+1 to LinearAlgebra:-RowDimension(DA) do
			for ell from 1 to LinearAlgebra:-ColumnDimension(DA) do 
			    #Due to symmetry in the underlying problem we need
			    #only N^2 choose 2 adjoint calls instead of N^2
			    if ell <> j then
			    
				active_vars := convert(PolynomialTools:-CoefficientVector(DA[k,ell],t),set) minus {0};
				active_vars := active_vars_ij union active_vars;
				#we will use a block-variant of the algorithm. It should be faster. 
			    
				#this is basically the variables we are ignoring for differentiation.
				
				inactive_pt := convert(convert(DA_vars,set) minus active_vars,Vector);   
				inactive_pt := Equate(inactive_pt, subs(Z_init,inactive_pt));
				
				#these are the variables we need for differentiation.
				active_pt := convert(active_vars,Vector);
				active_pt :=  Equate(active_pt, subs(Z_init,active_pt));
				
				
				active_vars := convert(active_vars,list);
				
			    
				A_copy := A+subs(inactive_pt,DA); #substitute all points not used in diff
			    
				#convert into a vectorized form.
				#note that this routine uses only one Adjoint call.
				detA_vec:=detA_vec+ vectorize_determinant(A,t,LinearAlgebra:-Determinant(A_copy));
			    fi:
			od:
		    od:
		    
		    active_vars_ij := convert(active_vars_ij,list);
		    if ArrayNumElems(J) =0 then 
			J:= VectorCalculus:-Jacobian(detA_vec,active_vars_ij);
		    else 
			J:= `<|>`(J,VectorCalculus:-Jacobian(detA_vec,active_vars_ij));
			#J:= J+VectorCalculus:-Jacobian(adjA_vec,DA_vars);
		    fi:
		    
		od:
	   od:
	    
	 #   J:= `<|>`(J,VectorCalculus:-Jacobian(hfs_vec,fh_vars)); 
	    
	    Lambda_vec := convert(subs(Z_init,Lambda),Vector[row]): 
	    
	    H:=VectorCalculus:-Jacobian( Lambda_vec.J, DA_vars);
	   # H:=VectorCalculus:-Hessian(Lambda_vec.adjA_vec,DA_vars);
	   H:=H+LinearAlgebra:-Transpose(H)+2*LinearAlgebra:-IdentityMatrix(LinearAlgebra:-Dimension(H)); #account for ${DA}
	   
	    return subs(Z_init,H);
end proc:



#proc: fast_hessian_full_determinant
#Input: Matrix poly in t, perturbation structure, listof vars, evaluation point
#Output Full Hessian Matrix of the KKT problem.
#Notes: this is for Min ||Delta A| s.t. det(A+Delta A)=0
#       This function basically computes and merges some matrices.

#Z = [X,Lambda] 
fast_hessian_full_determinant:= proc(A,t,DA,X,Z_init,Lambda)
        local   H,J;             
                
                J:= fast_gradient_determinant(A,t,DA,X,Z_init);
                H:= fast_hessianXX_determinant(A,t,DA,X,Z_init,subs(Z_init,Lambda));
                H:= `<,>`(H,J);
                H:=`<|>`(H,`<,>`(LinearAlgebra:-Transpose(J),Matrix(LinearAlgebra:-RowDimension(J))));
end proc:


#proc: fast_symbolic_gradient
#Input: usual arguments
#output: gradient of L, i.e. computes $\nabla L$.
#Notes: used for merit function stuff. 
fast_symbolic_gradient := proc(A,t,DA,X,Lambda,Z_init)
        local J,G;
        J:= fast_gradient_determinant(A,t,DA,X,Z_init);
        
         G:= 2*Vector(convert(subs(Z_init,X[1..nops(X)]),Vector),nops(X));
         G:= G+ LinearAlgebra:-Transpose(J) . convert(subs(Z_init,Lambda),Vector);
         G:=convert(`<,>`(G,vectorize_determinant(A,t,LinearAlgebra:-Determinant(A+subs(Z_init,DA)))),Vector);
        
        return G;
end proc:



#Proc: symbolic_numeric_lm
#Input: arguments in some form...
#Output: solution to Grad L =0
#Notes: This is like the fully implicit but with special functions.
#	It is not general purpose, and as such not ubiquitous. 
#	the variable handing is less compact snice we don't want to do parsing here and to get fast
#	function calls we need to pass explicit arguments. I.e. trade portability for speed :(
symbolic_numeric_lm := proc(my_args,tol:=1e-16,max_iter := 100000,lm_param:= 1,lm_amp_factor:=1e-1,lm_damp_factor:=1e1)
      local i,
	    grad,
	    J,
	    JTJ,
	    rhs_k, #first argument to linear solve
	    lhs_k,#second rgument to linear solve
	    current_pt,
	    x_next,
	    Imat,
	    residuals,
	    obj,
	    t_start,t_end,t_iter,
	    j,
	    dx,
	    x_next_candidate,
	    x_cur,
	    A,t,DeltaA,DeltaA_vars,fstar,h,Lambda,X,Z_init,Z,
	    Z_next_candidate,
	    mu_k,#the current levenberg-marquardt parameter
	    safety_factor #for LM
	    ;
	    t_start := time():
	    
	    A,t,DeltaA,DeltaA_vars,Lambda,X,Z_init,Z := op(my_args);
	    
	    #safety_factor:= 10^(-1*Digits);
	    safety_factor := 0:
	    #default h for now
	    #J:= numeric_hessian22(my_func,my_args,x,x_init);
	   
	    J:= fast_hessian_full_determinant(A,t,DA,X,Z_init,Lambda);
	    
	    JTJ := LinearAlgebra:-Transpose(J).J;

	    printf("Computed Initial Numerical Hessian...\n");
	    residuals := fast_symbolic_gradient(A,t,DeltaA,X,Lambda,Z_init);
	    
	   
	    obj := norm(residuals,2);
	    
	    
	    printf("Computed numerical gradient...\n");
	    
	    #hack to overcome some maple problems in previous versions.
	    #probably obsolete now.
	    Imat:=Matrix(LinearAlgebra:-Dimension(JTJ));
	    for j from 1 to LinearAlgebra:-RowDimension(Imat) do 
		Imat[j,j]:=1:
	    od:
	    
	  
	    
	    x_cur := convert(subs(Z_init,Z),Vector);
	    mu_k := lm_param;
	  
	 
	 
	    printf("Beginning iterations. Setuptime is %f \n", time()-t_start);
	  #  t_iter := time():
	    for i from 1 to max_iter do
		rhs_k := JTJ + mu_k * Imat;
		lhs_k := -1*LinearAlgebra:-Transpose(J).residuals;
		dx := LinearAlgebra:-LinearSolve(rhs_k,lhs_k);
		
		#printf("Computed dx\n");

		x_next_candidate := x_cur + dx;
		
		#residuals_next := numeric_gradient2(my_func,my_args,x,x_next_candidate);
		Z_next_candidate := Equate(Z,x_next_candidate): 
		residuals_next :=  fast_symbolic_gradient(A,t,DeltaA,X,Lambda,Z_next_candidate);
		obj_next := norm(residuals_next,2);
		
		#return obj_next, obj;
	
		if obj_next <= obj  then #valid step 
		#printf("Valid step. Proceeding to lower mu.\n");
		    mu_k :=max( mu_k*(lm_amp_factor),safety_factor); #take a smaller step next iteration
		   
		   
		    #this is based on a criteria from a paper by Wright.
		    #|| residuals||_2 approximates the solution, so the penalty term
		    #blows up at the same rate as the solution, giving asymptotic quadratic convergence.
		    if  norm(dx,2) <=  obj and obj <= 1.5*norm(dx,2) then 
			mu_k := obj_next;
		    fi:
		    
		    x_next := x_cur + dx;
		    obj := obj_next;
		    printf("k = %d, mu=%e, ||x_k - x_k+1|| = %e, ||f(x_k+1)||/||dx|| = %e, ||f(x_k)||=%e\n",i,mu_k, norm(dx,2), obj_next/norm(dx,2), obj_next);
		    
		    if norm(x_cur-x_next,2) < tol or obj < tol  then break fi:
		
		    x_cur := x_next;
		    
		    residuals := residuals_next; #recycle computation
		    
		    #J:= numeric_hessian22(my_func,my_args,x, x_cur);
		    Z_next := Z_next_candidate; #candidate was sucessful.
		    
                    J:= fast_hessian_full_determinant(A,t,DA,X,Z_next,Lambda);
                    JTJ := LinearAlgebra:-Transpose(J).J;
		else 
		    #printf("Invalid step. Decreasing mu\n");
		    mu_k := mu_k * (lm_damp_factor);  #take a smaller step and try again
		    
		    if norm(dx,2)<tol or  obj < tol  then break fi:
		
		fi:
		
		
		
	    
		
	    od:
	    
	    current_pt := subs(Equate(Vector(Z),x_cur));
	    t_end := time():
	    printf("Computation took %f seconds\n",t_end-t_start);
	    printf("||DA||=%e \n", poly_norm_opt(subs(current_pt,DeltaA),t));
	    return [obj ,current_pt, t_end-t_start];
	    
end proc:



kernelopts(printbytes=false):

#A:= <t+2,t+1|t+2,t+2>;

#A:= <t+1,t+1,t^2+t+1|1,3,4|3,2,t>;

aa:= LinearAlgebra:-Transpose(<0, 0.0400, 0.8900 | 0.1500 , -0.0200, 0 | 0.9200, 0.1100, 0.06600>);
ee:= <0,0,0|0,0,1|0,1,0>;


aa:= (<-1.79, .84, -0.89 | .10,-.54 ,.3 | -.6, .49, .74>);
ee:= <0,0,0|0,0,1|0,1,0>;
A:= collect(aa+t*ee,t);

DA,DA_vars := generate_perturbation(A,t,"degree");
X:= DA_vars; 

obj := poly_norm_opt(DA,t)^2;
n := LinearAlgebra:-RowDimension(A);
d := matrix_degree(A,t);

cons :=Vector( PolynomialTools:-CoefficientVector(LinearAlgebra:-Determinant(A+DA),t),n*d+1);

X_init := [seq(DA_vars[i]=0,i=1..nops(DA_vars))];


vectorize_determinant(A,t,LinearAlgebra:-Determinant(A));

#printlevel:=5:

fast_gradient_determinant(A,t,DA,X,X_init);

subs(X_init,VectorCalculus:-Jacobian(cons,X)); #verify we're not a fuck up
LinearAlgebra:-Norm(%-%%,2); #verify the "slow" and "fast" answers.

Lambda := Vector(ArrayNumElems(cons),i-> lambda[i] );

#now we need some fast hessian shit man

subs(X_init,VectorCalculus:-Hessian(poly_norm_opt(DA,t)^2+LinearAlgebra:-Transpose(Lambda). cons, X));
fast_hessianXX_determinant(A,t,DA,X,X_init,Lambda);


LinearAlgebra:-Norm(%-%%,2);


H_implicit := fast_hessian_full_determinant(A,t,DA,X,X_init,Lambda);

L := obj + LinearAlgebra:-Transpose(Lambda).cons:
Z:= [op(X), op(convert(Lambda,list))];
H_explicit := subs(X_init,VectorCalculus:-Hessian(L,Z));

LinearAlgebra:-Norm(H_implicit-H_explicit,2);

fast_symbolic_gradient(A,t,DA,X,Lambda,X_init);
Vector(subs(X_init,VectorCalculus:-Gradient(L,Z)));
LinearAlgebra:-Norm(%-%%,2);


lagrange_residuals :=  fast_symbolic_gradient(A,t,DA,X,Lambda,X_init)[1..nops(X)];
AA,bb := LinearAlgebra:-GenerateMatrix(Equate(Vector(subs(X_init,lagrange_residuals)),Vector(ArrayNumElems(lagrange_residuals))),convert(Lambda,list)):
Lambda_init :=  Equate(convert(Lambda,Vector), LinearAlgebra:-LeastSquares(evalf(AA),evalf(bb)));


Z_init := [op(X_init), op(Lambda_init)];

my_args := [A,t,DA,DA_vars,Lambda,X,Z_init,Z]:

my_result := symbolic_numeric_lm(my_args,1e-15)[2]: #use a modified levenberg-marquardt iteration to account for fast adjoint derivative computation.


Z_final := Equate(Vector(Z),subs(my_result,Z));
#Z_final := Z_init;

H:= fast_hessian_full_determinant(A,t,DA,X,Z_final,Lambda);

Hxx:= H[1.. nops(X),1..nops(X)];
J:= LinearAlgebra:-Transpose( H[1..nops(X), nops(X)+1..LinearAlgebra:-ColumnDimension(H)]);

LinearAlgebra:-Rank(J);

my_ker := `<|>`(op(LinearAlgebra:-NullSpace(subs(Z_final,J))));


sort(Re(LinearAlgebra:-Eigenvalues( LinearAlgebra:-Transpose(my_ker). Hxx.my_ker)));



my_residuals := Vector(VectorCalculus:-Gradient(L,Z)):

my_soln := levenberg_marquardt_implicit_merit(my_residuals,Z,Z_init,1e-12)[2];#iterative post-refinement


Z_final := Equate(Vector(Z),subs(my_soln,Z));
#Z_final := Z_init;

H:= fast_hessian_full_determinant(A,t,DA,X,Z_final,Lambda);

Hxx:= H[1.. nops(X),1..nops(X)];
J:= LinearAlgebra:-Transpose( H[1..nops(X), nops(X)+1..LinearAlgebra:-ColumnDimension(H)]);

LinearAlgebra:-Rank(J);

my_ker := `<|>`(op(LinearAlgebra:-NullSpace(subs(Z_final,J))));


sort(Re(LinearAlgebra:-Eigenvalues( LinearAlgebra:-Transpose(my_ker). Hxx.my_ker)));


subs(my_result,DA+A);
LinearAlgebra:-Determinant(%);
poly_norm_opt(subs(my_result,DA),t);

#get A0
evalf(subs(t=0,subs(my_soln,DA)),7);
latex(%);


#get A1
evalf(subs(t=0,Matrix(A,(i,j) -> diff(subs(my_soln,DA[i,j]),t))),7);
latex(%);

#get A2
evalf(subs(t=0,Matrix(A,(i,j) -> 1/2*diff(subs(my_soln,DA[i,j]),t$2))),7);
latex(%);

#get A3
evalf(subs(t=0,Matrix(A,(i,j) -> 1/6*diff(subs(my_soln,DA[i,j]),t$3))),7);
latex(%);
