#Copyright Joseph Haraldson, 2018
#All rights reserved.
#Not for redistribution.

#to run go to the command line and enter:
# > maple lower-mccoy.mpl

read("../lib/nearest_smith_functions.mpl");
read("../lib/alpha_optimization.mpl");
read("../lib/numeric_smith_prototype.mpl"): #get some helper functions

#A:= <t^2+.1*t+1,0,.2*t,0|0,.9*t^2+.2*t+1.3,0,.1*t^2+1.2| .3*t-.1,0,t^2+1.32+.03*t^3,0 | 0,.1,0,.89*t^2+.89>; #test example


A:= <t^3+3*t+1,0,t+1| 1,t^2+2*t+2,t+1|t+1,0,t^3+3*t+2*t+1>;



my_local_extrema := [fsolve( diff(LinearAlgebra:-Determinant(A)^2,t)=0,complex)];



#s_rank corresponds to the low mccoy rank approximation. 
s_rank :=3; #change this to 4 to get the second example

my_guess_index := min[index]([seq(LinearAlgebra:-SingularValues(subs(t=my_local_extrema[i],A))[LinearAlgebra:-RowDimension(A)-s_rank+1] ,i=1..nops(my_local_extrema))]);


A_init := A; #assume the input is adequate

DeltaA, DA_vars:= generate_perturbation(A,t,"support"): #type of coefficient perturbations



omega_init := my_local_extrema[my_guess_index]; #pick a guess.


my_residuals,Z,Z_init,omega:=op(lower_mccoy(A,t,A_init,omega_init,s_rank)): #get initial parameter values 


Digits := 35:
my_soln := levenberg_marquardt_implicit_merit(my_residuals,Z,Z_init,1e-25)[2]: #iterative post-refinement
Digits := 10:

omega_final := subs(my_soln,omega);

DA_final := subs(my_soln,DeltaA);
A_opt := DA_final + A;
LinearAlgebra:-SingularValues(subs(t=omega_final,A_opt)); #certify correctness


poly_norm_opt(DA_final,t); #get the size



#verify the computation with the adjoint method for s_rank = 2

my_guess := extract_guess_information(A,A_opt,t,1,2,"support"): #if s_rank =2 this should be a stationary point. 


A_opt_clean := clean_up(A_opt,t,1e-6); #clean output for paper


latex(A);
latex(omega_init);
latex(omega_final);
latex(evalf(A_opt_clean,5));




seq(degree(LinearAlgebra:-Adjoint(A+DeltaA)(i),t),i=1..ArrayNumElems(A));

seq(degree(LinearAlgebra:-Adjoint(A)(i),t),i=1..ArrayNumElems(A)); 

seq(degree(LinearAlgebra:-Adjoint(A_opt)(i),t),i=1..ArrayNumElems(A)); 

#these degree sequences should not all decrease or increase in degree from the original. If this happens then an unattainable GCD could happen.


poly_least_squares_division_implicit( [LinearAlgebra:-Determinant(A_opt)],  (t-Re(omega_final))^2,t);

poly_least_squares_division_implicit( [LinearAlgebra:-Determinant(A_opt)],  (t-Re(omega_final))^3,t);

S_eps := op(%[2]);

poly_least_squares_division_implicit( [LinearAlgebra:-Determinant(A_opt)],  (t-Re(omega_final))^4,t);

LinearAlgebra:-Determinant(A_opt);

Digits := 30:
zeros := [(solve(LinearAlgebra:-Determinant(A_opt) =0,t))]; #we only care about real zeros since the invariant factor is linear
Digits := 10:
[seq(LinearAlgebra:-SingularValues(subs(t=Re(zeros[i]), A_opt)),i=1..nops(zeros) )];
zeros;

eigs := [solve(S_eps = 0,t)];

seq(LinearAlgebra:-SingularValues(subs(t=eigs[j],A_opt)),j=1..nops(eigs));

eigs;
quit;

#######################################################################################3

#Second part


my_guess := extract_guess_information(A,A,t,1,2,"support"):



DeltaA,fstar,h :=op( my_guess[1]):
X,X_init := op(my_guess[2]):
Lambda,Lambda_init := op(my_guess[3]):
DeltaA_vars,fstar_vars,h_vars := op(my_guess[4]):




Z:= [op(X),op(Lambda)]:
Z_init := [op(X_init),op(Lambda_init)]:

my_args := [A,t,DeltaA,DeltaA_vars,fstar,h,Lambda,X,Z_init,Z]:

printlevel:=3:



#LinearAlgebra:-Adjoint(A+DeltaA);

my_result := symbolic_numeric_lm(my_args,1e-15)[2]: #use a modified levenberg-marquardt iteration to account for fast adjoint derivative computation.

poly_norm_opt(subs(my_result,DeltaA)-subs(my_soln,DeltaA),t);
DA_final := subs(my_result,DeltaA);
A_opt := DA_final + A;
my_guess := extract_guess_information(A,A_opt,t,1,2,"support"): #verify the answer again.





