read("../../../lib/alpha_optimization.mpl"):
with(AlphaOptimization);
read("../../../lib/nearest_singular_poly_functions.mpl"):
read("../../../lib/nearest_smith_functions.mpl"):
#with(NearestPoly);
with(NearestSmith);
read("../../../lib/rank-factor-lib.mpl");




with(NearestPoly):
with(AlphaOptimization):
with(NearestSmith):

kernelopts(printbytes=false):


#proc: compute_initial_h_guess
#Input: Matrix polynomial A in parameter $t$ with vector of supports alpha
#       mode: "real" or "complex" - used to guess if a solution is real or not. 
#Output:


compute_initial_h_guess := proc(A,t,Alpha, tol:=1e-7)
        local s_rank,i,omega_init,h_init,my_local_extrema,my_guess_index;
        s_rank := 0:

        for i from 1 to ArrayNumElems(Alpha) do
            if Alpha[i] <> 0 then s_rank := s_rank +1; fi:
        od:

        #get a list of the local extrema, these are suitable guesses
        my_local_extrema := [fsolve( diff(LinearAlgebra:-Determinant(A)^2,t)=0,complex)];
        #take a point that gives us something of approximately the desired McCoy rank
        my_guess_index := min[index]([seq(LinearAlgebra:-SingularValues(subs(t=my_local_extrema[i],evalf(A)))[LinearAlgebra:-RowDimension(A)-s_rank+1] ,i=1..nops(my_local_extrema))]);


        omega_init := my_local_extrema[my_guess_index]; #pick a guess.

        #tol:= 1e-8; #tolerance for degree 1 or degree 2 guesses

        if abs( Im(omega_init)) < tol then 
            h_init := t-Re(omega_init);
            h_vars := [hh[0]];
            h_init_pt := [hh[0] = subs(t=0,h_init)];
            h := hh[0]+t;
        else 
              h_init := Re(evalc((t-omega_init)*(t-conjugate(omega_init)))) assuming real; #this should be real!
; #this should be real!
            h_vars := [hh[0],hh[1]];
            h_init_pt := [hh[0] = subs(t=0,h_init), hh[1] = subs(t=0,diff(h_init,t))];
            h:= hh[0]+hh[1]*t+t^2;
        fi:

    return [h_init,h_vars,h_init_pt,h];
end proc:


#Proc: compute_initial_USV_guess
#Input: Matrix polynomial A in indeterminant $t$ and initial guess for h,  $h_init$ with vector of degrees
#Output: initial guesses for U  and SV
#        Some more infformation 
compute_initial_UV_guess := proc(A,t,h_init,Alpha)

        local   UU,U_init,U_init_pt,U_vars,
                SS,S_init,S_init_pt,S_vars,
                VS,V_init,V_init_pt,V_vars,
                m,n,d,s_rank,
                my_guess_index,
                U_degree_bound,V_degree_bound,sn_degree_bound,
                U_scale_factor,V_scale_factor,
                H,
                sn_vars,sn,s_vec,sn_init,sn_init_pt,
                USHV, #product of U,S,H and V in ANSATZ form.
                      #this expression has a polynomial blow up and can be handled with symbolic differentiation.
                Y, #U,S,V  
                Y_vars, #vars for U,S,V
                Y_init# initial USV init point
                ;
                m,n := LinearAlgebra:-Dimension(A);
        
        m,n := LinearAlgebra:-Dimension(A);
        d:= matrix_degree(collect(A,t),t);
        s_rank := add(Alpha[j],j=1..ArrayNumElems(Alpha));

#        U_degree_bound := n*d;
#        V_degree_bound := n*(d+1); #is this bound right?
       
        U_init,S_init,V_init := LinearAlgebra:-SmithForm(A,t,output=['U','S','V']):
 
 
        U_degree_bound := max(0,matrix_degree(U_init,t));
        V_degree_bound := max(0,matrix_degree(V_init,t)); #is this bound right?
       
        
        UU := Matrix(n,n, (i,j)-> add( u[i,j,k]*t^k, k=0..U_degree_bound ) );
        VV := Matrix(n,n, (i,j)-> add( v[i,j,k]*t^k, k=0..V_degree_bound ) );
        
        U_vars := [seq(seq(seq(u[i,j,k], k= 0..U_degree_bound), j=1..n),i=1..n)];
        V_vars := [seq(seq(seq(v[i,j,k], k= 0..V_degree_bound), j=1..n),i=1..n)];
        
        
        #get scaling factors.  
        
        U_scale_factor :=1;#LinearAlgebra:-Determinant(subs(t=0,U_init));
        V_scale_factor :=1;# LinearAlgebra:-Determinant(subs(t=0,V_init));
        
        #now they have determinant 1
        #and are feasible w.r.t. to our optimization. 
        #U_init := (U_init/(U_scale_factor^(1/n))); 
        
        
        #V_init := (V_init/V_scale_factor^(1/n));
        
        U_init_pt := op(solve(Equate(vectorize(UU-U_init,t),Vector(n^2*(U_degree_bound+1))),U_vars));
        
        #print(matrix_degree(VV,t));
        #print(matrix_degree(V_init,t));
        
        V_init_pt := op(solve(Equate(vectorize(VV-V_init,t),Vector(n^2*(V_degree_bound+1))),V_vars));
        
     
        ##U and V initial guess code is basically done here. 
        
        
        sn_degree_bound := n*d - s_rank; #the bound of the last invariant factor divided out by h^alpha
        
        print(s_rank);
        print(n*d);
        print(sn_degree_bound);
        
        
        sn := add( ss[n,j]*t^j,j=0..sn_degree_bound);
        
        s_degree_bound := 0; #assume these guys are constants for now!
        s_vec := Vector(n,j-> add(ss[j,i]*t^i,i=0..s_degree_bound));
        s_vec[n] := sn ;
        SS:= LinearAlgebra:-DiagonalMatrix(s_vec); #create ANSATZ for $S$

        
        #create a list of variables for the first n-1 entries
        S_vars := [ seq( op(convert(PolynomialTools:-CoefficientVector(SS[j,j],t),list)) ,j=1..n-1) ];
        #create a list of variables for the last guy
        sn_vars := [op(convert(PolynomialTools:-CoefficientVector(sn,t),list))];
        S_vars := [op(S_vars), op(sn_vars)];
        
        #the entries start off as 1 and are re-scaled by rescaling the unimodular multipliers.
        S_init_pt := [seq(S[i,i]=evalf(U_scale_factor*V_scale_factor), i = 1..n-1)];
        
        sn_init := 1;
        
        
        detA:= LinearAlgebra:-Determinant(A);
        #use my least squares polynomial division routine to get this guy
        if degree(h_init,t) + s_rank-1 <= degree(detA,t) then 
            sn_init:= op(poly_least_squares_division_implicit ([evalf(detA)],collect(h_init^s_rank,t),t)[2]); 
            
            print(degree(detA,t));
            print(degree(collect(h_init^s_rank,t),t));
            
            print((poly_least_squares_division_implicit ([evalf(detA)],collect(h_init^s_rank,t),t)));
        fi:
        #rescale this guy reasonably...
        sn_init := sn_init/lcoeff(sn_init,t);
        
        
        
        sn_init_pt := Equate(Vector(PolynomialTools:-CoefficientVector(sn,t), sn_degree_bound+1),Vector(PolynomialTools:-CoefficientVector(sn_init,t), sn_degree_bound+1));
#        sn_init_pt := op(solve(vectorize(sn-sn_init,t),sn_vars));
        
        S_init_pt := [op(S_init_pt), op(sn_init_pt)];
        #list of variables is now complete. 

        Y_init := [op(U_init_pt), op(S_init_pt), op(V_init_pt)];
        Y_vars := [op(U_vars), op(S_vars), op(V_vars)];
        Y := [UU,SS,VV];
        
        return evalf([Y,Y_init,Y_vars]);
        
end proc:



#A:= <t^3+.1*t+1,0,.2*t,0|0,-.9*t^2+.2*t+1.3,0,-.1*t^2+1.2| .3*t-.1,0,t^2+1.32+.03*t^3,0 | #0,.1*t,0,-.89*t^3+.89>; #test example

#A_init := A:
#A_init:= <(t+1.03)^2,0|(t+1.03)^2,(t+1.03)^2>;

#A:= <(t+1.01)^2,t^2+2|(t+.99)^2,(t+1.01)^2>;




A_d:= LinearAlgebra:-DiagonalMatrix(<(t+1)^2,(t+1)^2,(t+1)^4*(t^2+3)>);
U_guess := <1,t+1,t-3|0,1,t-4|0,0,1>;
V_guess := LinearAlgebra:-Transpose(<0,0,1|0,1,t+2|-3,t,1>);

A_init := collect(U_guess.A_d.V_guess,t);

#A:=collect(A_d,t);
#A_init := A_d;


A_noise := 1e-4*<0,0,1e-4|0,1e-3,1e-4|1e-3,1e-3,1e-3>;

A := A_init+ evalf(A_noise,30);


#A:= <t^3+.1*t+1,.2*t,0|0,-.9*t^2+.2*t+1.3,-.1*t^2+1.2| .3*t-.1,t^2+1.32+.03*t^3,0>; #test example
#A_init := A;
#A:= A_init;
DeltaA,DeltaA_vars := generate_perturbation(A,t,"support");
#A_init := A;

Alpha := <2,2,4>;  #degrees of h


m,n := LinearAlgebra:-Dimension(A);
d:= matrix_degree(A,t);



 #  U_init,S_init,V_init := LinearAlgebra:-SmithForm(A,t,output=['U','S','V']);
        
        


h_init,h_vars,h_init_pt,h := op(compute_initial_h_guess(A_init,t,Alpha));



Y,Y_init,Y_vars := op(compute_initial_UV_guess(A_init,t,h_init,Alpha)):

U,S,V := op(Y);


U_init:= subs(Y_init,U);
V_init := subs(Y_init,V);
U_degree_bound:= matrix_degree(U_init,t);
V_degree_bound:=matrix_degree(V_init,t);

H:= LinearAlgebra:-DiagonalMatrix(Vector(n,j->h^Alpha[j]));


DA_init := vectorize(A+DeltaA-A_init,t);
DA_init := op(solve(Equate(DA_init,Vector(ArrayNumElems(DA_init))),DeltaA_vars));

u_det_constraint := Vector(PolynomialTools:-CoefficientVector(LinearAlgebra:-Determinant(U)-LinearAlgebra:-Determinant(U_init),t),n*U_degree_bound+1);

v_det_constraint := Vector(PolynomialTools:-CoefficientVector(LinearAlgebra:-Determinant(V)-LinearAlgebra:-Determinant(V_init),t),n*V_degree_bound+1);

constraints := <vectorize(U.(A+DeltaA).V- S.H,t), u_det_constraint,v_det_constraint>;




X:= [op(DeltaA_vars), op(h_vars), op(Y_vars)];
X_init := [op(Y_init), op(h_init_pt), op(DA_init) ];

subs(X_init,LinearAlgebra:-Determinant(U));

norm(subs(X_init,u_det_constraint));
norm(subs(X_init,v_det_constraint));
norm(subs(X_init,constraints),2);




simplify(subs(X_init,U.S.H.V));

LinearAlgebra:-SmithForm(A_init,t);

subs(X_init,S);
subs(X_init,H);
subs(X_init, collect(S.H,t));

Lambda := Vector[row](ArrayNumElems(constraints),j->lambda[j]);

obj := poly_norm_opt(DeltaA,t)^2;

L:= obj + (Lambda.constraints)[1]:

Z:= [op(X),op(convert(Lambda,list))];

residuals := Vector(VectorCalculus:-Gradient(L,Z)):





#guess Lagrange multipliers
AA,bb := LinearAlgebra:-GenerateMatrix(convert(subs(X_init,residuals),list),convert(Lambda,list));

Lambda_init := Optimization:-LSSolve([bb,AA])[2];
Lambda_init := Equate(LinearAlgebra:-Transpose(Lambda),Lambda_init);

#testing
Lambda_init := Equate(Lambda,Vector[row](ArrayNumElems(Lambda)));



Z_init := [op(X_init), op(Lambda_init)];


norm(subs(Z_init,residuals),2);

my_soln :=levenberg_marquardt_implicit_merit(residuals,Z,Z_init,1e-14):

Digits := 30:
my_soln :=levenberg_marquardt_implicit_merit(residuals,Z,evalf(my_soln[2]),1e-27):

 
 subs(my_soln[2],U);
 LinearAlgebra:-Determinant(%);
 
 subs(my_soln[2],V);
 LinearAlgebra:-Determinant(%);
 
 subs(my_soln[2],H);
 subs(my_soln[2],S);
 subs(my_soln[2],DeltaA);




 
 J:= VectorCalculus:-Jacobian(convert(constraints,Vector),X);




 
 
 Hxx:=subs(my_soln[2],VectorCalculus:-Hessian(L,X));
 
 LinearAlgebra:-Rank(Hxx);
 
 J_ker :=`<|>`( op(LinearAlgebra:-NullSpace(subs(my_soln[2],J)))):
 
sort( Re(LinearAlgebra:-Eigenvalues( LinearAlgebra:-Transpose(J_ker).Hxx.J_ker)));
 min(%);
 

poly_norm_opt( subs(my_soln[2],DeltaA),t);
poly_norm_opt(subs(X_init,DeltaA),t);










quit;

#get a list of the local extrema, these are suitable guesses
my_local_extrema := [fsolve( diff(LinearAlgebra:-Determinant(A)^2,t)=0,complex)];



#s_rank corresponds to the low mccoy rank approximation. 
#s_rank :=2; #change this to 4 to get the second example
s_rank := add(Alpha[j],j=1..ArrayNumElems(Alpha));


#take a point that gives us something of approximately the desired McCoy rank
my_guess_index := min[index]([seq(LinearAlgebra:-SingularValues(subs(t=my_local_extrema[i],A))[LinearAlgebra:-RowDimension(A)-s_rank+1] ,i=1..nops(my_local_extrema))]);


A_init := A; #assume the input is adequate

DeltaA, DA_vars:= generate_perturbation(A,t,"support"): #type of coefficient perturbations

omega_init := my_local_extrema[my_guess_index]; #pick a guess.

tol:= 1e-8; #tolerance for degree 1 or degree 2 guesses

if abs( Im(omega_init)) < tol then 
    h_init := t-Re(omega_init);
    h_vars := [hh[0]];
    h_init_pt := [hh[0] = subs(t=0,h_init)];
    h := hh[0]+t;
else 
    h_init := Re(evalc((t-omega_init)*(t-conjugate(omega_init)))) assuming real; #this should be real!
    h_vars := [hh[0],hh[1]];
    h_init_pt := [hh[0] = subs(t=0,h_init), hh[1] = subs(t=0,diff(h_init,t))];
    h:= hh[0]+hh[1]*t+t^2;
fi:

h_vars;
h_init_pt;
h_init;
h;

s_n_degree_bound := n*d - s_rank; 

s_n := add( ss[n,j]*t^j,j=0..s_n_degree_bound);

s_degree_bound := 0; #assume these guys are constants for now!
s_vec := Vector(n,j-> add(ss[j,i]*t^i,i=0..s_degree_bound));
s_vec[n] := s_n ;

SS:= LinearAlgebra:-DiagonalMatrix(s_vec);

S_vars := [ seq( op(convert(PolynomialTools:-CoefficientVector(SS[j,j],t),list)) ,j=1..n) ];

#call the least squares polynomial division routine that I wrote.
s_n_init := op(poly_least_squares_division_implicit ([LinearAlgebra:-Determinant(A)],h_init^s_rank,t)[2]);
s_n_init := s_n_init / lcoeff(s_n_init,t); #make the S_n term monic for now. 


