#must read this first!

read("../../lib/nearest_singular_poly_functions.mpl");
read("../../lib/nearest_smith_functions.mpl"): #get some helper functions

read("../../lib/numeric_smith_prototype.mpl"): #get some helper functions
read("../../lib/alpha_optimization.mpl"): #get some helper functions


kernelopts(printbytes=false):


#proc: compute_initial_h_guess
#Input: Matrix polynomial A in parameter $t$ with vector of supports alpha
#       mode: "real" or "complex" - used to guess if a solution is real or not. 
#Output:


compute_initial_h_guess := proc(A,t,Alpha, tol:=1e-7)
        local s_rank,i,omega_init,h_init,my_local_extrema,my_guess_index;
        s_rank := 0:

        s_rank := add(Alpha[j],j=1..ArrayNumElems(Alpha));


        #take a point that gives us something of approximately the desired McCoy rank
        my_guess_index := min[index]([seq(LinearAlgebra:-SingularValues(subs(t=my_local_extrema[i],A))[LinearAlgebra:-RowDimension(A)-s_rank+1] ,i=1..nops(my_local_extrema))]);


        A_init := A; #assume the input is adequate

        DeltaA, DA_vars:= generate_perturbation(A,t,"support"): #type of coefficient perturbations

        omega_init := my_local_extrema[my_guess_index]; #pick a guess.

        tol:= 1e-8; #tolerance for degree 1 or degree 2 guesses

        if abs( Im(omega_init)) < tol then 
            h_init := t-Re(omega_init);
            h_vars := [hh[0]];
            h_init_pt := [hh[0] = subs(t=0,h_init)];
            h := hh[0]+t;
        else 
            h_init := (t-omega_init)*(t-conj(omega_init)); #this should be real!
            h_vars := [hh[0],hh[1]];
            h_init_pt := [hh[0] = subs(t=0,h_init), hh[1] = subs(t=0,diff(h_init,t))];
            h:= hh[0]+hh[1]*t+t^2;
        fi:

    return [h_init,h_vars,h_init_pt];
end proc:


#Proc: compute_initial_USV_guess
#Input: Matrix polynomial A in indeterminant $t$ and initial guess for h,  $h_init$ with vector of degrees
#Output: initial guesses for U  and SV
#        Some more infformation 
compute_initial_UV_guess := proc(A,t,h_init,Alpha)

        local   U,U_init,U_init_pt,U_vars,
                S,S_init,S_init_pt,S_vars,
                V,V_init,V_init_pt,V_vars,
                m,n,d,s_rank,
                my_guess_index,
                U_degree_bound,V_degree_bound,sn_degree_bound,
                U_scale_factor,V_scale_factor,
                H,
                USHV, #product of U,S,H and V in ANSATZ form.
                      #this expression has a polynomial blow up and can be handled with symbolic differentiation.
                Y, #USV vars 
                Y_init# initial USV init point
                ;
                m,n := LinearAlgebra:-Dimension(A);
        
        m,n := LinearAlgebra:-Dimension(A);
        d:= matrix_degree(A,t);
        s_rank := add(Alpha[j],j=1..ArrayNumElems(Alpha));

        
        U_init,S_init,V_init := LinearAlgebra:-SmithForm(A,t,output=['U','S','V']):
        
        U := Matrix(n,n, (i,j)-> add( u[i,j,k]*t^k, k=0..U_degree_bound ) );
        V := Matrix(n,n, (i,j)-> add( v[i,j,k]*t^k, k=0..V_degree_bound ) );
        
        U_vars := [seq(seq(seq(u[i,j,k], k= 0..U_degree_bound), j=1..n),i=1..n)];
        V_vars := [seq(seq(seq(v[i,j,k], k= 0..V_degree_bound), j=1..n),i=1..n)];
        
        
        #get scaling factors.  
        
        U_scale_factor :=LinearAlgebra:-Determinant(subs(t=0,U_init));
        V_scale_factor := LinearAlgebra:-Determinant(subs(t=0,V_init));
        
        #now they have determinant 1
        #and are feasible w.r.t. to our optimization. 
        U_init := U_init/U_scale_factor; 
        V_init := V_init/V_scale_factor;
        
        U_init_pt := op(solve(Equate(vectorize(U-U_init,t),Vector(n^2*(U_degree_bound+1))),U_vars));
        V_init_pt := op(solve(Equate(vectorize(V-V_init,t),Vector(n^2*(V_degree_bound+1))),V_vars));
        
        
        ##U and V initial guess code is basically done here. 
        
        my_guess_index := min[index]([seq(LinearAlgebra:-SingularValues(subs(t=my_local_extrema[i],A))[LinearAlgebra:-RowDimension(A)-s_rank+1] ,i=1..nops(my_local_extrema))]);

        
        #the re-scaling is needed for feasible point reasons.
        #S_init := S_init/lcoeff(S_init,t); #make this guy monic
        #S_init := S_init * U_scale_factor*V_scale_factor;  #rescale the entries of S
        
        
        sn_degree_bound := n*d - s_rank; #the bound of the last invariant factor divided out by h^alpha
        sn := add( ss[n,j]*t^j,j=0..s_n_degree_bound);
        
        s_degree_bound := 0; #assume these guys are constants for now!
        s_vec := Vector(n,j-> add(ss[j,i]*t^i,i=0..s_degree_bound));
        s_vec[n] := s_n ;
        S:= LinearAlgebra:-DiagonalMatrix(s_vec); #create ANSATZ for $S$

        
        #create a list of variables for the first n-1 entries
        S_vars := [ seq( op(convert(PolynomialTools:-CoefficientVector(S[j,j],t),list)) ,j=1..n-1) ];
        #create a list of variables for the last guy
        sn_vars := op(convert(PolynomialTools:-CoefficientVector(sn,t),list));
        S_vars := [op(S_vars), op(sn_vars)];
        
        #the entries start off as 1 and are re-scaled by rescaling the unimodular multipliers.
        S_init_pt := [seq(S[i]=U_scale_factor*V_scale_factor, i = 1..n-1)];
        
        #use my least squares polynomial division routine to get this guy
        sn_init:= op(poly_least_squares_division_implicit ([S_init[n,n]],h_init^s_rank,t)[2]); 
        #rescale this guy reasonably...
        sn_init := sn_init/lcoeff(sn_init,t)*U_scale_factor*V_scale_factor;
        
        sn_init_pt := op(solve(vectorize(sn-sn_init,t),sn_vars));
        
        S_init_pt := [op(S_init_pt), op(sn_init_pt)];
        #list of variables is now complete. 

        Y_init := [op(U_init_pt), op(S_init), op(V_init_pt)];
        Y_vars := [op(U_vars), op(S_varS), op(V_vars)];
        Y := [U,S,V];
        
        return [Y,Y_init,Y_vars];
        
end proc:







A_init:= <t+1,0|t+1,t+1>;

A:= <t+1,1e-2*(t^2)|t+1,t+1>;
DeltaA,DeltaA_vars := generate_perturbation(A,t,"entry");
m,n := LinearAlgebra:-Dimension(A);
d:= matrix_degree(A,t);

supports :=<1,1>;
Alpha:=supports;

h_init := compute_initial_h_guess(A,t,Alpha);
S_init := IdentityMatrix(n); #make an identity matrix 

d_A := LinearAlgebra:-Determinant(A);





d_h :=degree(h_init,t);

h := add( hh[i]*t^i, i=0..d_h-1 )+t^d_h; #make h monic for now.
h_vars := [seq(hh[i],i=0..d_h-1)];


sn_degree_bound := 0: #for now...
s_n := add( ss[j] * t^j, j=0..sn_degree_bound );

S_vars := [seq(seq(ss[i,j],j=0..sn_degree_bound[i]),i=1..n)];

S := LinearAlgebra:-DiagonalMatrix(Vector(n,i->add( ss[i,j] * t^j, j=0..sn_degree_bound )));
#S[n,n] := s_n:



H := LinearAlgebra:-DiagonalMatrix(Vector(n,i->h^(supports[i])));

U_degree_bound := n*d;
V_degree_bound := n*d;

U := Matrix(n,n, (i,j)-> add( u[i,j,k]*t^k, k=0..U_degree_bound ) );

V := Matrix(n,n, (i,j)-> add( v[i,j,k]*t^k, k=0..V_degree_bound ) );




U_vars := [seq(seq(seq(u[i,j,k], k= 0..U_degree_bound), j=1..n),i=1..n)];
V_vars := [seq(seq(seq(v[i,j,k], k= 0..V_degree_bound), j=1..n),i=1..n)];

vars := [op(DeltaA_vars), op(U_vars), op(S_vars),op(h_vars),op(V_vars)];

obj := poly_norm_opt(DeltaA,t)^2;



u_det_constraint := Vector(PolynomialTools:-CoefficientVector(LinearAlgebra:-Determinant(U)-1,t),n*U_degree_bound+1);

v_det_constraint := Vector(PolynomialTools:-CoefficientVector(LinearAlgebra:-Determinant(V)-1,t),n*V_degree_bound+1);

interface(rtablesize=100):
constraints := <vectorize(A+DeltaA- U.S.H.V,t), u_det_constraint,v_det_constraint>;

Lambda := Vector[row](ArrayNumElems(constraints),j->lambda[j]);

L:= obj + (Lambda.constraints)[1]:

Z:= [op(vars),op(convert(Lambda,list))];

residuals := Vector(VectorCalculus:-Gradient(L,Z)):


#for testing
x_init := random_pt(Z);



U_init,S_init,V_init := LinearAlgebra:-SmithForm(A_init,t,output=['U','S','V']):

h_init_pt := [hh[0]=.9];
S_guess := LinearAlgebra:-IdentityMatrix(n) * LinearAlgebra:-Determinant(subs(t=0,U_init)) * LinearAlgebra:-Determinant(subs(t=0,V_init));

S_init_pt := op(solve([seq( op(Equate(Vector(PolynomialTools:-CoefficientVector(S[i,i],t),sn_degree_bound+1) ,Vector(PolynomialTools:-CoefficientVector(S_guess[i,i],t),sn_degree_bound+1))) ,i=1..n) ],S_vars));



U_init := U_init/LinearAlgebra:-Determinant(subs(t=0,U_init));
V_init := V_init /LinearAlgebra:-Determinant(subs(t=0,V_init));


U_init_pt := op(solve(Equate(vectorize(U-U_init,t),Vector(n^2*(U_degree_bound+1))),U_vars));
V_init_pt := op(solve(Equate(vectorize(V-V_init,t),Vector(n^2*(V_degree_bound+1))),V_vars));



quit;


#DA_init := [seq(DeltaA_vars[i]=0,i=1..nops(DeltaA_vars))];

DA_init := vectorize(A+DeltaA-A_init,t);
DA_init := op(solve(Equate(DA_init,Vector(ArrayNumElems(DA_init))),DeltaA_vars));


X_init := [op(U_init_pt), op(V_init_pt), op(S_init_pt), op(h_init_pt), op(DA_init) ];

AA,bb := LinearAlgebra:-GenerateMatrix(convert(subs(X_init,residuals),list),convert(Lambda,list));

Lambda_init := Optimization:-LSSolve([bb,AA])[2];
Lambda_init := Equate(LinearAlgebra:-Transpose(Lambda),Lambda_init);

norm(subs(Lambda_init,Lambda),2);



#LinearAlgebra:-LeastSquares(subs(X_init,residuals),Lambda);

Z_init := [op(X_init), op(Lambda_init)];

#norm(subs(Z_init,residuals),2);

#quit;

my_soln :=levenberg_marquardt_implicit_merit(residuals,Z,Z_init,1e-14):
 
 subs(my_soln[2],U);
 LinearAlgebra:-Determinant(%);
 
 subs(my_soln[2],V);
 LinearAlgebra:-Determinant(%);
 
 subs(my_soln[2],H);
 subs(my_soln[2],S);
 subs(my_soln[2],DeltaA);

subs(my_soln[2],Lambda);
 
 J:= VectorCalculus:-Jacobian(convert(constraints,Vector),vars):




 
 
 Hxx:=subs(my_soln[2],VectorCalculus:-Hessian(L,vars));
 
 J_ker :=`<|>`( op(LinearAlgebra:-NullSpace(subs(my_soln[2],J)))):
 
sort( Re(LinearAlgebra:-Eigenvalues( LinearAlgebra:-Transpose(J_ker).Hxx.J_ker)));
 
 
 J_DA := VectorCalculus:-Jacobian(convert(constraints,Vector), DeltaA_vars);
 
 J_S :=  VectorCalculus:-Jacobian(convert(constraints,Vector), S_vars):
J_h :=   VectorCalculus:-Jacobian(convert(constraints,Vector), h_vars):
 
 LinearAlgebra:-Dimension(J);
 LinearAlgebra:-Rank(subs(my_soln[2],J));
LinearAlgebra:-Rank(subs(Z_init,J));

LinearAlgebra:-Rank(subs(x_init,J));

LinearAlgebra:-NullSpace(Hxx);



quit;







#some testing.
U_det := Vector(PolynomialTools:-CoefficientVector(LinearAlgebra:-Determinant(U),t), n*(U_degree_bound)+1);

J_U_det:=VectorCalculus:-Jacobian(U_det,U_vars);


V_det := Vector(PolynomialTools:-CoefficientVector(LinearAlgebra:-Determinant(V),t), n*(V_degree_bound)+1);

VectorCalculus:-Jacobian(V_det,V_vars);

#printlevel:=5;

det_U_vec := convert(convert(LinearAlgebra:-Transpose(LinearAlgebra:-Adjoint(U)),Vector[row]),Matrix);


J_U_det_direct := linearize_A_bound( det_U_vec ,t,n*d+1)[1];
J_U_det_direct - J_U_det; #verifies my Chapter 4 in my thesis!



