read("../lib/alpha_optimization.mpl"):
with(AlphaOptimization);
read("../lib/nearest_singular_poly_functions.mpl"):
read("../lib/nearest_smith_functions.mpl"):
#with(NearestPoly);
with(NearestSmith);
read("../lib/rank-factor-lib.mpl");




with(NearestPoly):
with(AlphaOptimization):
with(NearestSmith):

kernelopts(printbytes=false):


#proc: compute_initial_h_guess
#Input: Matrix polynomial A in parameter $t$ with vector of supports alpha
#       mode: "real" or "complex" - used to guess if a solution is real or not. 
#Output:


compute_initial_h_guess := proc(A,t,Alpha, tol:=1e-7)
        local s_rank,i,omega_init,h_init,my_local_extrema,my_guess_index;
        s_rank := 0:

        for i from 1 to ArrayNumElems(Alpha) do
            if Alpha[i] <> 0 then s_rank := s_rank +1; fi:
        od:

        #get a list of the local extrema, these are suitable guesses
        my_local_extrema := [fsolve( diff(LinearAlgebra:-Determinant(A)^2,t)=0,complex)];
        #take a point that gives us something of approximately the desired McCoy rank
        my_guess_index := min[index]([seq(LinearAlgebra:-SingularValues(subs(t=my_local_extrema[i],evalf(A)))[LinearAlgebra:-RowDimension(A)-s_rank+1] ,i=1..nops(my_local_extrema))]);


        omega_init := my_local_extrema[my_guess_index]; #pick a guess.

        #tol:= 1e-8; #tolerance for degree 1 or degree 2 guesses

        if abs( Im(omega_init)) < tol then 
            h_init := t-Re(omega_init);
            h_vars := [hh[0]];
            h_init_pt := [hh[0] = subs(t=0,h_init)];
            h := hh[0]+t;
        else 
              h_init := Re(evalc((t-omega_init)*(t-conjugate(omega_init)))) assuming real; #this should be real!
; #this should be real!
            h_vars := [hh[0],hh[1]];
            h_init_pt := [hh[0] = subs(t=0,h_init), hh[1] = subs(t=0,diff(h_init,t))];
            h:= hh[0]+hh[1]*t+t^2;
        fi:

    return [h_init,h_vars,h_init_pt,h];
end proc:

#Proc: compute_initial_S_guess
#Input: A,t, H_init
#output: initial guess in S
compute_initial_S_guess := proc(A,t,Alpha,h_init)
        local   SS,S_init,S_init_pt,S_vars,
                m,n,d,s_rank, sn_degree_bound,detA
                ;
                
                m,n := LinearAlgebra:-Dimension(A);
                d:= matrix_degree(collect(A,t),t);
                s_rank := add(Alpha[j],j=1..ArrayNumElems(Alpha));
                
                
                sn_degree_bound := n*d - s_rank; #the bound of the last invariant factor divided out by h^alpha
                
                 
                 
                sn := add( ss[n,j]*t^j,j=0..sn_degree_bound);
                s_degree_bound := 0; #assume these guys are constants for now!
                s_vec := Vector(n,j-> add(ss[j,i]*t^i,i=0..s_degree_bound));
                s_vec[n] := sn ;
                SS:= LinearAlgebra:-DiagonalMatrix(s_vec); #create ANSATZ for $S$

                 
                #create a list of variables for the first n-1 entries
                S_vars := [ seq( op(convert(PolynomialTools:-CoefficientVector(SS[j,j],t),list)) ,j=1..n-1) ];
                #create a list of variables for the last guy
                sn_vars := [op(convert(PolynomialTools:-CoefficientVector(sn,t),list))];
                S_vars := [op(S_vars), op(sn_vars)];
                
                #the entries start off as 1 and are re-scaled by rescaling the unimodular multipliers.
                S_init_pt := [seq( ss[i,0]=1, i = 1..n-1)];
                
                sn_init := 1;
                detA:= LinearAlgebra:-Determinant(A);
                
                        #use my least squares polynomial division routine to get this guy
                if degree(h_init,t) + s_rank-1 <= degree(detA,t) then 
                    sn_init:= op(poly_least_squares_division_implicit ([evalf(detA)],collect(h_init^s_rank,t),t)[2]); 
                    
#                    print(degree(detA,t));
 #                   print(degree(collect(h_init^s_rank,t),t));
                    
  #                  print((poly_least_squares_division_implicit ([evalf(detA)],collect(h_init^s_rank,t),t)));
                fi:
                #rescale this guy reasonably...
                sn_init := sn_init/lcoeff(sn_init,t);
                
                
                
                sn_init_pt := Equate(Vector(PolynomialTools:-CoefficientVector(sn,t), sn_degree_bound+1),Vector(PolynomialTools:-CoefficientVector(sn_init,t), sn_degree_bound+1));
        #        sn_init_pt := op(solve(vectorize(sn-sn_init,t),sn_vars));
                
                S_init_pt := [op(S_init_pt), op(sn_init_pt)];
                
                return SS,S_init_pt,S_vars;
        
end proc:



#Proc: compute_initial_USV_guess
#Input: Matrix polynomial A in indeterminant $t$ and initial guess for h,  $h_init$ with vector of degrees
#Output: initial guesses for U  and SV
#        Some more information
#Note that this uses an exact procedure, which may not be desirable. 
compute_initial_UV_guess := proc(A,t,h_init,Alpha)

        local   UU,U_init,U_init_pt,U_vars,
                SS,S_init,S_init_pt,S_vars,
                VS,V_init,V_init_pt,V_vars,
                m,n,d,s_rank,
                my_guess_index,
                U_degree_bound,V_degree_bound,sn_degree_bound,
                U_scale_factor,V_scale_factor,
                H,
                sn_vars,sn,s_vec,sn_init,sn_init_pt,
                USHV, #product of U,S,H and V in ANSATZ form.
                      #this expression has a polynomial blow up and can be handled with symbolic differentiation.
                Y, #U,S,V  
                Y_vars, #vars for U,S,V
                Y_init# initial USV init point
                ;
                m,n := LinearAlgebra:-Dimension(A);
        
        m,n := LinearAlgebra:-Dimension(A);
        d:= matrix_degree(collect(A,t),t);
        s_rank := add(Alpha[j],j=1..ArrayNumElems(Alpha));

#        U_degree_bound := n*d;
#        V_degree_bound := n*(d+1); #is this bound right?
       
        U_init,S_init,V_init := LinearAlgebra:-SmithForm(A,t,output=['U','S','V']):
 
 
        U_degree_bound := max(0,matrix_degree(U_init,t));
        V_degree_bound := max(0,matrix_degree(V_init,t)); #is this bound right?
       
        
        UU := Matrix(n,n, (i,j)-> add( u[i,j,k]*t^k, k=0..max(0,degree(U_init[i,j],t)) ) );
        VV := Matrix(n,n, (i,j)-> add( v[i,j,k]*t^k, k=0..max(0,degree(V_init[i,j],t)) ) );
        
        U_vars := [seq(seq(seq(u[i,j,k], k= 0..max(0,degree(U_init[i,j],t))), j=1..n),i=1..n)];
        V_vars := [seq(seq(seq(v[i,j,k], k= 0..max(0,degree(V_init[i,j],t))), j=1..n),i=1..n)];
        
        
        #get scaling factors.  
        
        U_scale_factor :=1;#LinearAlgebra:-Determinant(subs(t=0,U_init));
        V_scale_factor :=1;# LinearAlgebra:-Determinant(subs(t=0,V_init));
        
        #now they have determinant 1
        #and are feasible w.r.t. to our optimization. 
        #U_init := (U_init/(U_scale_factor^(1/n))); 
        
        
        #V_init := (V_init/V_scale_factor^(1/n));
        
        U_init_pt :=vectorize(UU-U_init,t);
        U_init_pt := op(solve(Equate(U_init_pt,Vector(ArrayNumElems(U_init_pt))),U_vars));
        
        
        
        #print(matrix_degree(VV,t));
        #print(matrix_degree(V_init,t));
        
        V_init_pt :=vectorize(VV-V_init,t);
        V_init_pt := op(solve(Equate(V_init_pt,Vector(ArrayNumElems(V_init_pt))),V_vars));
        
     
        ##U and V initial guess code is basically done here. 
        
        
        sn_degree_bound := n*d - s_rank; #the bound of the last invariant factor divided out by h^alpha
        
       # print(s_rank);
      #  print(n*d);
      #  print(sn_degree_bound);
        
        
        sn := add( ss[n,j]*t^j,j=0..sn_degree_bound);
        
        s_degree_bound := 0; #assume these guys are constants for now!
        s_vec := Vector(n,j-> add(ss[j,i]*t^i,i=0..s_degree_bound));
        s_vec[n] := sn ;
        SS:= LinearAlgebra:-DiagonalMatrix(s_vec); #create ANSATZ for $S$

        
        #create a list of variables for the first n-1 entries
        S_vars := [ seq( op(convert(PolynomialTools:-CoefficientVector(SS[j,j],t),list)) ,j=1..n-1) ];
        #create a list of variables for the last guy
        sn_vars := [op(convert(PolynomialTools:-CoefficientVector(sn,t),list))];
        S_vars := [op(S_vars), op(sn_vars)];
        
        #the entries start off as 1 and are re-scaled by rescaling the unimodular multipliers.
        S_init_pt := [seq(S[i,i]=evalf(U_scale_factor*V_scale_factor), i = 1..n-1)];
        
        sn_init := 1;
        
        
        detA:= LinearAlgebra:-Determinant(A);
        #use my least squares polynomial division routine to get this guy
        if degree(h_init,t) + s_rank-1 <= degree(detA,t) then 
            sn_init:= op(poly_least_squares_division_implicit ([evalf(detA)],collect(h_init^s_rank,t),t)[2]); 
            
            #print(degree(detA,t));
          #  print(degree(collect(h_init^s_rank,t),t));
            
         #   print((poly_least_squares_division_implicit ([evalf(detA)],collect(h_init^s_rank,t),t)));
        fi:
        #rescale this guy reasonably...
        sn_init := sn_init/lcoeff(sn_init,t);
        
        
        
        sn_init_pt := Equate(Vector(PolynomialTools:-CoefficientVector(sn,t), sn_degree_bound+1),Vector(PolynomialTools:-CoefficientVector(sn_init,t), sn_degree_bound+1));
#        sn_init_pt := op(solve(vectorize(sn-sn_init,t),sn_vars));
        
        S_init_pt := [op(S_init_pt), op(sn_init_pt)];
        #list of variables is now complete. 

        Y_init := [op(U_init_pt), op(S_init_pt), op(V_init_pt)];
        Y_vars := [op(U_vars), op(S_vars), op(V_vars)];
        Y := [UU,SS,VV];
        
        return evalf([Y,Y_init,Y_vars]);
        
end proc:



#Proc: compute_initial_USV_guess
#Input: Matrix polynomial A in indeterminant $t$ and initial guess for h,  $h_init$ with vector of degrees
#Output: initial guesses for U  and SV
#        Some more information
#Note that this uses an exact procedure, which may not be desirable. 
compute_initial_UV_guess_dense := proc(A,t,h_init,Alpha)

        local   UU,U_init,U_init_pt,U_vars,
                SS,S_init,S_init_pt,S_vars,
                VS,V_init,V_init_pt,V_vars,
                m,n,d,s_rank,
                my_guess_index,
                U_degree_bound,V_degree_bound,sn_degree_bound,
                U_scale_factor,V_scale_factor,
                H,
                sn_vars,sn,s_vec,sn_init,sn_init_pt,
                USHV, #product of U,S,H and V in ANSATZ form.
                      #this expression has a polynomial blow up and can be handled with symbolic differentiation.
                Y, #U,S,V  
                Y_vars, #vars for U,S,V
                Y_init# initial USV init point
                ;
                m,n := LinearAlgebra:-Dimension(A);
        
        m,n := LinearAlgebra:-Dimension(A);
        d:= matrix_degree(collect(A,t),t);
        s_rank := add(Alpha[j],j=1..ArrayNumElems(Alpha));

#        U_degree_bound := n*d;
#        V_degree_bound := n*(d+1); #is this bound right?
       
        U_init,S_init,V_init := LinearAlgebra:-SmithForm(A,t,output=['U','S','V']):
 
 
        U_degree_bound := max(0,matrix_degree(U_init,t));
        V_degree_bound := max(0,matrix_degree(V_init,t)); #is this bound right?
       
        
        UU := Matrix(n,n, (i,j)-> add( u[i,j,k]*t^k, k=0..max(0,U_degree_bound) ) );
        VV := Matrix(n,n, (i,j)-> add( v[i,j,k]*t^k, k=0..max(0,V_degree_bound) ) );
        
        U_vars := [seq(seq(seq(u[i,j,k], k= 0..max(0,U_degree_bound)), j=1..n),i=1..n)];
        V_vars := [seq(seq(seq(v[i,j,k], k= 0..max(0,V_degree_bound)), j=1..n),i=1..n)];
        
        
        #get scaling factors.  
        
        U_scale_factor :=1;#LinearAlgebra:-Determinant(subs(t=0,U_init));
        V_scale_factor :=1;# LinearAlgebra:-Determinant(subs(t=0,V_init));
        
        #now they have determinant 1
        #and are feasible w.r.t. to our optimization. 
        #U_init := (U_init/(U_scale_factor^(1/n))); 
        
        
        #V_init := (V_init/V_scale_factor^(1/n));
        
        U_init_pt :=vectorize(UU-U_init,t);
        U_init_pt := op(solve(Equate(U_init_pt,Vector(ArrayNumElems(U_init_pt))),U_vars));
        
        
        
        #print(matrix_degree(VV,t));
        #print(matrix_degree(V_init,t));
        
        V_init_pt :=vectorize(VV-V_init,t);
        V_init_pt := op(solve(Equate(V_init_pt,Vector(ArrayNumElems(V_init_pt))),V_vars));
        
     
        ##U and V initial guess code is basically done here. 
        
        
        sn_degree_bound := n*d - s_rank; #the bound of the last invariant factor divided out by h^alpha
        
       # print(s_rank);
      #  print(n*d);
      #  print(sn_degree_bound);
        
        
        sn := add( ss[n,j]*t^j,j=0..sn_degree_bound);
        
        s_degree_bound := 0; #assume these guys are constants for now!
        s_vec := Vector(n,j-> add(ss[j,i]*t^i,i=0..s_degree_bound));
        s_vec[n] := sn ;
        SS:= LinearAlgebra:-DiagonalMatrix(s_vec); #create ANSATZ for $S$

        
        #create a list of variables for the first n-1 entries
        S_vars := [ seq( op(convert(PolynomialTools:-CoefficientVector(SS[j,j],t),list)) ,j=1..n-1) ];
        #create a list of variables for the last guy
        sn_vars := [op(convert(PolynomialTools:-CoefficientVector(sn,t),list))];
        S_vars := [op(S_vars), op(sn_vars)];
        
        #the entries start off as 1 and are re-scaled by rescaling the unimodular multipliers.
        S_init_pt := [seq(S[i,i]=evalf(U_scale_factor*V_scale_factor), i = 1..n-1)];
        
        sn_init := 1;
        
        
        detA:= LinearAlgebra:-Determinant(A);
        #use my least squares polynomial division routine to get this guy
        if degree(h_init,t) + s_rank-1 <= degree(detA,t) then 
            sn_init:= op(poly_least_squares_division_implicit ([evalf(detA)],collect(h_init^s_rank,t),t)[2]); 
            
            #print(degree(detA,t));
          #  print(degree(collect(h_init^s_rank,t),t));
            
         #   print((poly_least_squares_division_implicit ([evalf(detA)],collect(h_init^s_rank,t),t)));
        fi:
        #rescale this guy reasonably...
        sn_init := sn_init/lcoeff(sn_init,t);
        
        
        
        sn_init_pt := Equate(Vector(PolynomialTools:-CoefficientVector(sn,t), sn_degree_bound+1),Vector(PolynomialTools:-CoefficientVector(sn_init,t), sn_degree_bound+1));
#        sn_init_pt := op(solve(vectorize(sn-sn_init,t),sn_vars));
        
        S_init_pt := [op(S_init_pt), op(sn_init_pt)];
        #list of variables is now complete. 

        Y_init := [op(U_init_pt), op(S_init_pt), op(V_init_pt)];
        Y_vars := [op(U_vars), op(S_vars), op(V_vars)];
        Y := [UU,SS,VV];
        
        return evalf([Y,Y_init,Y_vars]);
        
end proc:


#Proc: compute_USV_guess_LS
#Input: A, S, t
#Output: Guesses for U and V such that UAV \approx S
#Notes: This assumes $S$ as a priori and attempts to compute in a least squares sense approximatly 
#       unimodular $U$ and $V$. The goal is to set this up as an optimization problem.

compute_USV_guess_LS := proc(A,t,S,degree_bound)
        local   UU,U_init,U_init_pt, 
                VV,V_init,V_init_pt,
                m,n,d,
                U_degree_bound,V_degree_bound,
                obj,
                V_constraint, V0
                ;
                
                
                m,n := LinearAlgebra:-Dimension(A);
                d := matrix_degree(A,t);
                
                #nothing fancy for now.
                U_degree_bound := degree_bound;
                V_degree_bound := degree_bound; 
                
                UU := Matrix(n,n, (i,j)-> add( u[i,j,k]*t^k, k=0..U_degree_bound ) );
                
                
                U_vars := [seq(seq(seq(u[i,j,k], k= 0..U_degree_bound), j=1..n),i=1..n)];
                
                VV := Matrix(n,n);
                
                V_vars := [];
                for i from 1 to n do 
                    for j from i+1 to n do 
                        VV[i,j]:= add( v[i,j,k]*t^k, k=0..V_degree_bound );
                        V_vars := [op(V_vars), seq(v[i,j,k],k=0..V_degree_bound)];
                    od:
                od:
                for i from 1 to n do 
                    VV[i,i] := 1; #normalize diagonal to force it to be unimodular
                od:
                
                
                obj := convert(vectorize(UU.A-S.VV,t),list);
                
                
                
                my_soln := Optimization:-LSSolve(obj);
                
                return UU,VV, [op(U_vars), op(V_vars)] ,my_soln[2],my_soln[1];
                
end proc:



#tweak this example.
A:=Matrix(3, 3, [[4*t^4-9*t^3+17*t^2-23*t+11, 4*t^4-4*t^3+8*t^2-20*t+12, -7/5+(28/5)*t^3-(63/5)*t^2+(42/5)*t], [-3/2-(7/2)*t^4-5*t^3+21*t^2-11*t, -8-(7/2)*t^4-(15/2)*t^3+(31/2)*t^2+(7/2)*t, -77/10-(49/10)*t^3+(7/2)*t^2+(91/10)*t], [4*t^4-4*t^3-9*t^2+12*t-3, 4*t^4+t^3-8*t^2-t+4, 7+(28/5)*t^3-(28/5)*t^2-7*t]]);

#this example has a different SNF, and the supports are different.

#A := Matrix(3, 3, [[-3*t^3+7*t^2+23*t+13, 3*t^3+12*t^2+15*t+6, 7/5+(7/5)*t^2+(14/5)*t], [-3/2+2*t^4-(7/2)*t^3-(29/2)*t^2-(21/2)*t, -8+2*t^4-(11/2)*t^3-25*t^2-(51/2)*t, -77/10+(14/5)*t^3-(21/10)*t^2-(63/5)*t], [t^3-t^2-5*t-3, 7*t^3+18*t^2+15*t+4, 7*t^2+14*t+7]]);

A_init := A;







DeltaA,DeltaA_vars := generate_perturbation(A,t,"support");

A_noise := random_pt(DeltaA_vars);


A:= evalf(A+1e-3*subs(A_noise,DeltaA),10);



Alpha := <1,2,2>;  #degrees of h


m,n := LinearAlgebra:-Dimension(A);
d:= matrix_degree(A,t);

  
  
        


h_init,h_vars,h_init_pt,h := op(compute_initial_h_guess(A_init,t,Alpha));

#printlevel:=2;
SS,S_init_pt,S_vars := compute_initial_S_guess(A_init,t,Alpha,h_init);

H := LinearAlgebra:-DiagonalMatrix( Vector(n, j->h^Alpha[j]));

S_guess := subs(S_init_pt, SS). subs(h_init_pt,H);

UU,VV, my_vars, my_soln, residual := compute_USV_guess_LS(A_init,t,S_guess,n*d):

residual;

LinearAlgebra:-Determinant(subs(my_soln,UU));

V_det := LinearAlgebra:-Determinant(subs(my_soln,VV));

V_inv :=  1/V_det * LinearAlgebra:-Adjoint(subs(my_soln,VV)): #V_det is 1 via normalization

clean_up(subs(my_soln,UU).A_init.V_inv,t,1e-8);



