#Author: Joseph Haraldson
#Version ALPHA 0
#alpha_optimization.mpl
#rudimentary implementation of different optimization techniques for (un)constrained optimization problems.
#These are implementations of different optimization algorithms in Maple. The main purpose is rapid prototyping of algorithms.

AlphaOptimization:= module()
option package;
export   levenberg_marquardt_implicit_merit, newton_vanilla;
#constrained methods...

#Gradient algorithms...
#let's try to  




#Second order algorithms...
#we can basically copy/paste newton from other code...


#proc: newton_vanilla
#Input: objective function and residuals, some other parameters
#Output: result of a newton iteration
#Notes: this isn't regularized or particularily optimized. It uses a truncated SVD to solve the arising linear system. This may not always be wise. 


newton_vanilla := proc(residuals,vars,x_init,tol:=1e-16,max_iter := 100000)
             local i,
	    grad,
	    J,
	    rhs_k, #first argument to linear solve
	    lhs_k,#second rgument to linear solve
	    current_pt,
	    x_next,
	    x_cur,
	    mu_k#the current levenberg-marquardt parameter
	    ;
	    
	    J := VectorCalculus:-Jacobian(residuals,vars);
	   
	  
	    
	    x_cur := subs(x_init, Vector(vars));
	    
	    #printf("Beginning LM Iteration\n");
	    for i from 1 to max_iter do
		
		current_pt := subs(Equate(Vector(vars),x_cur));
		
		
		#J_cur := subs(current_pt,J);
		rhs_k := subs(current_pt,J);
		
		lhs_k := -1*subs(current_pt,residuals);
		
		dx := LinearAlgebra:-LeastSquares(rhs_k,lhs_k, optimize=true);
		
		x_next_candidate := Equate(Vector(vars),x_cur + dx);
		
		    
		    x_next := x_cur + dx;
	
		printf("k = %d,  ||dx|| = %e, ||f(x_k)||=%e\n",i, norm(dx,2), norm(subs(current_pt,residuals),2));
		
		    if norm(x_cur-x_next,2) < tol or norm(subs(current_pt,residuals),2) < tol   then break fi:
		    
		    if norm(subs(current_pt,residuals),2) < norm(subs(x_next_candidate,residuals),2) then 
                        printf("Warning: ||f(x_k)|| < ||f(x_k+1)||: %e < %e Improvement not monotone. Aborting\n", norm(subs(current_pt,residuals),2),norm(subs(x_next_candidate,residuals),2)  );
                        break;
		    fi:
		
		    x_cur := x_next;
		
		
		
	    od:
	    
	    current_pt := subs(Equate(Vector(vars),x_cur));
	    printf("Newton returning with ||f(x_k)|| %e  and k=%d iterations.\n",norm(subs(current_pt,residuals),2) , i);
	    
	    return [norm(subs(Equate(vars,x_cur),residuals),2) , Equate(vars,x_cur)];
                    
end proc:



#Proc: levenberg_marquardt_implicit_merit
#Input: objective function and residuals, some other parameters
#Output: result of a Levenberg-Marquardt iteration
#Notes: implicit merit function being norm of residuals.

levenberg_marquardt_implicit_merit := proc(residuals,vars,x_init,tol:=1e-16,max_iter := 100000,lm_param:= 1,lm_amp_factor:=1e-1,lm_damp_factor:=1e1)
      local i,
	    grad,
	    J,
	    JTJ,
	    rhs_k, #first argument to linear solve
	    lhs_k,#second rgument to linear solve
	    current_pt,
	    x_next,
	    I,
	    x_cur,
	    mu_k#the current levenberg-marquardt parameter
	    ;
	    
	    J := VectorCalculus:-Jacobian(residuals,vars);
	    #JTJ := LinearAlgebra:-Transpose(J).J;
            
	    Ident:=Matrix(LinearAlgebra:-ColumnDimension(J));
	    for j from 1 to LinearAlgebra:-RowDimension(Ident) do 
		Ident[j,j]:=1:
	    od:
	    
	  #  printf("Initial guess has dimension %d,%d and rank %d\n", LinearAlgebra:-Dimension(JTJ),LinearAlgebra:-Rank(subs(x_init,JTJ)));

	  
	    
	    x_cur := subs(x_init, Vector(vars));
	    
	    mu_k := lm_param;
	    #printf("Beginning LM Iteration\n");
	    for i from 1 to max_iter do
		
		current_pt := subs(Equate(Vector(vars),x_cur));
		
		#mu_k := max(norm(subs(current_pt,residuals),1),lm_param);
		
		J_cur := subs(current_pt,J);
		JTJ := LinearAlgebra:-Transpose(J_cur).J_cur; 
		
		rhs_k := subs(current_pt,JTJ + mu_k * Ident);
		
		lhs_k := -1*LinearAlgebra:-Transpose(J_cur).subs(current_pt,residuals);
		
		dx := LinearAlgebra:-LinearSolve(rhs_k,lhs_k);
		
		x_next_candidate := Equate(Vector(vars),x_cur + dx);
		
		if norm(subs(x_next_candidate, residuals),2) <= norm(subs(current_pt,residuals),2)  then #valid step 
		    #mu_k := max(mu_k*(lm_amp_factor), 10^(-.5*Digits)); #take a smaller step next iteration
		   mu_k := max(mu_k*(lm_amp_factor), 0); #take a smaller step next iteration
		   
		   
		    #this is based on a criteria from a paper by Wright.
		    #|| residuals||_1 approximates the solution, so the penalty term
		    #blows up at the same rate as the solution, giving asymptotic quadratic convergence.
		    if norm(dx,2)<  norm(subs(x_next_candidate,residuals),2) and  norm(subs(x_next_candidate,residuals),2) <= 1.5 * norm(dx,2)   then 
			mu_k := norm(subs(x_next_candidate,residuals),2);
		    fi:
		    
		    x_next := x_cur + dx;
		    printf("k = %d, mu=%e, ||x_k - x_k+1|| = %e, ||dx|| = %e, ||f(x_k)||=%e\n",i,mu_k,norm(x_cur - x_next,2), norm(dx,2), norm(subs(current_pt,residuals),2));
		
		    if norm(x_cur-x_next,2) < tol or norm(subs(current_pt,residuals),2) < tol then break fi:
		
		    x_cur := x_next;
		
		
		else 
		    mu_k := mu_k * (lm_damp_factor);  #take a smaller step and try again
		fi:
		if mu_k >= 10^Digits then
                    WARNING("Unable to find a suitable step for improvement. Terminating.");
		break fi:
		
		
		
	    od:
	    
	    current_pt := subs(Equate(Vector(vars),x_cur));
	    return [norm(subs(Equate(vars,x_cur),residuals),2) , Equate(vars,x_cur)];
	    
end proc:

#unconstrained methods...


#Proc: levenberg_marquardt
#Input: objective function and residuals, some other parameters
#Output: result of a Levenberg-Marquardt iteration
#Notes: I have no fucking clue if this will work.

levenberg_marquardt := proc(obj,residuals,vars,x_init,tol:=1e-16,max_iter := 100000,lm_param:= 1,lm_amp_factor:=1e-1,lm_damp_factor:=1e1)
      local i,
	    grad,
	    J,
	    JTJ,
	    rhs_k, #first argument to linear solve
	    lhs_k,#second rgument to linear solve
	    current_pt,
	    x_next,
	    I,
	    x_cur,
	    mu_k#the current levenberg-marquardt parameter
	    ;
	    
	    J := VectorCalculus:-Jacobian(residuals,vars);
	    JTJ := LinearAlgebra:-Transpose(J).J;

	    Ident:=Matrix(LinearAlgebra:-Dimension(JTJ));
	    for j from 1 to LinearAlgebra:-RowDimension(Ident) do 
		Ident[j,j]:=1:
	    od:
	    
	  #  printf("Initial guess has dimension %d,%d and rank %d\n", LinearAlgebra:-Dimension(JTJ),LinearAlgebra:-Rank(subs(x_init,JTJ)));

	  
	    
	    x_cur := subs(x_init, Vector(vars));
	    
	    mu_k := lm_param;
	    for i from 1 to max_iter do
		
		current_pt := subs(Equate(Vector(vars),x_cur));
		
		#mu_k := max(norm(subs(current_pt,residuals),1),lm_param);
		
		
		rhs_k := subs(current_pt,JTJ + mu_k * Ident);
		
		lhs_k := subs(current_pt,-1*LinearAlgebra:-Transpose(J).residuals);
		
		dx := LinearAlgebra:-LinearSolve(rhs_k,lhs_k);
		
		x_next_candidate := Equate(Vector(vars),x_cur + dx);
		
		if subs(x_next_candidate, obj) <= subs(current_pt,obj)  then #valid step 
		    mu_k := max(mu_k*(lm_amp_factor), 10^(-.5*Digits)); #take a smaller step next iteration
		   
		   
		    #this is based on a criteria from a paper by Wright.
		    #|| residuals||_1 approximates the solution, so the penalty term
		    #blows up at the same rate as the solution, giving asymptotic quadratic convergence.
		    if norm(dx,2)<  subs(x_next_candidate,obj) and  subs(x_next_candidate,obj) <= 1.5 * norm(dx,2)   then 
			mu_k := norm(subs(x_next_candidate,residuals),2);
		    fi:
		    
		    x_next := x_cur + dx;
		    printf("k = %d, mu=%e, ||x_k - x_k+1|| = %e, ||dx|| = %e, ||f(x_k)||=%e\n",i,mu_k,norm(x_cur - x_next,2), norm(dx,2), sqrt(subs(current_pt,obj)));
		
		    if norm(x_cur-x_next,2) < tol or sqrt(subs(current_pt,obj)) < tol then break fi:
		
		    x_cur := x_next;
		
		
		else 
		    mu_k := mu_k * (lm_damp_factor);  #take a smaller step and try again
		fi:
		
		
		
	    od:
	    
	    current_pt := subs(Equate(Vector(vars),x_cur));
	    return [sqrt(subs(Equate(vars,x_cur),obj)) , Equate(vars,x_cur)];
	    
end proc:

#Proc: gauss_newton
#Input: objective function and residuals, some other parameters
#Output: result of a gauss-newton minimization
#Notes: I have no fucking clue if this will work.

gauss_newton := proc(obj,residuals,vars,x_init,tol:=1e-8,max_iter := 100)
      local i,
	    grad,
	    J,
	    JTJ,
	    rhs_k, #first argument to linear solve
	    lhs_k,#second rgument to linear solve
	    GN_step,
	    current_pt,
	    x_next,
	    x_cur
	    ;
	    
	    J := VectorCalculus:-Jacobian(residuals,vars);
	    JTJ := LinearAlgebra:-Transpose(J).J;

	    
	    printf("Initial guess has dimension %d,%d and rank %d\n", LinearAlgebra:-Dimension(JTJ),LinearAlgebra:-Rank(subs(x_init,JTJ)));

	    x_cur := subs(x_init, Vector(vars));
	    
	    for i from 1 to max_iter do
		
		current_pt := subs(Equate(Vector(vars),x_cur));
		rhs_k := subs(current_pt,JTJ);
		
		lhs_k := subs(current_pt,-1*LinearAlgebra:-Transpose(J).residuals);
		
		dx := LinearAlgebra:-LinearSolve(rhs_k,lhs_k);
		
		x_next := x_cur + dx;
		printf("k = %d, ||x_k - x_k+1|| = %e, ||dx|| = %e, ||f(x_k)||=%e\n",i,norm(x_cur - x_next,2), norm(dx,2), sqrt(subs(current_pt,obj)));
		
		if norm(x_cur-x_next,2) < tol or sqrt(subs(current_pt,obj)) < tol then break fi:
		
		
		
		
		x_cur := x_next;
		
	    od:
	    
	    return [subs(Equate(vars,x_cur),obj) , Equate(vars,x_cur)];
	    
end proc:



#Proc: gradient_descent
#Input: Objective function with vars and step size parameter eta.
#Notes: Tolerance for a solution is low (8 decimal points) because gradient methods are shitty.

gradient_steepest_descent := proc(obj,vars,x_init,step_size:=1e-2,tol:=1e-8,max_iter := 1000)
      local i,
	    grad,
	    step,
	    x_cur,
	    vars_vec,
	    x_next
	    ;
	    
	    grad := VectorCalculus:-Gradient(obj,vars); #compute the gradient 
	    
	    vars_vec := Vector(vars);
	    
	    x_cur := subs(x_init, Vector(vars));
	    
	    
	    for i from 1 to max_iter do 
	
		  #The step computation is the bottle beck.
		  #The number of arithemtic operations here is pretty bad
		  step :=  (-1)*step_size* subs( Equate(vars_vec,x_cur),grad); 
		  x_next := x_cur +step; 
		  
		  
		  if i=1 or i mod 50 = 0 then 
		      printf("k=%d,  ||x_k-x_k+1||=%e  , ||grad(x_k)||=%e, ||f(x_k)||=%e\n", i,norm(subs( Equate(vars_vec,x_cur),grad),2),norm(x_cur-x_next,2),subs( Equate(vars_vec,x_cur),obj) );
		  fi:
		  
		  #3 different stopping criteria, all are mostly equivalent if the objective is Lipschitz
		  if 	norm(x_cur-x_next,2) < tol or norm(subs( Equate(vars_vec,x_cur),grad),2) < tol or 
			subs( Equate(vars_vec,x_cur),obj)<tol
			
			then break;
		  fi:
		  
		  if i = max_iter then 
		      WARNING("Maximum number of iterations achieved\n");
		      break;
		  fi:
		  x_cur := x_next; #update after checking for convergence.
		  
		  #debugging symbols
		 # printf("k=%d,  ||x_k-x_k+1||=%e  , ||grad(x_k)||=%e, ||f(x_k)||=%e\n", i,norm(subs( Equate(vars_vec,x_cur),grad),2),norm(x_cur-x_next,2),subs( Equate(vars_vec,x_cur),obj) );
		  
	    od:
	    #WARNING("performed maximum number of iterations. Aborting.");
	    
	    
	    return [subs( Equate(vars_vec,x_cur),obj), Equate(vars_vec,x_cur), i,norm(subs( Equate(vars_vec,x_cur),grad),2),norm(x_cur-x_next,2)  ];
end proc:

#Proc: gradient_armijo_descent
#Input: Objective function with vars and step size parameter eta.
#Notes: Tolerance for a solution is low (8 decimal points) because gradient methods are shitty.

gradient_armijo_descent := proc(obj,vars,x_init,step_size:=1e-2,tol:=1e-8,max_iter := 1000)
      local i,
	    j,
	    grad,
	    step,
	    x_cur,
	    vars_vec,
	    eta,#step param
	    h, #armijo step function
	    x_next
	    ;
	    
	    grad := VectorCalculus:-Gradient(obj,vars); #compute the gradient 
	    
	    vars_vec := Vector(vars);
	    
	    x_cur := subs(x_init, Vector(vars));
	    
	    
	    for i from 1 to max_iter do 
	
		  #The step computation is the bottle beck.
		  #The number of arithemtic operations here is pretty bad
		  step :=  (-1)*eta* subs( Equate(vars_vec,x_cur),grad); 
		  
		  h := subs(Equate(vars_vec,x_cur+step),obj);  
		  
		  
		  
		  eta_k := [fsolve(diff(h,eta)=0,eta)]; #need to find maximizers
		  
		  eta_cur := eta_k[1];
		  #max[index] not support in my version...
		  for j from 2 to nops (eta_k) do 
		      if subs(eta_k[j], h) < subs(eta_cur,h) then 
			  eta_cur := eta_k[j];
		      fi:
		  od:
		  eta_k := eta_cur;
		  
		  x_next := x_cur +subs(eta=eta_k,step); 
		  
		  if i=1 or i mod 50 = 0 then 
		      printf("k=%d,  eta_k = %e, ||x_k-x_k+1||=%e  , ||grad(x_k)||=%e, ||f(x_k)||=%e\n", i,eta_k,norm(subs( Equate(vars_vec,x_cur),grad),2),norm(x_cur-x_next,2),subs( Equate(vars_vec,x_cur),obj) );
		  fi:
		  
		  #3 different stopping criteria, all are mostly equivalent if the objective is Lipschitz
		  if 	norm(x_cur-x_next,2) < tol or norm(subs( Equate(vars_vec,x_cur),grad),2) < tol or 
			sqrt(subs( Equate(vars_vec,x_cur),obj))<tol
			
			then break;
		  fi:
		  
		  if i = max_iter then 
		      WARNING("Maximum number of iterations achieved\n");
		      break;
		  fi:
		  x_cur := x_next; #update after checking for convergence.
		  
		  #debugging symbols
		 # printf("k=%d,  ||x_k-x_k+1||=%e  , ||grad(x_k)||=%e, ||f(x_k)||=%e\n", i,norm(subs( Equate(vars_vec,x_cur),grad),2),norm(x_cur-x_next,2),subs( Equate(vars_vec,x_cur),obj) );
		  
	    od:
	    #WARNING("performed maximum number of iterations. Aborting.");
	    
	    
	    return [sqrt(subs( Equate(vars_vec,x_cur),obj)), Equate(vars_vec,x_cur), i,norm(subs( Equate(vars_vec,x_cur),grad),2),norm(x_cur-x_next,2)  ];
end proc:




#Proc: initial_guess_lagrangian
#Input: stuff we need to compute the Lagrangian (Gradient, initial guess, lambda variables)
#Output: initial guess for Lagrange multipliers (computed via linear least squares approximation)

initial_guess_lagrangian := proc(Gx,X_init,Lambda)
	local lagrangian_system,
	      my_eqns,
	      cf_mat,
	      cf_vec,
	      ineq_cons,
	      ls_soln,
	      Lambda_init
	      ; 

    #solving a linear system for initial guess
	lagrangian_system :=  convert(subs(X_init,Gx),Vector);
	my_eqns := Equate(lagrangian_system, Vector(ArrayNumElems(lagrangian_system)));
	cf_mat,cf_vec := LinearAlgebra:-GenerateMatrix(my_eqns,convert(Lambda,list));
	

	
	ls_soln := Optimization:-LSSolve([cf_vec,cf_mat])[2];
	Lambda_init := Equate(convert(Lambda,Vector), ls_soln);

	
	return Lambda_init;
    

end proc:

end module;