###Nearest singular polynomial matrix helper functions
####Author: Joseph Haraldson
###Purpose: Utility functions used to compute the nearest singular structured matrix.

NearestPoly:=module()
description "some functions for nearest singular matrix polynomial";
option package;
export generate_dense_unstructured_perturbation, generate_unstructured_perturbation,constrained_newton_arbitrary,  norm_opt, poly_norm_opt,clean_up, clean_list, matrix_degree,nearest_unstructured_distance, convolution_matrix, linearize_A_bound,vectorize, linearize_A_fat,linearize_A,linearize_A_minimal, linearize_A_minimal_kernel, generate_initial_point, compute_initial_guess_kernel;

#input: dimension n and indeterminant x of degree d 
#output: a matrix in undetermined coefficient form  with a list of vars 
generate_structured_ansatz := proc(A,x:='xx',symbol:='a')
local M:=Matrix(A);
local vars;
local i,j,k;
vars:=[];
for i from 1 to LinearAlgebra:-RowDimension(A) do 
  for j from 1 to LinearAlgebra:-ColumnDimension(A) do
    M[i,j]:= add(symbol[i,j,k]*x^k,k=0..degree(A[i,j],x));
    vars:= [op(vars),seq(symbol[i,j,k],k=0..degree(A[i,j],x))];
  od:
od:

return M,vars;
end proc:

#input: dimension n and indeterminant x of degree d 
#output: a matrix in undetermined coefficient form  with a list of vars 
generate_ansatz := proc(n,d,x,symbol:='a')
local M:=Matrix(n);
local vars;
local i,j;
vars:=[];
for i from 1 to n do 
  for j from 1 to n do 
       M[i,j] := add(symbol[i,j,k]*x^k,k=0..d); #generate a dense, unstructured undetermined coefficient form
       vars:= [op(vars),seq(symbol[i,j,k],k=0..d)]; #keep a list of vars, needed for eqn solving
  od:
od:

return M,vars;
end proc:


#procedure to compute 2-norm without absolute values, for optimization
norm_opt:=proc(f,x,p:=2)
local v;
local nm;
v:=PolynomialTools:-CoefficientVector(f,x);

nm := add(v[i]^p,i=1..ArrayNumElems(v));

return (nm)^(1/p);
end proc:


#procedure to compute Fronebnius norm without absolute values
poly_norm_opt :=proc(A,x,p:=2)

local nm;
local i;

nm := add(norm_opt(A(i),x,2)^2, i=1..ArrayNumElems(A));
return sqrt(nm);

end proc:





#proc: conv
#pomputes a convolution matrix of f of degree d
#conv := proc (f,d,x:='t')
#  local M, i;
# M := Matrix(d+1, degree(f,x)+d+1);
#  for i to d+1 do
#    M[i] := Vector(CoefficientVector(x^(i-1)*f, x), degree(f,x)+d+1):
#  end do;
#  return LinearAlgebra:-Transpose(M);
# end proc:



#proc: block_conv
#computes an inflated block system for multiplication of polynomial matrices.

#conv := proc(A,d,x:='t')
#  local M, #matrix we return, well transpose thereof
#	i,
#	j,
#	k,
#	;
#end proc:



#proc: generate_factor
#computes a matrix polynomial with indeteminants
#input: symbol = u for indeterminants, t is the variable of matrix polynomial,d is the degree, n, m are dimension parameters.

generate_factor :=proc(u,t,d,n,m,structure:=NULL)
  local 	my_vars:=[], #vars for the computation
		U, # the matrix
		i,
		j;
		
	U:=Matrix(n,m):
	my_vars := []:
	
	
	
	for i from 1 to n do
	  for j from 1 to m do 
	      if(structure <> NULL) then
		  U[i,j] := add(u[i, j, k]*t^k, k = 0 .. degree(structure[i, j], t));
		  my_vars:= [op(my_vars), seq(u[i, j, k], k = 0 .. degree(structure[i,j],t))];
	      else
		  U[i,j] := add(u[i, j, k]*t^k, k = 0 .. d);
		  my_vars:= [op(my_vars), seq(u[i, j, k], k = 0 .. d)];      
	      fi:
	  od:
	od:
	
	return U, my_vars
end proc;

generate_triangular_factor :=proc(u,t,d,n,m,unit:=true)
  local 	my_vars:=[], #vars for the computation
		U, # the matrix
		i,
		j;
		
	U:=Matrix(n,m):
	my_vars := []:
	
	
	
	for i from 1 to n do
	  for j from 1 to m do 
	      if i > j then 
		  U[i,j] := add(u[i, j, k]*t^k, k = 0 .. d);
		  my_vars:= [op(my_vars), seq(u[i, j, k], k = 0 .. d)];
	      elif i= j then
		  U[i,j]:=1;
	      fi:
	  od:
	od:
	
	return U, my_vars
end proc;


#proc: generate_smith
#computes a generic rank deficient smith form.
#TODO: make sure I have correct degree bounds...


# n,m - dimension
#r - rank of matrix
#t -poly var 
#d  = degree of original,  used to compute non-trivial entry, usually degree n*d

generate_smith:=proc(n,m,r,t,d)
    local S, #smith form
	  s_vars, #vars!111
	  i,
	  j;

	  S:=Matrix(n,m);
	  
	  for i from 1 to r do 
	      S[i,i] := 1;
	  od:
	  
	  S[r+1,r+1] := add('s'[r+1, r+1, k]*t^k, k = 0 .. n*d);
	  s_vars := [seq('s'[r+1,r+1,k],k=0..n*d)];
	  
	  
	  return S,s_vars;
	  
end proc:


#todo:
#generate a sparse version of algorithms that uses the structure precisely, and has several trivial variables.
#this greatly restricts the search direction to places we like.

#proc: degree_cons
#given Phi, enforce some degree constraints...
#if objective[i,j, k] > degree(A[i,j])+1, then add a zero constraint.

#by default allow a dense formulation of the problem.
degree_cons := proc(A,objective,t, dense:=true)
      local i,
	    j,
	    cons, #the constraints
#	    cons_vec,
	    cf_vec,
	    d,
	    Md
	    ;
	    cons:=[]:

	    Md:=matrix_degree(A,t);
	    
	    
	  for i from 1 to ArrayNumElems(A) do
	      cf_vec := PolynomialTools:-CoefficientVector(objective(i),t);
	    
	      
	      if dense = true then 
		d:= Md;
	      else
		d:= degree(A(i),t);
	      fi:
	      
	      
	      #this is a fix to the problem of 0 degree entries have degree -infty
	      #for implementation purposes this is much more convenient.
	      if d < 0 then d := -1 fi:
	      
	      cf_vec:=cf_vec[d+2..ArrayNumElems(cf_vec)]; #okay to perturb lc's, but not above that!, the +2 looks weird, but ther are d+1 non-trivial elements, so d+2 and above is trivial.
	      cons:=[op(cons), op(Equate(cf_vec, Vector(ArrayNumElems(cf_vec))))];
	  od:
	    
	    
    return cons;
end proc:



#proc: full_rank_degree_system
#used to take null space constraints and prevent construction of a rank deficient matrix. 
#low order terms should not appear.
full_rank_degree_system := proc( null_system, u_vars,v_vars)
      
      local i,
	    j,
	    u_copy,
	    v_copy,
	    AA_u,
	    BB_v
	    ;
	    
	  AA_u := LinearAlgebra:-GenerateMatrix(null_system,v_vars)[1]; #unknowns are in u
	  BB_v := LinearAlgebra:-GenerateMatrix(null_system,u_vars)[1]; #unknowns are in v
	  
	  
	  u_copy := {};
	  v_copy := {};
	  
	  for i from 1 to ArrayNumElems(AA_u) do
	    u_copy := u_copy union {AA_u(i)};
	  od:
	  
	  for j from 1 to ArrayNumElems(BB_v) do
	    v_copy := v_copy union  {BB_v(j)};
	  od:
	  
	  #ordering of this method might be fucked up a bit.
	  return convert(u_copy minus {0,1},list), convert(v_copy minus {0,1},list)
	  

end proc:



#proc: unimodular_cons
#used to generate constraints that force $U$ to be unimodular_cons

unimodular_cons :=proc(U,t)
    
    local i,
	  j,
	  cons,
	  cons_vec,
	  #I,
	  dimen
	  ;
	  
	  cons_vec:= PolynomialTools:-CoefficientVector(LinearAlgebra:-Determinant(U),t);
	  
	  #this might be bad, make it $nd$  is another choice. normalize determinant to be a constant...
	  cons:=[cons_vec[1]=1];
	  cons:=[op(cons), op(Equate(cons_vec[2..ArrayNumElems(cons_vec)], Vector(ArrayNumElems(cons_vec)-1)))];
	  
	  return cons;

end proc:



#proc linearize_uv
#linearizes A-UV into the form vec(A)-M(U)vec(V) or something like that

linearize_uv := proc(A,U,V,t)
    local 	UV,
		i,
		my_eqns,
		cf_vec,
		avec
		  ;
    UV := U.V;
    my_eqns:=[];
    
    
    for i from 1 to ArrayNumElems(UV) do 
	cf_vec := PolynomialTools:-CoefficientVector(UV(i),t);
	avec:= PolynomialTools:-CoefficientVector(A(i),t);
	avec:= Vector(avec,ArrayNumElems(cf_vec));
	my_eqns:=[op(my_eqns), op(Equate(avec,cf_vec))];
    od:


    return my_eqns;
end proc;


#proc: alternating_gradient_desc
#implement the alternating gradient descent method. 
#partially implemented in a work sheet, polish later.

a_grad := proc(A,U,V,t,u_Vars,v_vars,u_guess:=[],v_guess:=[],tol:=1e-5,max_iter:=100)

    local Phi,
	  u_grad,
	  v_grad,
	  step,
	  u_prev,
	  v_prev,
	  u_next,
	  v_next,
	  u_list,
	  v_list,
	  i,
	  my_step,
	  err_next,
	  err_prev,
	  phi_dist, #compute distance in an easily stored thingy...
	  uv_dist #distance between u and v 
	  ;
    
    
	Phi := poly_norm_opt(A-U.V, t)^2;
	u_grad := Vector(VectorCalculus:-Gradient(Phi, u_vars));
	v_grad := Vector(VectorCalculus:-Gradient(Phi, v_vars));
	
	  
	  
	u_prev := convert(subs(u_guess,u_vars),Vector);
	v_prev := convert(subs(v_guess,v_vars),Vector);
	
	  if nops(u_guess) =0  then 
	      u_prev := subs([seq(u_vars[i] = RandomTools:-MersenneTwister:-GenerateFloat(), i = 1 .. nops(u_vars))],convert(u_vars,Vector)); 
	  fi:
	  if nops(v_guess)=0 then
	      v_prev := subs([seq(v_vars[i] = RandomTools:-MersenneTwister:-GenerateFloat(), i = 1 .. nops(v_vars))],convert(v_vars,Vector));
	  fi:
	  
	  u_next := u_prev;
	  v_next := v_prev;
	
	my_step := .05; #use this fixed value for now? could do something adaptive maybe...fuck it.
	err_next := -1;
	
	    u_list := Equate(convert(u_vars, Vector), convert(u_prev, Vector)); 
	    v_list := Equate(convert(v_vars, Vector), convert(v_prev, Vector));
	

	for i from 1 to max_iter do 
	
	    u_list := Equate(convert(u_vars, Vector), convert(u_prev, Vector)); 
	    v_list := Equate(convert(v_vars, Vector), convert(v_prev, Vector));

	    v_next := v_prev - my_step * subs( [op(v_list),op(u_list)], v_grad);
	    u_next := u_prev - my_step * subs( [op(v_list),op(u_list)], u_grad);
	    
	    
	    if norm(u_prev - u_next) + norm(v_next - v_prev) < tol then 
		  u_list := Equate(convert(u_vars, Vector), convert(u_prev, Vector)); 
		  v_list := Equate(convert(v_vars, Vector), convert(v_prev, Vector));
		  break
	    fi:
	    
	    u_prev := u_next;
	    v_prev := v_next;
	    
	od:

	return [subs([op(u_list),op(v_list)],Phi), [op(u_list),op(v_list)], i, norm(u_prev - u_next) + norm(v_next - v_prev) ]:

end proc:


 


#proc: alternating_LS
#implement alternating LS to solve the problem. Constraints are linear, so it works nicely.
#this implementation is with respect to the 2-norm, however one could adjust this to compute the 1-norm, if it is of interest.
#this shit is production code.

alternating_LS := proc(A,U,V,t,u_vars,v_vars,u_guess:=[],v_guess:=[],tol:=1e-16,max_iter:=1000)

    local AA_u, #linear system A_u (vec V) = b_u
	  bb_u,
	  BB_v, #linear system B_v (vec B) = c_v
	  cc_v,
	  AA, #ansatz version
	  BB,
	  cc,
	  bb,
	  u_prev,
	  v_prev,
	  u_init,
	  v_init,
	  u_next,
	  v_next,
	  A_max,
	  i,
	  uv_lin, #linearize U.V
	  uv_degree_cons, #linearize the degree constraints that deg(UV[i,j]) <= deg(A[i,j])
	  u_con, #uv_degree_cons with u subbed in
	  v_con, #ditto with v
	  Phi, #objective func
	  Xu, #u-coordinate stuff
	  Xv, #v-corrdinate stuff for projective method
	  u_soln,
	  v_soln,
	  u_non_zero,
	  v_non_zero,
	  u_degree_vars,
	  v_degree_vars,
	  test,
	  m,
	  n,
	  non_zero_soln_con
	  ;
	 
	 
	  n,m := LinearAlgebra:-Dimension(A);
	  #WARNING
	  #maximum element... This is not efficient for massive matrices, however I am lazy.
	  #A_max:= max([seq( norm(A(i),2)^2,i=1..ArrayNumElems(A))]);
	  A_max := 1;
	  
	  
	  #assign randm #'s as an init guess?
	  #we scale by A_max so the expected value will be A_max/2. this prevents clustering around 0, which is a bad solution,
	  #but the degree constraints like to enforce it. This makes us avoid the stationary point corresponding to 0.
	  #I can also add another constraint to prevent the 0 soln, but then Maple fucking shits the bed and we've got a huge snafu.
	  
	  u_init := u_guess;
	  v_init := v_guess;
	  
	  print("constructing initial guess...");
	  if nops(u_guess) =0  then 
	      u_init := [seq(u_vars[i] = RandomTools:-MersenneTwister:-GenerateFloat()*A_max, i = 1 .. nops(u_vars))]; 
	  fi:
	  #print(u_init);
	  if nops(v_guess)=0 then
	      v_init := [seq(v_vars[i] = RandomTools:-MersenneTwister:-GenerateFloat()*A_max, i = 1 .. nops(v_vars))];
	  fi:
	  #print(v_init);
	  print("initialization done");
	  
	  #initilization is now done.
	  
	  Phi:= poly_norm_opt(A-U.V,t)^2;
	  print("Phi initialized");
	  
	  #linearize
	  uv_lin := linearize_uv(A, U, V, t);
	  
	  #prevent perturbing degrees higher.
	  uv_degree_cons := degree_cons(A, simplify(U.V), t,false);
	  
	  u_degree_vars, v_degree_vars:= full_rank_degree_system(uv_degree_cons,u_vars,v_vars);
	  
	  
	  #prevent the zero solution...

	#make  Tr(V^T V) >= m to prevent the 0 constraint. If $V$ has rank m then we allow VV^T = diag(blah not zero),
	#and we include orthogonal matrices in the solution if we reduce to lower polynomial degree.
	#print(U.V);
	  non_zero_soln_con := [];#[poly_norm_opt(U,t)^2 <= 10];
	#[ PolynomialTools:-CoefficientVector(A[1, 1], t)[degree(A[1, 1], t)+1] =
	#			 PolynomialTools:-CoefficientVector((simplify(U.V))[1, 1], t)[degree(A[1, 1], t)+1]];
	#				 print(non_zero_soln_con);
	  
	  print("Problem linearized");
	  
	  
	  #we use the degree_vars here because we want the linearized block-Toeplitz system to have full rank,
	  #which prevents degenerate least square solutions.
	  #AA_u = AA(u), bb_u = bb(u), AA_u . vec(V) = bb_u
	  AA, bb := LinearAlgebra:-GenerateMatrix( uv_lin,v_degree_vars);
	 
	  
	  #BB_v = BB(v), cc_v = cc(v), BB_v . vec(U) = cc_v
	  BB, cc := LinearAlgebra:-GenerateMatrix( uv_lin,u_degree_vars);
	  
	  
	  #fo shizzle
	  u_prev := convert(u_init,Vector);
	  v_prev := convert(v_init,Vector);
	  
	  #the double init snafu is because we need to mine big data in the cloud
	  #or... well symmetry and stuf, and I don't want to debug symmetries when I could everage the novel computing cloud paradigms with hadoop.
	  u_next :=u_init;
	  v_next :=v_init;
	  
	  print("iteration commencing");
	  
	  #now we do the iteritive snafu.
	  for i from 1 to max_iter do
	      #remember to meme with integrity at uwaterloo
	      
	      u_prev := u_next;
	      v_prev := v_next;
 
	      AA_u := subs( u_next, AA);
	      bb_u := subs( u_next,bb);
	      u_con :=[]: [op(subs(u_next,uv_degree_cons)), op(subs(u_next,non_zero_soln_con))];
	     
	      
	      v_next := Optimization:-LSSolve( convert(AA_u.convert(v_degree_vars,Vector)-bb_u,list), [op(u_con)])[2];

	      
	      BB_v :=   subs( v_next,BB);
	      cc_v :=   subs( v_next,cc);
	      v_con := [op(subs( v_next, uv_degree_cons)),op(subs(v_next,non_zero_soln_con))];
	      u_next :=  Optimization:-LSSolve(convert(BB_v.convert(u_degree_vars,Vector)-cc_v,list),[op(v_con)])[2];
	      
	      test := norm(subs(u_next, convert(u_vars,Vector))-subs(u_prev, convert(u_vars,Vector))) +  
		      norm(subs(v_next, convert(v_vars,Vector))-subs(v_prev, convert(v_vars,Vector))) ;  
	 # print(norm(subs(u_next, convert(u_vars,Vector))-subs(u_prev, convert(u_vars,Vector))));
	 # print(norm(subs(v_next, convert(v_vars,Vector))-subs(v_prev, convert(v_vars,Vector))));  
	  
	  
	      if test < tol  or abs(subs([op(u_next),op(v_next)],Phi) - subs([op(u_prev),op(v_prev)],Phi))  <tol  then 
		    break;
	      fi: 
	      
	  od:
	  
	
	  
	  #return stuff to the abuser
	  return  [ subs([op(u_next),op(v_next)],Phi) ,[op(u_next), op(v_next)],i,test,abs(subs([op(u_next),op(v_next)],Phi) - subs([op(u_prev),op(v_prev)],Phi))];
end proc:


#proc clean_up
#cleans up small terms

clean_up := proc (A, t, tol:=1e-4)

    local i,
	  j,
	  cf_vec
	  ;

	  for i from 1 to ArrayNumElems(A) do 
	      cf_vec:= PolynomialTools:-CoefficientVector(A(i),t);
	      for j from 1 to ArrayNumElems(cf_vec) do 
		  if abs(cf_vec(j)) < tol then 
		  cf_vec(j) := 0;
		  fi:
	      od:
	      A(i):= PolynomialTools:-FromCoefficientVector(cf_vec,t);
	  od:
	  
	  return A;
	  
end proc:

#proc clean_list: sanitize list inputs if needed.

clean_list :=proc (L, tol:=1e-6)
    local i,
	  M
	  ;
	  M:=[]:
	  for i from 1 to nops(L) do 
	      
	      if abs(rhs(L[i])) < tol then 
		M:=[op(M), lhs(L[i]) = 0];
		  
	      else
		 M:=[op(M), L[i]];
	      fi:
	  od:
    
    return M;
end proc:



#proc: matrix degree - compute degree of matrix polynomial A in t

matrix_degree := proc( A,t)

    local i,
	  d;
	
    d := degree(A(1),t);  
    for i from 2 to ArrayNumElems(A) do 
	d := max(degree(A(i),t),d);
    od:
    return d;
end proc:



#proc: nearest_unstructured
#solves the QM problem  min norm(Ab).
#note that this is an unstructured singular value.
#this is a debugging function to get rank information,
#since min |Ab| is a quadratic minimization problem for fixed a

nearest_unstructured_distance := proc(A,t,tol:=1e-8)

    local i,
	  j,
	  bb,
	  bb_vars,
	  d,
	  n,
	  m,
	  my_obj,
	  my_eqns,
	  cf_vec,
	  Ab,
	  AA,
	  mu,
	  Sigma,
	  rank
	  ;
	  
	  d := matrix_degree (A,t); #get the degree for bounds
	  n,m := LinearAlgebra:-Dimension(A);

	  bb := Vector(m, i-> add(b[i,j]*t^j,j=0..2*max(m,n)*d));
	  bb_vars := [seq(seq(b[i,j],j=0..2*max(m,n)*d),i=1..m)];
	  
	  mu := 2*n*d;

	  
	  #simplify this shiznit dawg
	  Ab := simplify(A.bb); #no seriously, maples degree function will fail if I don't :(
	  
	  my_eqns := []:
	  for i from 1 to ArrayNumElems(Ab) do 
	      cf_vec := PolynomialTools:-CoefficientVector(Ab(i),t);
	      my_eqns  := [op(my_eqns), op(Equate(cf_vec, Vector(ArrayNumElems(cf_vec)))) ];
	  od:
	  
	  #vector should be null vector!
	  AA:=LinearAlgebra:-GenerateMatrix(my_eqns,bb_vars)[1];
	  
	  #this gives us the "inflated" version of A that we want to find the nearest rank deficient matrix to.
	  #if A is rank deficient, this this too will be rank deficient.
	  #I might code this up later as an initial guess routine.
	  
	  #rank computation
	  #this does not account for scaling and is not particularly reliable for huge matrices.
	  #however, it is good for a debugging procedure on small matrices, like this one.
	  Sigma:= LinearAlgebra:-SingularValues(evalf(AA)); #do an SVD
	  rank:=0;
	  for j from 1 to ArrayNumElems(Sigma) do 
	      if abs(Sigma[j]) < tol then break fi: #numeric rank of the inflated matrix.
	  od:
	  
	  rank:= ceil(evalf(j-1)/evalf(mu+d+1));#this one is important
	  
	  
	  return [rank, Sigma[j-1],j];
end proc:




#proc covolution_matrix
#creates a convolution matrix for polynomial multiplication

convolution_matrix := proc(a,d,n,t)
      local mat,
	    i,
	    j,
	    mu,
	    cf_vec
	    ;
  
	  mu:= n*d+1;
	  mat  := Matrix(mu,mu+d);
	  
	  for i from 1 to mu do 
	      mat [i] := PolynomialTools:-CoefficientVector(a*t^(i-1),t);
	  od:
	  
	 return LinearAlgebra:-Transpose(mat);
end proc:



#proc: linearize_A_fat
#linearizes $A$ according to the block-Toeplitz transformation.

linearize_A_fat := proc(A,t)

local 	d,
	n,
	m,
	bb,
	bb_vars,
	Ab,
	my_eqns,
	cf_Vec,
	AA,
	b_lin,
	mu,
	i
	;


	d:= matrix_degree(A,t);
	n,m := LinearAlgebra:-Dimension(A);
	
	#WARNING: assumes square matrix, i.e. m=n
	AA:=`<,>`(seq(`<|>`(seq(convolution_matrix(A[i, j], d, n, t), j = 1 .. n)), i = 1 .. n));
	

	return AA;
	
end proc:



#proc: linearize_A
#linearizes $A$ according to nullspace criteria

#A is singular <=> Ab =0 <=> Ahat. bhat =0 
linearize_A := proc(A,t)
    local d,
	  n,
	  m,
	  bb,
	  bb_vars,
	  Ab,
	  my_eqns,
	  cf_vec,
	  AA,
	  b_lin,
	  mu,
	  i
	  ;
	  
    
    	  d := matrix_degree (A,t); #get the degree for bounds
	  n,m := LinearAlgebra:-Dimension(A);

	  
	  #WARNING: using the max here could over-estimate
	  
	  mu:=1*min(m,n)*d+1;
	  
	  bb := Vector(m, i-> add('b'[i,j]*t^j,j=0..mu-1));
	  bb_vars := [seq(seq('b'[i,j],j=0..mu-1),i=1..m)];

	  
	  Ab := (A.bb); #no seriously, maples degree function will fail if I don't :(
	  
	  my_eqns := []:
	  for i from 1 to ArrayNumElems(Ab) do 
	      cf_vec := PolynomialTools:-CoefficientVector(Ab(i),t);
	      my_eqns  := [op(my_eqns), op(Equate(cf_vec, Vector(ArrayNumElems(cf_vec)))) ];
	  od:
	  
	  #vector should be null vector!
	  AA:=LinearAlgebra:-GenerateMatrix(my_eqns,bb_vars)[1];

	  return AA, convert(bb_vars,Vector);
end proc:



#proc: linearize_A_minimal
#linearizes $A$ according to nullspace criteria

#A is singular <=> Ab =0 <=> Ahat. bhat =0 
linearize_A_minimal := proc(A,t,b_guess,symbol:='b',return_vector := false)
    local d,
	  n,
	  m,
	  bb,
	  bb_vars,
	  Ab,
	  my_eqns,
	  cf_vec,
	  AA,
	  b_lin,
	  mu,
	  i,
	  k
	  ;
    
    	  d := matrix_degree (A,t); #get the degree for bounds
	  n,m := LinearAlgebra:-Dimension(A);

	  mu:=1*min(m,n)*d+1;
	  
	  
	  #TODO: Need to set some variables to 0
	 # bb := Vector(m, i-> add(symbol[i,j]*t^j,j=0..degree(b_guess[i],t)));
	#  bb_vars := [seq(seq(symbol[i,j],j=0..degree(b_guess[i],t)),i=1..m)];
	  
	  bb_vars := []:
	  #print(generate_structured_perturbation(b_guess,t,'c'));
	  bb:=Vector(m);
	  
	  
	   for i from 1 to m do   
		cf_vec := PolynomialTools:-CoefficientVector(b_guess[i],t);
		
		bb_vars := [op(bb_vars), seq(symbol[i,k], k=0..degree(b_guess[i],t))];
		bb[i] := add(symbol[i,k]*t^k,k=0..degree(b_guess[i],t));
		
		
		for k from 1 to ArrayNumElems(cf_vec) do
		    if(cf_vec[k]=0) then
			#bb_vars := convert(convert(bb_vars,set) minus convert(symbol[i,k-1],set),list);
			bb_vars := convert(convert(bb_vars,set) minus {symbol[i,k-1]},list);
			bb[i]:=subs(symbol[i,k-1]=0,bb[i]);
		    fi:
		od:
	      od:
	# bb := Vector(m, i-> add(symbol[i,j]*t^j,j=0..degree(b_guess[i],t)));
	#  bb_vars := [seq(seq(symbol[i,j],j=0..degree(b_guess[i],t)),i=1..m)];
	  
	
	  
	  
	  #simplify this shiznit dawg
	  Ab := (A.bb); #no seriously, maples degree function will fail if I don't :(
	  
	  my_eqns := []:
	  for i from 1 to ArrayNumElems(Ab) do 
	      cf_vec := PolynomialTools:-CoefficientVector(Ab(i),t);
	      my_eqns  := [op(my_eqns), op(Equate(cf_vec, Vector(ArrayNumElems(cf_vec)))) ];
	  od:
	# 
	#  my_eqns:=convert(my_eqns,set) minus {};
	  #remove duplicates
	 # my_eqns:=convert(my_eqns,list);
	 # 
	  
	  #vector should be null vector!
	  AA:=LinearAlgebra:-GenerateMatrix(my_eqns,bb_vars)[1];

	  if return_vector = true then 
	      return  AA, convert(bb_vars,Vector), bb;
	  fi:
	  
	  return AA, convert(bb_vars,Vector);
end proc:

#proc: compute_initial_guess_kernel
compute_initial_guess_kernel := proc (B_init,B,t)
	local i,
	      j,
	      B_vec,
	      B_init_vec
	      ;

	      B_init_vec := convert(B_init,Vector);
	      B_vec := convert(B,Vector);
	      
	      
	      B_init_vec := `<,>`( seq( PolynomialTools:-CoefficientVector(B_init_vec(i),t), i=1..ArrayNumElems(B_init_vec)));
	      B_vec := `<,>`( seq( PolynomialTools:-CoefficientVector(B_vec(i),t), i=1..ArrayNumElems(B_vec)));
	      
	    
	      
	      return convert(convert(Equate(B_vec,B_init_vec),set) minus {0=0},list); #lets hope this works...
	      
	      
	      
end proc:

#Proc: linearize_A_minimal_kernel 
#try to compute a minimally embedded $A$ by linearizing the entire kernel (minimally) at once.
linearize_A_minimal_kernel := proc(A,t,B,symbol:='F')
	local i,
	      j,
	      k,
	      d,
	      n,
	      m,
	      BB,
	      BB_vars,
	      mu
	      ;
	      
	  d := matrix_degree (A,t); #get the degree for bounds
	  
	  B_temp := convert(B,Matrix);
	  
	  n,m := LinearAlgebra:-Dimension(B_temp);


	  BB:=Matrix(LinearAlgebra:-Dimension(B_temp));
	  
	  #fill up the unknown coefficient matrix.
	  for i from 1 to LinearAlgebra:-RowDimension(B_temp) do 
	      for j from 1 to LinearAlgebra:-ColumnDimension(B_temp) do 
		    if B_temp[i,j] <> 0 or B_temp[i,j] <> 0. then  
			#need to preserve the support as well... Not done yet...
			BB[i,j] := add(symbol[i,j,k]*t^k,k=0..degree(B_temp[i,j],t));
			
			cf_vec := PolynomialTools:-CoefficientVector(B_temp[i,j],t);
			for k from 0 to ArrayNumElems(cf_vec)-1 do 
			    if cf_vec[k+1] = 0 or cf_vec[k+1] = 0.0 then 
				BB[i,j] := subs(symbol[i,j,k] =0, BB[i,j]);
			    fi:
			od:
			
		    fi:
	      od:
	  od:
	  
	  BB_vars := convert(convert(`<,>`( seq(PolynomialTools:-CoefficientVector(BB(i),t),i=1..ArrayNumElems(BB))),set) minus {0,0.},list);
	  
	  #BB_vars := convert(convert(BB_vars,set) minus {0},list); #POSSIBLE BUG HERE?
	  
	  
	  AB:= A.BB; #generate residual vector. 
	  my_eqns := [];
	  for i from 1 to LinearAlgebra:-RowDimension(AB) do
	      for j from 1 to LinearAlgebra:-ColumnDimension(AB) do 
		  cf_vec := PolynomialTools:-CoefficientVector(AB[i,j],t);
		  my_eqns := [op(my_eqns), op(Equate(cf_vec, Vector(ArrayNumElems(cf_vec))))];
	      od:
	  od:
	  
	  my_eqns := convert(convert(my_eqns,set) minus {0=0},list);
	  
	  
	  A_lin := LinearAlgebra:-GenerateMatrix(my_eqns, BB_vars)[1];
	  
	  
	  return A_lin, convert(BB_vars,Vector), BB;
	  
	
end proc:




#Proc: linearize_A_minimal_kernel_alt 
#try to compute a minimally embedded $A$ by linearizing the entire kernel (minimally) at once.
#this method differs from the previous in that it creates the system of equations by stacking each vector individually,
#instead of linearizing the whole kernel. The system of equations generated by this method have fewer singularities in the Jacobian.
#This is the only reason to use this. 
linearize_A_minimal_kernel_alt := proc(A,t,B,symbol:='F')
	local i,
	      j,
	      k,
	      d,
	      n,
	      m,
	      BB,
	      A_lin,
	      BB_vars
	      ;
	      
	  d := matrix_degree (A,t); #get the degree for bounds
	  
	  B_temp := convert(B,Matrix);
	  
	  m := LinearAlgebra:-RowDimension(B);
	  n := LinearAlgebra:-ColumnDimension(B);
	  
	  BB_vars := [];
	  my_eqns := []; #keep track of the equations and shit. 
	  BB:= Matrix(0,0);
	  
	  residuals :=[];
	  
	  
	  
	  #we will just minimally embed each column vector, then stack the equations together...
	  for i from 1 to n do 
	      A_lin,b_lin,b_vec := linearize_A_minimal(A,t,Matrix(B)[1..m,i],symbol[i],true);
	      
	      #get the minimal residuals...
	      residuals := [op(residuals),op(convert(convert(A_lin.b_lin,set) minus {0},list))]; 
	     #residuals := [op(residuals),op(convert(A_lin.b_lin,list))]; 
	     
	      BB_vars := [op(BB_vars), op( convert(b_lin,list))];
	      
	      
	      if ArrayNumElems(BB) = 0 then 
		  BB:=Matrix(b_vec);
	      else
		  BB:= <BB|b_vec>; 
	      fi:
	  od:
	  
	  A_lin := LinearAlgebra:-GenerateMatrix(residuals,BB_vars)[1];
	  
	  

	  
	  return A_lin, convert(BB_vars,Vector), BB;
	  	
end proc:


#proc: rank_factor_A_lin
#assume $U$ is symmetric
#compute an ansatz for a rank factorization of A
rank_factor_A_lin := proc(A_lin,r)
  local 	 U_lin,
		 V_lin,
		 U_lin_vars,
		 V_lin_vars,
		 n,
		 m,
		 i,
		 j
		 ;


	n,m := LinearAlgebra:-Dimension(A_lin);
	U_lin := Matrix(n,r);
	V_lin := Matrix(r,m);
	
	for i from 1 to n do
	    for j from 1 to r do 
		U_lin[i,j] := 'U'[i,j];
	    od:
	od:
	
#	for i from 1 to n do 
#	    for j from 1 to r do 
#		U_lin[i,j] := 'U'[i,j];
#		#U_lin[j,i] := -1*'U'[i,j];
#	    od:
#	od:
	
	for i from 1 to r do 
	    for j from 1 to m do 
		 V_lin[i,j] := 'V'[i,j];
	    od:
	od:
	
	#print(U_lin);
	U_lin_vars := [seq( seq('U'[i,j],j=1..r),i=1..n)];
	#U_lin_vars:=[seq(,j=)];
	#[seq( seq('U'[i,j], 'U'[j,i],j=1..r),i=1..r)];
	V_lin_vars := [seq( seq('V'[i,j],j=1..m),i=1..r)];

	
	return U_lin,V_lin, U_lin_vars,V_lin_vars;
end proc:

#proc: enforce_polynomial_constraints
#forces U_lin . V_lin to have polynomial structure. 
enforce_polynomial_constraints := proc( A_lin, U_lin, V_lin,t)
    local 	i,
		j,
		constraints,
		UV
		;

	constraints := [];
	
	#ouch!
	UV := simplify(U_lin.V_lin);
	#sad panda is sad about the complexity here :(
	
	for i from 1 to ArrayNumElems(A_lin) do 
	    if A_lin(i) =0 then 
		constraints:=[op(constraints), UV(i)=0]:
	    fi:
	od:
	
	return constraints;
end proc:

#proc random_pt
#plugs in a random point to a list
#monte carlo alg for detecting rank of a matrix.

random_pt := proc(L)
    local i, M := []:
    
    for i from 1 to nops(L) do 
	M:= [op(M), L[i] = RandomTools:-MersenneTwister:-GenerateFloat()];
    od:
    return M;
end proc:


#proc: generate_structured_perturbation
#generates strutured ansatz for the problem...




generate_structured_perturbation := proc(A,t,symbol:='a')
	local i,
	      M,
	      j,
	      m,
	      n,
	      k,
	      a_vars,
	      cf_vec;
	      ;
	      
	      
	      n,m := LinearAlgebra:-Dimension(A);
	      M:= Matrix(n,m);
	      a_vars := []:
	      
	      for i from 1 to n do 
		  for j from 1 to m do 
			cf_vec := PolynomialTools:-CoefficientVector(A[i,j],t);
			
			a_vars := [op(a_vars), seq(symbol[i,j,k], k=0..degree(A[i,j],t))];
			M[i,j] := add(symbol[i,j,k]*t^k,k=0..degree(A[i,j],t));
			
			
			for k from 1 to ArrayNumElems(cf_vec) do
			    if(cf_vec[k]=0) then
				a_vars := convert(convert(a_vars,set) minus convert(symbol[i,j,k-1],set),list);
				M[i,j]:=subs(symbol[i,j,k-1]=0,M[i,j]);
			    fi:
			od:
			
		  od:
	      od:
	      
	      
	      return M, a_vars;
end proc:

#vectorizes stuff
vectorize := proc(b,t)
    local i,
	  j,
	  vec,
	  ell
	  ;
	  
	  ell := []:
	  for i from 1 to ArrayNumElems(b) do 
	      vec := PolynomialTools:-CoefficientVector(b(i),t);
	      for j from 1 to ArrayNumElems(vec) do 
		  if(vec[j] <>0) then
		      ell := [op(ell),vec[j]];
		  fi:
	      od:
	  od:
	  
	  return convert(ell,Vector);

end proc:


#proc constrained_newton
#implements a constrained version of Newtons method.

#input:
#A- matrix A that we want to optimize over 
#DeltaA - perturbation structure
#A_init - rank deficient or near rank deficient matrix poly
#b_init -  null or approximate null vector for A_init 

#perturbation structure specified by the user.
constrained_newton := proc (A,DeltaA,DeltaA_vars, A_init,b_init,num_iter := 1000,tol:=1e-9,divergence:=1e2)
	local A_vars, #variables
	      b_vars, #variables
	      Lambda,  #Lagrange multiplier variables
	      Lambda_init, #we solve for this via Jacobian/residual equations.
	      A_lin, #minimal embedded A
	      A_init_lin,
	      b_lin, #minimal embedded b
	      DA_lin, #perturbation structure. #pull variables from here!
	      X, #<A_vars,b_vars>
	      X_init,
	      Z, #<X,Lambda>
	      Z_init,
	      ZZ_next,
	      DeltaZ, #change in coordinates for Newton
	      ZZ, #used to perform the iteration, stores values of Z in a vectorized form. 
	      residuals, # vector of residuals
	      L, #Lagrangian
	      Phi, #objective function.
	      H, #FULL Hessian of Lagrangian
	      G, #FULL Gradient of Lagrangian
	      Gx,
	      Gy,
	      lagrangian_system,
	      cf_vec,
	      cf_mat,
	      ell,
	      default_DeltaA_pt,
	      i,
	      active_vars,
	      J,
	      ZZ_prev,
	      my_eqns,
	      ls_soln,
	      unactive_vars,
	      trivial_eqns
	      ;
	      
	      #WARNING: Assuming b_init has no "small" elements in it.
	      #cleanup code to be added after initial prototyping process.
	      
	     A_lin := linearize_A_minimal(A,t,b_init)[1];
	     A_init_lin := linearize_A_minimal(A_init,t,b_init)[1];
	     DA_lin,b_lin := linearize_A_minimal(DeltaA,t,b_init);
	     
	     
	    
	     b_vars:= convert(b_lin,list);
	  #  print(b_vars);
	    
	    #initial guess gets equationed.
	    
	    default_DeltaA_pt := [seq(DeltaA_vars[i]=0,i=1..nops(DeltaA_vars))];


		
		
	#    print(A_init_lin);
	#    print(A_lin);
	#    print(A);
	#    print(A_init);
	    
		
	     X_init :=[ op(Equate(DA_lin, A_init_lin-A_lin)), op(Equate(b_lin, vectorize(b_init,t)))];
	     #prin(X_init);
	     
	    
	     
	     #somewhat convoluted; need to remove duplicates and 0=0 equations to get full-rank Jacobian.
	     residuals := convert( (A_lin + DA_lin).b_lin,list);
	     residuals := convert( [op(residuals), poly_norm_opt(b_lin,t)^2-1],set) minus {0}; #remove the zero equation is important
	     residuals := convert(residuals,list); #go back to the list form.
	     
	     
	     active_vars:= convert(convert([seq( lhs(X_init[i]),i=1..nops(X_init))],set) minus {0,0.0},list);
	     
	     unactive_vars := convert(convert(DeltaA_vars,set) minus convert(active_vars,set),list);
	     trivial_eqns := [seq(unactive_vars[i]=0,i=1..nops(unactive_vars))];
	  # print(active_vars);
	   
	   
	 #   print(X_init);
	   
	     Lambda := [seq('lambda'[i],i=1..nops(residuals))];
	     X:=active_vars ; #[op(DeltaA_vars), op(b_vars)];
	    # print(X);
	     Z:= [op(X),op(Lambda)];
	   

	    #WARNING:
	    #Note the 1/2 in the objective function!!!! - I since removed it.
	     Phi:= poly_norm_opt(DeltaA,t)^2 +(poly_norm_opt(b_lin,t)^2-1);
	     #note the minus sign here is important!
	     L:=Phi - LinearAlgebra:-DotProduct(convert(Lambda,Vector), convert(residuals,Vector), conjugate=false);
	     
	     #print(Z);
	#print(active_vars);
	
	     G:= VectorCalculus:-Gradient(L,Z);
	     H:=VectorCalculus:-Hessian(L,Z);
	     Gx := (VectorCalculus:-Gradient(L,X));
	     Gy := (VectorCalculus:-Gradient(L,Lambda));
	     
	     
	     
	     #TODO: write code to make sure zero-rows in the Jacobian don't appear.
	     #this doesn't affect convergence but looks bad for testing.
	      J:= VectorCalculus:-Jacobian(residuals,X);
	     if( LinearAlgebra:-Rank(subs(X_init,J)) <> LinearAlgebra:-RowDimension(J)) then
		WARNING("RANK DEFICIENT JACOBIAN DETECTED. PREPARE FOR CRITICAL FAIL");
		#WARNING(subs(X_init,J));
		#print(LinearAlgebra:-SingularValues(subs(X_init),J)[LinearAlgebra:-RowDimension(J)]);
		WARNING("Possibly not a minimally embedded nullspace vector");
	     fi:

	     #print(LinearAlgebra:-ConditionNumber( subs(X_init, J.Transpose(J)),2));
	     
	     
	     #solving a linear system for initial guess
	     lagrangian_system := subs(default_DeltaA_pt, convert(subs(X_init,Gx),Vector));
	     my_eqns := Equate(lagrangian_system, Vector(ArrayNumElems(lagrangian_system)));
	     cf_mat,cf_vec := LinearAlgebra:-GenerateMatrix(my_eqns,Lambda);
	     #print(cf_mat);
	     #print(cf_vec);
	     #print(X_init);
	     ls_soln := Optimization:-LSSolve([cf_vec,cf_mat])[2];
	     
	     Lambda_init := Equate(convert(Lambda,Vector), ls_soln);
	     #we now have initial points and matrices computed. Now we do heavy lifting w/ equation solving.
	     Z_init := [op(X_init), op(Lambda_init)];
	     

	     
	   #  if( (Rank(subs(Z_init,H))) <> LinearAlgebra:-RowDimension(H)) then
		#WARNING("Possibly rank deficient Hessian. Newton may not converge.");
		#print(LinearAlgebra:-ConditionNumber(subs(Z_init,H),2));
			     	     
		
		#print(LinearAlgebra:-SingularValues(subs(X_init,J))[1]);
		#print(LinearAlgebra:-SingularValues(subs(X_init,J))[LinearAlgebra:-RowDimension(J)-1..LinearAlgebra:-RowDimension(J)]);
		#print(LinearAlgebra:-SingularValues(subs(Z_init,H))[LinearAlgebra:-RowDimension(H)]);
		#print(LinearAlgebra:-SingularValues(subs(Z_init,H))[1]);
	      
		#WARNING("Please ignore this is warning if the above condition number is  \" small \"");
	   # fi:
	     
	     #a bit of a hack for now, but some perturbation coefficients just won't appear.
	     ZZ:= subs(default_DeltaA_pt, convert(subs(Z_init,Z),Vector));
	     ZZ_next := Equate(Z,ZZ);
	     for i from 1 to num_iter do
		#solve it via iterative least squares for stability purposes
		DeltaZ:= Optimization:-LSSolve([subs(ZZ_next,-1*G)  , subs(ZZ_next, H)])[2];
		ZZ_prev := ZZ;
		ZZ:= ZZ+DeltaZ; #update
		ZZ_next := Equate(Z,ZZ);

		print(norm(ZZ-ZZ_prev,2));
		if norm(ZZ-ZZ_prev,2) < tol then break fi:
		
		#divergence works for normalized examples...
		if norm(ZZ-ZZ_prev,2) > divergence then WARNING("Newton Diverged"); break; fi:

	     od:
	     
	     
	     
	     return [subs( default_DeltaA_pt, subs(ZZ_next,Phi)),[ op(ZZ_next), op(trivial_eqns)],i,norm(ZZ-ZZ_prev,2)];

end proc:



#used to generate an initial point in a non-hack way for the unstructured newton method. 
generate_initial_point := proc(A,A_init,DA,t)
	local i,
	      j,
	      n,
	      cf_vec,
	      perturb_vec; 
	      eqns
	      ;
	      
	      eqns:=[]:
	      for i from 1 to ArrayNumElems(A) do 
		    cf_vec := PolynomialTools:-CoefficientVector(-A(i)+A_init(i),t);
		    perturb_vec := PolynomialTools:-CoefficientVector(DA(i),t);
		     n := max(ArrayNumElems(cf_vec),ArrayNumElems(perturb_vec));
		     
		    cf_vec := Vector(cf_vec, n);
		    perturb_vec:=Vector(perturb_vec,n);
		    eqns:= [op(eqns), op(Equate(perturb_vec,cf_vec))];
	      od:
	 return eqns;     
end proc:

#proc constrained_newton
#implements a constrained version of Newtons method.

#input:
#A- matrix A that we want to optimize over 
#DeltaA - perturbation structure
#A_init - rank deficient or near rank deficient matrix poly
#b_init -  null or approximate null vector for A_init 

#this function is for arbitrary linear structures... it's not for the paper. just for me.
#perturbation structure specified by the user.
constrained_newton_arbitrary := proc (A,DeltaA,DeltaA_vars, A_init,b_init,num_iter := 1000,tol:=1e-9,divergence:=1e2)
	local A_vars, #variables
	      b_vars, #variables
	      Lambda,  #Lagrange multiplier variables
	      Lambda_init, #we solve for this via Jacobian/residual equations.
	      A_lin, #minimal embedded A
	      A_init_lin,
	      b_lin, #minimal embedded b
	      DA_lin, #perturbation structure. #pull variables from here!
	      X, #<A_vars,b_vars>
	      X_init,
	      Z, #<X,Lambda>
	      Z_init,
	      ZZ_next,
	      DeltaZ, #change in coordinates for Newton
	      ZZ, #used to perform the iteration, stores values of Z in a vectorized form. 
	      residuals, # vector of residuals
	      L, #Lagrangian
	      Phi, #objective function.
	      H, #FULL Hessian of Lagrangian
	      G, #FULL Gradient of Lagrangian
	      Gx,
	      Gy,
	      lagrangian_system,
	      cf_vec,
	      cf_mat,
	      ell,
	      default_DeltaA_pt,
	      i,
	      active_vars,
	      J,
	      ZZ_prev,
	      my_eqns,
	      ls_soln,
	      unactive_vars,
	      trivial_eqns
	      ;
	      
	      #WARNING: Assuming b_init has no "small" elements in it.
	      #cleanup code to be added after initial prototyping process.
	      
	     A_lin := linearize_A_minimal(A+DeltaA,t,b_init)[1];
	     A_init_lin := linearize_A_minimal(A_init+DeltaA,t,b_init)[1];
	     DA_lin,b_lin,b_poly := linearize_A_minimal(DeltaA,t,b_init,'Bb',true);
	     
	     
	    
	     b_vars:= convert(b_lin,list);
	  #  print(b_vars);
	    
	    #initial guess gets equationed.
	    
	    default_DeltaA_pt := [seq(DeltaA_vars[i]=0,i=1..nops(DeltaA_vars))];
	    
	    A_lin := subs(default_DeltaA_pt,A_lin); #remove Delta terms
	    A_init_lin := subs(default_DeltaA_pt,A_init_lin); #remove Delta terms
		
		
	#    print(A_init_lin);
	#    print(A_lin);
	#    print(A);
	#    print(A_init);
	    
		
	     X_init :=[ op(Equate(DA_lin, A_init_lin-A_lin)), op(Equate(b_lin, vectorize(b_init,t)))];
	     #prin(X_init);
	     
	    
	     
	     #somewhat convoluted; need to remove duplicates and 0=0 equations to get full-rank Jacobian.
	     residuals := convert( (A_lin + DA_lin).b_lin,list);
	     residuals := convert( [op(residuals), poly_norm_opt(b_lin,t)^2-1],set) minus {0}; #remove the zero equation is important
	     residuals := convert(residuals,list); #go back to the list form.
	     
	     
	     active_vars:= convert(convert([seq( lhs(X_init[i]),i=1..nops(X_init))],set) minus {0,0.0},list);
	     
	     unactive_vars := convert(convert(DeltaA_vars,set) minus convert(active_vars,set),list);
	     trivial_eqns := [seq(unactive_vars[i]=0,i=1..nops(unactive_vars))];
	  # print(active_vars);
	   

	   
	 #   print(X_init);
	   
	     Lambda := [seq('lambda'[i],i=1..nops(residuals))];
	     X:=active_vars ; #[op(DeltaA_vars), op(b_vars)];
	    # print(X);
	     Z:= [op(X),op(Lambda)];
	   

	    #WARNING:
	    #Note the 1/2 in the objective function!!!! - I since removed it.
	     Phi:= poly_norm_opt(DeltaA,t)^2 +(poly_norm_opt(b_lin,t)^2-1);
	     #note the minus sign here is important!
	     L:=Phi - LinearAlgebra:-DotProduct(convert(Lambda,Vector), convert(residuals,Vector), conjugate=false);
	     
	     #print(Z);
	#print(active_vars);
	
	     G:= VectorCalculus:-Gradient(L,Z);
	     H:=VectorCalculus:-Hessian(L,Z);
	     Gx := (VectorCalculus:-Gradient(L,X));
	     Gy := (VectorCalculus:-Gradient(L,Lambda));
	     
	     
	     
	     #TODO: write code to make sure zero-rows in the Jacobian don't appear.
	     #this doesn't affect convergence but looks bad for testing.
	      J:= VectorCalculus:-Jacobian(residuals,X);
	     if( LinearAlgebra:-Rank(subs(X_init,J)) <> LinearAlgebra:-RowDimension(J)) then
		WARNING("RANK DEFICIENT JACOBIAN DETECTED. PREPARE FOR CRITICAL FAIL");
		#WARNING(subs(X_init,J));
		#print(LinearAlgebra:-SingularValues(subs(X_init),J)[LinearAlgebra:-RowDimension(J)]);
		WARNING("Possibly not a minimally embedded nullspace vector");
	     fi:

	     #print(LinearAlgebra:-ConditionNumber( subs(X_init, J.Transpose(J)),2));
	     
	     
	     #solving a linear system for initial guess
	     lagrangian_system := subs(default_DeltaA_pt, convert(subs(X_init,Gx),Vector));
	     my_eqns := Equate(lagrangian_system, Vector(ArrayNumElems(lagrangian_system)));
	     cf_mat,cf_vec := LinearAlgebra:-GenerateMatrix(my_eqns,Lambda);
	     #print(cf_mat);
	     #print(cf_vec);
	     #print(X_init);
	     ls_soln := Optimization:-LSSolve([cf_vec,cf_mat])[2];
	     
	     Lambda_init := Equate(convert(Lambda,Vector), ls_soln);
	     #we now have initial points and matrices computed. Now we do heavy lifting w/ equation solving.
	     Z_init := [op(X_init), op(Lambda_init)];
	     

	     
	   #  if( (Rank(subs(Z_init,H))) <> LinearAlgebra:-RowDimension(H)) then
		#WARNING("Possibly rank deficient Hessian. Newton may not converge.");
		#print(LinearAlgebra:-ConditionNumber(subs(Z_init,H),2));
			     	     
		
		#print(LinearAlgebra:-SingularValues(subs(X_init,J))[1]);
		#print(LinearAlgebra:-SingularValues(subs(X_init,J))[LinearAlgebra:-RowDimension(J)-1..LinearAlgebra:-RowDimension(J)]);
		#print(LinearAlgebra:-SingularValues(subs(Z_init,H))[LinearAlgebra:-RowDimension(H)]);
		#print(LinearAlgebra:-SingularValues(subs(Z_init,H))[1]);
	      
		#WARNING("Please ignore this is warning if the above condition number is  \" small \"");
	   # fi:
	   
	   print("----------------------------------------------");
	     
	     #a bit of a hack for now, but some perturbation coefficients just won't appear.
	     ZZ:= subs(default_DeltaA_pt, convert(subs(Z_init,Z),Vector));
	     ZZ_next := Equate(Z,ZZ);
	     for i from 1 to num_iter do
		#solve it via iterative least squares for stability purposes
		DeltaZ:= Optimization:-LSSolve([subs(ZZ_next,-1*G)  , subs(ZZ_next, H)])[2];
		ZZ_prev := ZZ;
		ZZ:= ZZ+DeltaZ; #update
		ZZ_next := Equate(Z,ZZ);

		print(norm(ZZ-ZZ_prev,2));
		if norm(ZZ-ZZ_prev,2) < tol then break fi:
		
		#divergence works for normalized examples...
		if norm(ZZ-ZZ_prev,2) > divergence then WARNING("Newton Diverged"); break; fi:

	     od:
	 
	     
	     return [subs( default_DeltaA_pt, subs(ZZ_next,Phi)),subs(default_DeltaA_pt, subs(ZZ_next, DeltaA)),subs(ZZ_next,b_poly),[ op(ZZ_next), op(trivial_eqns)],i,norm(ZZ-ZZ_prev,2)];

end proc:




#proc: remove_minimal_jacobian_singularity
#input: list of constraints with initial guess
#output:list of possibly fewer constraints with corresponding list of possibly fewer variables.
#this function will remove (in an exact setting), singularities in the Jacobian matrix when the kernel is minimally embedded
#caused by columns having common block entries. 
#this can be augmented for a numerical version that solves a nearby problem exactly that is better conditioned, as possible future work.


#this function is inteded to be used on the kernel variables
remove_minimal_jacobian_singularity:= proc(residuals, vars, vars_init)
      #currently I am thinking of an efficient implementation....
end proc:




#proc constrained_newton_arbitrary_kernel
#implements a constrained version of Newtons method for arbitrary linear structures and rank.
#this one tries to deal with a rank-at-most approximation, opposed to rank deficient.

#input:
#A- matrix A that we want to optimize over 
#DeltaA - perturbation structure
#A_init - rank deficient or near rank deficient matrix poly
#b_init -  null or approximate null vector for A_init 

#Notes: There is either a bug or "feature" occuring in this function where the matrix A_lin has more than one kernel Vector.
#	The matrix looks something like 
#A_lin = [ A_1 ]
#	---------
#	 [ A_2 ]
#with a kernel resembling 
# B_lin = c_1 [ B_1 ] + c_2 [ 0 ] 
#	      [ 0   ]       [ B_2 ],
#that is the problem is that it has two kernel vectors. I need to come up with a way to "fix" this...

#this seems like approximate GC(R)D, in that if I fix some "principal" coefficients, like leading coefficients in each Vector, I avoid this problem. 
#unfortunately, this can lead to undesirable numerical behavior, such as infima occuring.





#this function is for arbitrary linear structures... it's not for the paper. just for me.
#perturbation structure specified by the user.
constrained_newton_arbitrary_kernel := proc (A,DeltaA,DeltaA_vars, A_init,B_init,num_iter := 10,tol:=1e-9,divergence:=1e2)
	local B_lin, #list of kernel vector with minimally embedded matrix
	      B_lin_init, # list of variables for an initial guess...
	      B_lin_vars, #variables for kernel vector. Not A_vars 
	      B_var,
	      B_row_dim, #row dimension 
	      default_DeltaA_pt,
	      DA_init,
	      X, #concatenation of a bunch of vars
	      X_init, #initial guess for...
	      Z, #concatenation of lagrangian lambda values 
	      Z_init, #initial guess as well...
	      Lambda,
	      Phi,
	      L
	      ;
	      


	      
	      B_var := 'BB'; #use this symbol for now... 
	      
	     #symbol for kernel variables will be 'BB' for now...
	   
	    default_DeltaA_pt := [seq(DeltaA_vars[i]=0,i=1..nops(DeltaA_vars))];
	   
	    A_lin, B_lin,B_ker := linearize_A_minimal_kernel_alt(A+DeltaA,t,B_init,'F'); #the 'F' was chosen for "FUCK FUCK FUCK FUCK FUCK... FUCK?"
	    A_lin := subs(default_DeltaA_pt,A_lin); #this is a hack to get the right number of equations...
	
	    
	    B_lin_vars := convert(B_lin,list);
	    
	    

	    DA_init := generate_initial_point(A,A_init,DeltaA,t);
	    
	    
	    DA_init := convert(convert(DA_init,set) minus {0=0},list);
	    
	
	    
	    #The bug occurs in this code segment right here;
	    #there are trivial equations here
	    DA_lin := linearize_A_minimal_kernel_alt(DeltaA,t,B_init)[1];
	    
	    B_temp:= convert(B_init,Vector); 
	    
	    B_lin_init := Equate(B_lin,vectorize(B_temp,t));
	    
	    #print(B_lin_init);
	    
	    #some coefficients do not appear in the minimization and are trivially zero. 
	    
	    
	    
	    #initial guesses and default variables are now sat up.
	    
	    
	    residuals := convert( (DA_lin-A_lin).B_lin,list);
	    
	    print( nops(convert(convert((A_lin.B_lin),set) minus {0,0.0}, list)));
	    
	    #multi lined the code by popular request/demand.
	    residuals := convert(convert(residuals,set) minus {0,0. },list); #get rid of 0=0 type equalities, and redundant equations
	    residuals := convert(residuals,list);
	    
	
	  
	    scale_constraints := [seq( poly_norm_opt(B_ker[1..LinearAlgebra:-RowDimension(B_ker),j],t)^2-1,j=1..LinearAlgebra:-ColumnDimension(B_ker))];
	   

	   
	   
	   residuals := [op(residuals),op(scale_constraints) ];
	    #the objective, albeit slightly tweaked. 
	   Phi := poly_norm_opt(DeltaA,t)^2 + poly_norm_opt(B_lin,t)^2-LinearAlgebra:-ColumnDimension(B_init);
	   
	    
	   #the set up is mostly done now...
	   
	   Lambda := [seq('lambda'[i],i=1..nops(residuals))];
	   
	   
	   printf("There are %d Lagrangian variables\n",nops(Lambda));
	   printf("There are  %d Kernel variables and  %d Delta Variables\n",nops(B_lin_vars), nops(DeltaA_vars));
	  L:=Phi - LinearAlgebra:-DotProduct(convert(Lambda,Vector), convert(residuals,Vector), conjugate=false);
	   
	   X:= [ op(DeltaA_vars), op(B_lin_vars)];
	   X_init := [op(DA_init), op(B_lin_init)]; #WARNING: inactive variables may be present...
	   
	   printf("B_lin_init has %d vars and B_lin_vars is %d\n",nops(B_lin_init), nops(B_lin_vars));
	   printf("DA_init has %d vars and DeltaA is %d\n",nops(DA_init), nops(DeltaA_vars));
	   
	   
	   Z:= [op(X), op(Lambda)];
	
	
	#create differential quantities...

	    #up to this point everything looks good....
	
	    G:= VectorCalculus:-Gradient(L,Z);
	    H:=VectorCalculus:-Hessian(L,Z);
	    Gx := (VectorCalculus:-Gradient(L,X));
	    Gy := (VectorCalculus:-Gradient(L,Lambda));
	    J:= VectorCalculus:-Jacobian(residuals,X);
	    
	    
	    #maybe make this a seperate routine?
	    
	    #solving a linear system for initial guess
	    lagrangian_system := subs(default_DeltaA_pt, convert(subs(X_init,Gx),Vector));
	    my_eqns := Equate(lagrangian_system, Vector(ArrayNumElems(lagrangian_system)));
	    cf_mat,cf_vec := LinearAlgebra:-GenerateMatrix(my_eqns,Lambda);
	    #print(cf_mat);
	    #print(cf_vec);
	    #print(X_init);
	    ls_soln := Optimization:-LSSolve([cf_vec,cf_mat])[2];
	    Lambda_init := Equate(convert(Lambda,Vector), ls_soln);
	    
	    Z_init := [op(X_init), op(Lambda_init)]; #al the initial guesses should be done...
	    
	    print(subs(Z_init,L));

	    
	    printf("The initial value of the Lagrangian is %e. This should be \"close \" to zero.\n",subs(Z_init,L));
	    
	    printf("The initial value of the gradient Lagrangian is %e. This should be \"close \" to zero.\n",norm(subs(Z_init,G),2));
	    
	      
	    printf("X_init has %d vars and Lambda_init has %d vars\n",nops(X_init) ,nops(Lambda_init));  
	    printf("Total number of variables is %d and total initial point information is %d"\n, nops(Z),nops(Z_init) );
	    
	  print(  nops(convert(convert(X_init,set),list)));
	    
	      
	    print("------------------BEGINNING ITERATION--------------------");
	    
	    #problem looks like it's in the Hessian... 
	
	    ZZ:= subs(default_DeltaA_pt, convert(subs(Z_init,Z),Vector));
	    ZZ_next := Equate(Z,ZZ);

	      

	     for i from 1 to num_iter do
	     
		#solve it via iterative least squares for stability purposes
		DeltaZ:= Optimization:-LSSolve([subs(ZZ_next,-1*G)  , subs(ZZ_next, H)])[2];
		ZZ_prev := ZZ;
		ZZ:= ZZ+DeltaZ; #update
		ZZ_next := Equate(Z,ZZ);

		print(norm(ZZ-ZZ_prev,2));
		if norm(ZZ-ZZ_prev,2) < tol then break fi:
		
		#divergence works for normalized examples...
		if norm(ZZ-ZZ_prev,2) > divergence then WARNING("Newton Diverged"); break; fi:

	     od:

	    printf("The final Lagrangian guess is accurate to %e \n", norm(subs(ZZ_next,G),2)); #verify stationary point property.
	    
	    printf("The initial Jacobian rank is %d and final Jacobian rank is %d\n",LinearAlgebra:-Rank(subs(Z_init,J)),LinearAlgebra:-Rank(subs(ZZ_next,J)));
	    
	    printf("The dimension of the Hessian is %d and has condition number %e and rank %d\n",LinearAlgebra:-RowDimension(H),LinearAlgebra:-ConditionNumber(subs(ZZ_next,H),2),LinearAlgebra:-Rank(subs(Z_init,H)));
	    printf("The dimension of the Jacobian is %d by %d and has smallest singular Value %f and largest singular value  %f\n",LinearAlgebra:-Dimension(J), LinearAlgebra:-SingularValues(subs(Z_init,J))[LinearAlgebra:-RowDimension(J)], LinearAlgebra:-SingularValues(subs(Z_init,J))[1]);
	
	
	   
	   printf("The final residual is %e and the initial residual was %e\n", poly_norm_opt(subs(ZZ_next,DeltaA),t),poly_norm_opt(subs(Z_init,DeltaA),t));
	     
	     return [subs( default_DeltaA_pt, subs(ZZ_next,Phi)), subs(ZZ_next,DeltaA), subs(ZZ_next,B_ker), [ op(ZZ_next), op(trivial_eqns)],i,norm(ZZ-ZZ_prev,2)];

end proc:


constrained_newton_arbitrary_kernel_dbg2:= proc(A,DA,DA_vars,A_init,K_init,num_iter := 10,tol:=1e-9,divergence:=1e2)

    default_DeltaA_pt:= [seq(DA_vars[i]=0,i=1..nops(DA_vars))]:
    DA_lin,B_lin,B_ker_lin:= linearize_A_minimal_kernel(A+DA,t,K_init,'FF'):
    #DA_lin := linearize_A_minimal_kernel_alt(DA,t,K_init)[1]:

   # A_lin := subs(default_DeltaA_pt,A_lin);

    

    residuals := convert(convert( simplify((DA_lin) .B_lin),set) minus {0},list):
    vars := [ op(DA_vars),op(convert(B_lin,list))]: #same for both of them...


    scale_constraints := [seq( poly_norm_opt(B_ker_lin[1..LinearAlgebra:-RowDimension(B_ker_lin),j],t)^2-1,j=1..LinearAlgebra:-ColumnDimension(B_ker_lin))];
   # scale_constraints:= [poly_norm_opt(B_ker_lin,t)^2-1];
    

    residuals := [op(residuals),op(scale_constraints) ];
   Phi := poly_norm_opt(DA,t)^2 + poly_norm_opt(B_ker_lin,t)^2-LinearAlgebra:-ColumnDimension(K_init);

  #  Phi := poly_norm_opt(DA,t)^2 + poly_norm_opt(B_ker_lin,t)^2-1;

    Lambda := [seq('lambda'[i],i=1..nops(residuals))];
    L:=Phi - LinearAlgebra:-DotProduct(convert(Lambda,Vector), convert(residuals,Vector), conjugate=false);
    

  B_lin_init := compute_initial_guess_kernel(K_init, B_ker_lin,t):

    DA_init := generate_initial_point(A,A_init,DA,t):
    init_pt := [op(B_lin_init),  op(DA_init)];


    
    Z:= [op(vars),op(Lambda)];


    
    G:= VectorCalculus:-Gradient(L,Z);
    H:=VectorCalculus:-Hessian(L,Z);
    Gx := (VectorCalculus:-Gradient(L,vars));
    Gy := (VectorCalculus:-Gradient(L,Lambda));
  
    J:= VectorCalculus:-Jacobian(residuals,vars):

    #solving a linear system for initial guess
	lagrangian_system := subs(default_DeltaA_pt, convert(subs(init_pt,Gx),Vector));
	my_eqns := Equate(lagrangian_system, Vector(ArrayNumElems(lagrangian_system)));
	cf_mat,cf_vec := LinearAlgebra:-GenerateMatrix(my_eqns,Lambda);
	ls_soln := Optimization:-LSSolve([cf_vec,cf_mat])[2];
	Lambda_init := Equate(convert(Lambda,Vector), ls_soln);
	
	Z_init := [op(init_pt), op(Lambda_init)]; #al the initial guesses should be done...
	

    
    print( "init gradient", norm(subs(Z_init,G),2));

    print( "init gradient X", norm(subs(Z_init,Gx),2));
    
    
    print( "init gradient Y", norm(subs(Z_init,Gy),2));
    
    
     ZZ:= subs(default_DeltaA_pt, convert(subs(Z_init,Z),Vector));
	    ZZ_next := Equate(Z,ZZ);

	    print(subs(ZZ_next,Lambda));
	      

	     for i from 1 to num_iter do
	     
		#solve it via iterative least squares for stability purposes
		DeltaZ:= Optimization:-LSSolve([subs(ZZ_next,-1*G)  , subs(ZZ_next, H)])[2];
		
		#print("ls residual is ", Optimization:-LSSolve([subs(ZZ_next,-1*G)  , subs(ZZ_next, H)])[1]);
		
		ZZ_prev := ZZ;
		ZZ:= ZZ+DeltaZ; #update
		ZZ_next := Equate(Z,ZZ);

		
		printf("||X-X_next|| = %e, ||Phi - Phi_next|| = %e\n", norm(ZZ-ZZ_prev,2),abs(poly_norm_opt( subs(ZZ_next ,DA),t) - poly_norm_opt(subs(Equate(Z,ZZ_prev),DA),t)));
		if norm(ZZ-ZZ_prev,2) < tol then break fi:
		
		#divergence works for normalized examples...
		if norm(ZZ-ZZ_prev,2) > divergence then WARNING("Newton Diverged"); break; fi:

	     od:
	      print( "final gradient", norm(subs(ZZ_next,G),2));
	      
	      
	      print("initial residuals", norm(subs(ZZ_next,convert(residuals,Vector)),2));
    

	      print("final residuals", norm(subs(ZZ_next,convert(residuals,Vector)),2));
        print("Jacobian-rank,dimension", LinearAlgebra:-Rank(subs(ZZ_next, J)),LinearAlgebra:-Dimension(J));
        
        printf("Jacobian has initial smallest singular value %e and final smallest singular value %e\n", LinearAlgebra:-SingularValues(subs(Z_init,J))[LinearAlgebra:-RowDimension(J)],LinearAlgebra:-SingularValues(subs(ZZ_next,J))[LinearAlgebra:-RowDimension(J)] );
	
	#print(LinearAlgebra:-SingularValues(subs(Z_init,J)));
	#print(LinearAlgebra:-SingularValues(subs(ZZ_next,J)));
	
	printf("The Hessian has dimension %d, rank %d and Condition Number %e\n", LinearAlgebra:-RowDimension(H),LinearAlgebra:-Rank(subs(ZZ_next, H)),LinearAlgebra:-ConditionNumber(subs(ZZ_next,H)));
    
	    
	    # print("final....", subs(ZZ_next,poly_norm_opt((DA_lin+A_lin).subs(ZZ_next,B_lin)),t)));
	    
	     return [subs( default_DeltaA_pt, subs(ZZ_next,Phi)), subs(ZZ_next,DA), subs(ZZ_next,B_ker_lin), [ op(ZZ_next)],i,norm(ZZ-ZZ_prev,2)];


end proc:




#generates a singular matrix polynomial of dimension n,m with rank r 
# and degree  at most d in t.
#used for testing code.

generate_singular_matrix:= proc( n,m, r,d,t,structure:=dense,lower:=-10,upper:=10)
    local i,
	  j,
	  Ml,
	  Mr
	  ;


	  #rank factorization always gives a rank deficient matrix!
	  Ml := Matrix(n,r);
	  Mr :=  Matrix(r,m);
	  
      for i from 1 to n do 
	  for j from 1 to  r do 
	      Ml [i,j] := randpoly(t,degree= floor(d/2.0),coeffs=rand(lower..upper),structure);
	  od:
      od:

      for i from 1 to r do 
	  for j from 1 to m do 
	      Mr [i,j] := randpoly(t,degree= ceil(d/2.0), coeffs=rand(lower..upper),structure);
	  od:
      od:
      return simplify(Ml.Mr);
end proc:


#generates a singular matrix polynomial of dimension n,m with rank r 
# and degree  at most d in t.
#used for testing code.

generate_singular_matrix_rank_factored:= proc( n,m, r,d,t,structure:=dense,lower:=-10,upper:=10)
    local i,
	  j,
	  Ml,
	  Mr
	  ;


	  #rank factorization always gives a rank deficient matrix!
	  Ml := Matrix(n,r);
	  Mr :=  Matrix(r,m);
	  
      for i from 1 to n do 
	  for j from 1 to  r do 
	      Ml [i,j] := randpoly(t,degree= floor(d/2.0),coeffs=rand(lower..upper),structure);
	  od:
      od:

      for i from 1 to r do 
	  for j from 1 to m do 
	      Mr [i,j] := randpoly(t,degree= ceil(d/2.0), coeffs=rand(lower..upper),structure);
	  od:
      od:
      return simplify(Ml.Mr),Ml,Mr;
end proc:


#proc clear denoms
#clears denominators of vectors to get polynomials
#warning: this is for exact input, as I don't have the time to debug approximate GCD stuff.
#don't expect it to work super well right now for non-test problems.

clear_denoms := proc(b_input,t)

      local i,
	    bb
	    ;
	   # print(b_input);
  
	    
	    bb := simplify((b_input)); #this should clear rational functions!!!
	    for i from 1 to ArrayNumElems(bb) do 
		  bb := simplify(bb * denom(bb(i)));#iteratively get rid of denominators.
	    od:
	

      return bb;

end proc:


#proc: generate_unstructured_perturbation
#generates an unstructured perturbation for the problem....
#note that this is semi structured in that the perturbations are of the same degree, not same degree structure.
#allowing larger degree perturbations can lead to weird situtations and severely ill-conditioned (but not singular) matrices.

generate_unstructured_perturbation := proc(A,t,symbol:='a')

      local i,
	    M,
	    j,
	    a_vars
	    ;
	    
	    M:= Matrix(LinearAlgebra:-Dimension(A));
	    a_vars := []:
	    
	    for i from 1 to LinearAlgebra:-RowDimension(A) do
		for j from 1 to LinearAlgebra:-ColumnDimension(A) do 
		    M[i,j] := add(symbol[i,j,k]*t^k, k=0..degree(A[i,j],t));
		    a_vars := [op(a_vars), seq(symbol[i,j,k], k=0..degree(A[i,j],t))];
		od:
	    od:

	    return M,a_vars;
end proc:


#used to generate a dense perturation matrix that has all entries as an ANSATZ of degree d
#this of course destroys the support...
generate_dense_unstructured_perturbation:= proc(A,t, symbol:='a')
      local i,
	    M,
	    j,
	    d,
	    a_vars
	    ;
	    
	    M:= Matrix(LinearAlgebra:-Dimension(A));
	    a_vars := []:
	    d:= matrix_degree(A,t);
	    
	    for i from 1 to LinearAlgebra:-RowDimension(A) do
		for j from 1 to LinearAlgebra:-ColumnDimension(A) do 
		    M[i,j] := add(symbol[i,j,k]*t^k, k=0..d);
		    a_vars := [op(a_vars), seq(symbol[i,j,k], k=0..d)];
		od:
	    od:

	    return M,a_vars;
end proc:

#used to generate random perturbations for examples and testing. 
#generated in the interval [-|magnitude|,|magnitue|]
generate_perturbation := proc(A,magnitude:=1)
      local DA,
	    DA_vars,
	    random_pt
	    ;
	    
	    
	    DA,DA_vars:=generate_structured_perturbation(A,t);
	    random_pt := [seq(DA_vars[i]= (-1)^(RandomTools:-MersenneTwister:-GenerateInteger())*RandomTools:-MersenneTwister:-GenerateFloat()*magnitude , i=1..nops(DA_vars))];
	   
	    return subs(random_pt, DA);
	    
end proc:



#paper_examples
#generate some examples and output key statistics in latex format.

#takes no args since I'm hard coding this shit in. 
paper_examples := proc()
	local 	n,#dimension
		d, #degree
		r, #rank deficiency
		e, #exponent
		my_file,
		num_iter,
		DA,
		DA_vars,
		AA,
		C,
		bb,
		fd,
		z,#WARNING: Do not make $t$ local... :(
		newton_error,
		Al,Ar,QQ,Ql,Qr,Alp,Arp,A_init,b_init,rel_error, converged
		;
		
		
		my_file := "paper_examples_out.txt"; #output file name
		
		# constrained_newton(B, DA, a_vars, A, b2, 100, 0.1e-34)

		#do some for loops here and print the progress
		#write stuff out to a file or something.
		r:=1; #Rank deficient by 1
		
		fd:=fclose(my_file); #close it...
		fd:= fopen(my_file,WRITE);
		Digits:=35: #quad precision?
		for n from  3 to 12 by 3 do 
		    for d from 2 to 10 by 2 do
			for e from -6 to -1 by 1 do
			    
			    #my format string, it should give me latex
			    #("%d & %d & %d & %d & 1e-%f & %f\n", n,d,r,num_iter,e,newton_error)
			    #writeline(blah);
			    
		AA, Al, Ar := generate_singular_matrix_rank_factored(n, n, n-r, d, t);
		QQ, Ql, Qr := generate_singular_matrix_rank_factored(n, n, n-r, d, t); 
		#AA := evalf(AA);
		DA, DA_vars := generate_unstructured_perturbation(AA, t);
		C := evalf(generate_perturbation(AA));
		C := C/poly_norm_opt(C, t); 
		C := evalf(C*10^e);
		Al := Al/poly_norm_opt(Al, t);
		Ar := Ar/poly_norm_opt(Ar, t); 
		Alp := Ql;
		Alp := Alp*10^e/poly_norm_opt(Alp, t);
		Arp := Qr;
		Arp := Arp*10^e/poly_norm_opt(Arp, t);
		A_init := (Al+Alp).(Ar+Arp);
		
		#BUG:!!!!
		b_init := clear_denoms(op(NullSpace(A_init)), t); #apperantly there be a bug here?
		#	
		
		b_init := make_primitive(b_init, t);
		A_init := evalf(A_init/poly_norm_opt(A_init, t));
		AA := evalf(AA/poly_norm_opt(AA, t));
		b_init := collect(evalf(b_init/poly_norm_opt(b_init, t)), t);
		
			    
			
			    
			    z := constrained_newton(AA+C, DA, DA_vars, A_init, b_init, 10, 0.1e-34);
			    newton_error := evalf(z[1]/poly_norm_opt(AA+C, t)^2);
			    rel_error := poly_norm_opt(AA+C-A_init, t)^2/poly_norm_opt(A_init, t)^2;
			    num_iter := z[3];
			    converged := z[4];
			    fprintf(fd,"%d & %d & %d & %d & 1e-%f & %f\n", n,d,r,num_iter,e,newton_error);
			   
			od:
		    od:
		od:
		Digits := 15:
		fclose(fd);
end proc:


#proc: make_primitive
#makes a vector primitive.

#this can be done faster, but I don't care for testing.
make_primitive := proc(b,t)
	local c,
	      i,
	      cont
	      ;
	     
	      c:=Vector(ArrayNumElems(b));
	      cont := b(1):
	      for i from 2 to ArrayNumElems(b) do 
		  cont := gcd( cont,b(i));
	      od:

	      for i from 1 to ArrayNumElems(b) do 
		  c[i] := quo(b(i),cont,t);
		  if rem(b(i),cont,t) <> 0 then 
		      WARNING("FAILED TO COMPUTE GCD");
		  fi:
	      od:
	      
	      return c;
end proc:


#check_support
#function to make sure two matrices have the same degree structure (ignores trailing coefficients)
#used to generate examples, weird shit happens some times and you just need to be prepared.

#checks if B has the same structure as A

check_support := proc(A,B,t)
    local i;
    
    for i from 1 to ArrayNumElems(A) do 
	if degree(A(i),t) <> degree(B(i),t) then 
	  return false; 
	fi:
    od:
  return true;
end proc:

 
#A is singular <=> Ab =0 <=> Ahat. bhat =0 
linearize_A_bound := proc(A,t,d_bound)
    local d,
	  n,
	  m,
	  bb,
	  bb_vars,
	  Ab,
	  my_eqns,
	  cf_vec,
	  AA,
	  b_lin,
	  mu,
	  i
	  ;
	  
    
    	  d := matrix_degree (A,t); #get the degree for bounds
	  n,m := LinearAlgebra:-Dimension(A);

	  
	  #WARNING: using the max here could over-estimate
	  
	  #mu:=1*min(m,n)*d+1;
	  mu := d_bound;
	  
	  bb := Vector(m, i-> add('b'[i,j]*t^j,j=0..mu-1));
	  bb_vars := [seq(seq('b'[i,j],j=0..mu-1),i=1..m)];

	  
	  Ab := (A.bb); #no seriously, maples degree function will fail if I don't :(
	  
	  my_eqns := []:
	  for i from 1 to ArrayNumElems(Ab) do 
	      cf_vec := PolynomialTools:-CoefficientVector(Ab(i),t);
	      my_eqns  := [op(my_eqns), op(Equate(cf_vec, Vector(ArrayNumElems(cf_vec)))) ];
	  od:
	  
	  #vector should be null vector!
	  AA:=LinearAlgebra:-GenerateMatrix(my_eqns,bb_vars)[1];

	  return AA, convert(bb_vars,Vector);
end proc:

end module;