#Author: Joseph Haraldson
#some code to perform numeric polynomial linear algebra. 
#depends on the nearest_singular_poly code base for some basic linearization functions.

#preliminary verison

#outline of tasks:
#1.) Compute an approximate minimally embedded kernel vector from a matrix polynomial that is (almost rank) deficient.
#2.) Compute a minimally embedded approximate kernel basis
#3.) Generalize code to other types of linear(affinely) structured matrices


#utitlity function to clean up noise and shit.
clean_up_noise:=proc(A,tol:=1e-4)
local i;
    for i from 1 to ArrayNumElems(A) do 
	if abs(A(i)) < tol then 
	    A(i):=0;
	fi:
    od:
    return A:
end proc:



#numeric rank code
#use a reliable method to comput ethe rank of a matrix numericlaly instead of thresholding like ealrlier...

#Input: Vecor of singular values
#output: the rank of the matrix
numeric_rank_relative := proc(S, m, tol:=1e-4)
    local i,
	  r
	  ;
 
	  r:=m; #default value
	  for i from 1 to m-1 do 
	      if S[i+1]/S[i]  < tol * evalf(sqrt(m)) then 
		  r:= i+1;
	      fi:
	  od:
	  return r;
end proc:

#Input: Vecor of singular values
#output: the rank of the matrix
#by default we use 35 digits of precision, and the matrix is exactly rank deficient, so the smallest singular vlaues should
#be "around" 1e-35. default tolerance of 1e-30 is a safety factor. 
numeric_rank_absolute := proc(S, m, tol:=1e-30)
    local i,
	  r
	  ;
 
	  r:=m; #default value
	  for i from 1 to m-1 do 
	     if S[i] < tol then
		r := i;
		break;
	     fi:
	  od:
	  return r;
end proc:



#used to compute an orthonormal kernel of the matrix $A_lin$
#tol is used to check which singular values are small or something.
compute_orthonormal_kernel := proc(A_lin,tol:=1e-4)
    
    local  UU, #from svd
	  SS, #from svd
	  VVt,  #from svd
	  n, #row dim 
	  m, #col dim
	  i,
	  singular_column_start_index,
	  singular_column_end_index
	  ;
    

    
    UU,SS,VVt :=LinearAlgebra:-SingularValues(evalf(A_lin),output = ['U','S','Vt']); #do the SVD

    n,m := LinearAlgebra:-Dimension(A_lin);
    singular_column_start_index := 1;
    singular_column_end_index := m; 
    
    #naive numeric rank code

  #  for i from 1 to ArrayNumElems(SS) do 
 #     if abs(SS[i]) < tol then singular_column_start_index := i;
#	break;
    #  fi:
   # od:
  # print(singular_column_start_index);
   
  # singular_column_start_index := numeric_rank_relative(SS,m,tol);
    singular_column_start_index := numeric_rank_absolute(SS,m,tol);

  # print(singular_column_start_index);
      #interface(rtablesize=50);
     # print([singular_column_start_index]);
      
     
      
      #this is an approximate kernel.
    return LinearAlgebra:-Transpose(VVt[singular_column_start_index..singular_column_end_index]);
end proc:



#procedure to compute a minimally embedded kernel vector
#column basis K of the kernel of a (scalar)
#WARNING: what we really want is a QL factorization and not an RQ factorization
#THIS PERFORMS AN RQ FACTORIZATION
#THIS DOES NOT GIVE A MINIMALLY DEGREE EMBEDDED VECTOR
#THE RESULTING VECTOR WILL NOT BE PRIMITIVE IF IT CORRESPONDS TO A MATRIX POLYNOMIAL
#THIS WARNING IS FUCKING IMPORTANT.
extract_minimally_embedded_kernel_vector := proc(K)
    local Q, 
	  R,
	  n
	  ;

	  #this shit really doesn't matter right now.
	 # WARNING("LQ unavailable; using QR instead. Minimally embedded vector may not be primitive");

	 Q,R := LinearAlgebra:-QRDecomposition(LinearAlgebra:-Transpose(K));
	#  Q,R := QLDecomposition(LinearAlgebra:-Transpose(K));
	  
	  return LinearAlgebra:-Transpose(R[LinearAlgebra:-RowDimension(R)]);
	 # print(R);
	 # return LinearAlgebra:-Transpose(R[1]);
end proc:


#devectorize
#procedure to de-linearize a vector into a vector of polynomials...
#this assumes every entry has degree $nd$ and is padded with zeros.
#if this assumption fails we are pretty fucked with the answers that we will get.  
devectorize_custom := proc(b,n,degree_bound,t,tol:=0)
local 	bb,
	bb_vars,
	i,
	j
	;
	


	bb := Vector(n, i-> add('c'[i,j] *t^j, j=0..degree_bound-1));


	bb_vec := NearestPoly:-vectorize(bb,t);
	my_eqns := Equate(bb_vec, clean_up_noise(b,tol));
	
	return (subs(my_eqns, bb)); #devectorization complete
end proc:





#devectorize
#procedure to de-linearize a vector into a vector of polynomials...
#this assumes every entry has degree $nd$ and is padded with zeros.
#if this assumption fails we are pretty fucked with the answers that we will get.  
devectorize := proc(b,n,d,t,tol:=1e-6)
local 	bb,
	bb_vars,
	i,
	j
	;
	


	bb := Vector(n, i-> add('c'[i,j] *t^j, j=0..n*d));


	bb_vec := NearestPoly:-vectorize(bb,t);
	my_eqns := Equate(bb_vec, clean_up_noise(b,tol));
	
	return (subs(my_eqns, bb)); #devectorization complete
end proc:


#here we put some code to minimally embed an entire kernel...
#idea behind minimal kernel embedding is 


#BROKEN DONT CALL
#proc: QLDecomposition
#Given a matrix $M$, it does QL decomp on it or something
#let's assume the input is real for now.
QLDecomposition_v1:= proc(A)
    local U,
	  R,
	  m,
	  n,
	  i,
	  j,
	  x
	  ;
	  
	  m,n := LinearAlgebra:-Dimension(A);
	  Q:= Matrix(LinearAlgebra:-IdentityMatrix(m)); #bit of a hack
	
	  print(Q);
	  R:= evalf(A);
	  
	  for k from 1 to min(m,n) do 
		x := R[1..k,k];
		e:=Vector(ArrayNumElems(x));
		e(ArrayNumElems(e)) := 1;
		u:= sign(x(1))*norm(x,2)*e+x;
		u:=(1.0/norm(u,2))*u;
		
		
		R[1..k,k..n] := R[1..k,k..n] - 2*(u.LinearAlgebra:-Transpose(u)).R[1..k,k..n];
		Q[1..k]:= Q[1..k]-2*u.LinearAlgebra:-Transpose(u).Q[1..k]; 
	  od:
	  #do the zeroing out of super diagonal entries
	  for i from 1 to min(m,n) do 
	      for j from 1 to min(m,n) do 
		   if j>i then R[i,j] := 0; fi:
	      od:
	  od:
	  #WARNING: I am not sure that this computes $R$ correctly... yet.
	  return evalf(R);
end proc:




#Proc: structured_gauss_jordan_R
#Procedure to do a structured elimination on the triangular kernel.
#it essentially minimally embeds every vector. 

structured_gauss_jordan_R := proc(R)
 
	return LinearAlgebra:-Transpose(LinearAlgebra:-ReducedRowEchelonForm(LinearAlgebra:-Transpose(R)));
end proc:


#ProC: devectorize_matrix_kernel
#takes the a linearized matrix kernel and converts it to a  matrix polynomial and converts it into a matrix polynomial
#go team redundancy team?

#Input: K is a linearized kernel and d is the maximum degree (note these terms could possibly be zero, resulting in zero-rows in the kernel)
#	X is the variable of the polynomial
devectorize_matrix_kernel := proc(K,n,d,x:='t')
      local i,
	    j,
	    M,#matrix poly dimension 
	    N,#more dimensions
	    K_poly
	    ;
	    
	    
	    M,N := LinearAlgebra:-Dimension(K);
	    
	    K_poly := Matrix(n,N); #same # columns, the number of rows is a bit queer though. 
	    
	    for i from 1 to N do 
		K_poly[1..n,i] := devectorize(K[1..M,i],n,d,x,0); #the zero argument is not to annihilate small terms. 
	    od:
	    
	    return K_poly;
	    
end proc:


#Proc: minimal_spanning_set 
#Input: matrix polynomial, not necessarily sqaure
#Output: a linearly independent spanning set of the vectors. 
#Notes: Runtime is currently sub-optimal. 

minimal_spanning_set := proc(A,t)
      local i,
	    j,
	    A_lin,
	    A_min,
	    A_lin_min,
	    A_temp,
	    m,
	    n,
	    r
	    ;
	    m,n := LinearAlgebra:-Dimension(A);
	    
	    A_min := A[1..m,n]; #for our implementation the last vector is the "nicest" since it's  already minimally embedded.  
	    A_lin_min := A_lin[1..m,n];
	  
	    for i from n-1 to 1 by -1 do 

	    
		A_temp :=  <A[1..m,i] | A_min>;
		
	
		A_lin := linearize_A(A_temp,t)[1];  #there might be some problems with fat instead of tall matrices here... fuck it. that's another day.
		

		
		r:= LinearAlgebra:-Rank(A_lin); #we are using Maples built in rank here. This could not always be correct for very ill-conditioned matrices
						#a WARNING is warranted here if bugs occur...
						
						#of course we can compute the rank with the SVD and using our favourite relative/absolute tolerance scheme...
						
		if r = LinearAlgebra:-ColumnDimension(A_lin) then A_min := <A[1..m,i]|A_min> fi; #update the current kernel vectors. 
	    od:
	    
	    
	    return A_min;
  
end proc:


#helper proc for something  I forget why its here...
stack_vector := proc(b,t,d)
      local i;

	    return `<,>`(seq(Vector(PolynomialTools:-CoefficientVector(b(i),t),d+1),i=1..ArrayNumElems(b)));
end proc:



#Proc: compute_matrix_poly_kernel
#it does what it says more or less. 
#the kernel compute is somewhat special, since it is "column reduced" but not in CRREF
#It's column reduced in the sense that (up to normalization), the kernel is represented with the minimum 
#number of non-zero entries. 
compute_matrix_poly_kernel := proc(A,t,tol:=1e-8)
      local 	A_lin,
		b_lin,
		K,
		K_rref,
		n,
		d,
		K_poly
		;
	    A_lin,b_lin := NearestPoly:-linearize_A(A,t);
	    
	    d:= matrix_degree(A,t);
	    n := min(LinearAlgebra:-Dimension(A));
	    
	    
	    #these double transposes are needed;
	    #we are performing column operations on the kernel 
	    #but maple's linear algebra routines only support row operations... 
	    
	    K:=LinearAlgebra:-Transpose(compute_orthonormal_kernel(A_lin,tol));
	    K_rref := LinearAlgebra:-Transpose(LinearAlgebra:-ReducedRowEchelonForm((K)));
	    
	    
        
	    
            K_poly := devectorize_matrix_kernel( clean_up_noise(K_rref,tol),n,d);

	    
	    return  minimal_spanning_set(K_poly,t);
	    
end proc:



#scale columns to be unit vectors...
rescale_columns := proc(A,t)
    local m,
	  n,
	  i,
	  A_temp
	  ;
	  A_temp := A;
	  m,n := LinearAlgebra:-Dimension(A);
	  
	  
	  for i from 1 to n do 
	      A_temp[1..m,i] := A_temp[1..m,i]/poly_norm_opt(A_temp[1..m,i],t);
	  od:
	  return A_temp;
end proc:



#Proc: row_rank_profile
#Input: Matrix M over $RR$
#Output: row-rank profile of M
#Notes: This is a floating point implementation and relies on the SVD for stability purposes. The runtime is currently $O(n^5),
#	but this can be reduced to O(n^4) with a simple divide and conquer trick. 
#	If I wanted to be really fast ( O(n^3) or faster for structured matrices) I could implement a rank revealing QR decomposition to get the
#	information, but maple doesn't have this. For example, if the input matrix is sparse or structured then a structured rank revealing QR could solve this
#	in time O(s^2) where s is the sparsity or number of basis elements in the (presumably linear) structure.  


#the SVD rank profile method is straight forward; Take rows, add one  more row.  If the smallest singular value is small, discard the last row. repeat until done.

#the divide and conquer method involves doing a binary search on indices. It's akin to a union-find, and achieves an order of magnitude improvement in run time by
#adding blocks of rows instead of a single row. 

#A QR based algorithm would perform a (rank revealing) QR factorization and keep track of indicies where pivoting was needed. Of course fuck coding this for
#prototyping. I'll code this for production code if it comes to it.  

#An uber fast rank revealing  QR would exploit the Jacobian structure and solve the problem in 
#quadratic time (for this particular problem)


#todo: implement this with SVD for beta testing.
row_rank_profile:= proc(M,tol:=1e-8)
      local i,
	    r,
	    m,
	    n,
	    N,
	    Sigma, #for SVD
	    indices;

	    m := LinearAlgebra:-RowDimension(M);
	    n := LinearAlgebra:-ColumnDimension(M); #do I even need this?
	    
	    
	    N := M[1]; #assume non-zero first row for now.
	    
	    indices := [];
	
	    Sigma := LinearAlgebra:-SingularValues(N);
	    r:=1;
	    if(Sigma[1] < tol) then r := 0; indicies := [1] fi: #zero row detected
	    
	    for i from 2 to m do 

		if LinearAlgebra:-SingularValues(<N,M[i]>)[LinearAlgebra:-RowDimension(N)] >=tol then 
			indicies := [op(indicies),i];
			N := <N,M[i]>; #maybe optimize this part later...
			r:=r+1;
		fi:
	    od:
	    
	    return [N, indices];
  #TODO: go implement a naive rank profile method here.
  #TODO: later go implement the QR rank profile method (this is essentially a rank revealing QR factorization method)

  end proc:

#The method works, but has a runtime of $O(n^5)$
row_rank_profile_slow:= proc(M,tol:=1e-8)
      local i,
	    r,
	    m,
	    n,
	    N,
	    Sigma, #for SVD
	    indices;

	    m := LinearAlgebra:-RowDimension(M);
	    n := LinearAlgebra:-ColumnDimension(M); #do I even need this?
	    
	    
	    N := M[1]; #assume non-zero first row for now.
	    
	    indices := [];
	
	    Sigma := LinearAlgebra:-SingularValues(N);
	    r:=1;
	    if(Sigma[1] < tol) then r := 0; indices := [1] fi: #zero row detected
	    
	    
	    for i from 2 to m do 
		Sigma:=LinearAlgebra:-SingularValues(<N,M[i]>); 
		
		#print(Sigma[LinearAlgebra:-RowDimension(N)+1]);
		  
		if Sigma[LinearAlgebra:-RowDimension(N)+1] <= tol then 
			indices := [op(indices),i];
			
			#print("detected singularity");
			
		else 
			N := <N,M[i]>; #maybe optimize this part later...
			r:=r+1; #for debugging purposes... 
		fi:
	
	    od:
	    
	    return N, indices;
  #TODO: go implement a naive rank profile method here.
  #TODO: later go implement the QR rank profile method (this is essentially a rank revealing QR factorization method)

  end proc:
