#Author: Joseph Haraldson
#Project: numeric_smith_prototype
#Purpose: Evaluate Approximate SNF implementation with derivative approximations.


#Skeleton:
#4th and 2nd order numeric diff schemes

#read("nearest_smith_functions.mpl"); #get some helper functions
#Note that we are not including optimization since we are hard-coding LM for the implicit problem, as I can't figure out function pointers in Maple.

#change pathing depending on the current directory...

#read("lib/nearest_smith_functions.mpl"): #get some helper functions



#Proc: nearest_snf
#Input: Matrix polynomial A in indeterminant t with perturbation structure DeltaA (and initial guess)
#Output: Nearby matrix polynomial with non-trivial SNF or indication of failure
#Notes: DeltaA <= A or else we'll have infima and other problems. 
nearest_snf := proc(A,t,DeltaA,DeltaA_init)
      local i,
	    j,
	    Lambda,
	    Lambda_init,
	    adj,
	    adj_vec,
	    fstar,
	    h,
	    X,
	    Z,
	    X_init,
	    Z_init
	    ;
	    
	    ;
end proc:




#Proc: symbolic_numeric_lm
#Input: arguments in some form...
#Output: nearest snf obviously.
#Notes: This is like the fully implicit but with special functions.
#	It is not general purpose, and as such not ubiquitous. 
#	the variable handing is less compact snice we don't want to do parsing here and to get fast
#	function calls we need to pass explicit arguments. I.e. trade portability for speed :(
symbolic_numeric_lm := proc(my_args,tol:=1e-16,max_iter := 100000,lm_param:= 1,lm_amp_factor:=1e-1,lm_damp_factor:=1e1)
      local i,
	    grad,
	    J,
	    JTJ,
	    rhs_k, #first argument to linear solve
	    lhs_k,#second rgument to linear solve
	    current_pt,
	    x_next,
	    Imat,
	    residuals,
	    obj,
	    t_start,t_end,t_iter,
	    j,
	    dx,
	    x_next_candidate,
	    x_cur,
	    A,t,DeltaA,DeltaA_vars,fstar,h,Lambda,X,Z_init,Z,
	    Z_next_candidate,
	    mu_k,#the current levenberg-marquardt parameter
	    safety_factor #for LM
	    ;
	    t_start := time():
	    
	    A,t,DeltaA,DeltaA_vars,fstar,h,Lambda,X,Z_init,Z := op(my_args);
	    
	    safety_factor:= 10^(-1*Digits);
	    #default h for now
	    #J:= numeric_hessian22(my_func,my_args,x,x_init);
	   
	    J:= fast_symbolic_hessian_full(A,t,DeltaA,DeltaA_vars,fstar,h,Lambda,X,Z_init);
	    JTJ := LinearAlgebra:-Transpose(J).J;

	    printf("Computed Initial Numerical Hessian...\n");
	    #residuals := numeric_gradient2(my_func,my_args,x,x_init);
	    residuals := fast_symbolic_gradient(A,t,DeltaA,DeltaA_vars,fstar,h,Lambda,X,Z_init);
	    obj := norm(residuals,2);
	    
	    
	    printf("Computed numerical gradient...\n");
	    
	    #hack to overcome some maple problems in previous versions.
	    #probably obsolete now.
	    Imat:=Matrix(LinearAlgebra:-Dimension(JTJ));
	    for j from 1 to LinearAlgebra:-RowDimension(Imat) do 
		Imat[j,j]:=1:
	    od:
	    
	  
	    
	    x_cur := convert(subs(Z_init,Z),Vector);
	    mu_k := lm_param;
	    
	    printf("Beginning iterations. Setuptime is %f \n", time()-t_start);
	  #  t_iter := time():
	    for i from 1 to max_iter do
		rhs_k := JTJ + mu_k * Imat;
		lhs_k := -1*LinearAlgebra:-Transpose(J).residuals;
		dx := LinearAlgebra:-LinearSolve(rhs_k,lhs_k);
		
		#printf("Computed dx\n");

		x_next_candidate := x_cur + dx;
		
		#residuals_next := numeric_gradient2(my_func,my_args,x,x_next_candidate);
		Z_next_candidate := Equate(Z,x_next_candidate): 
		residuals_next :=  fast_symbolic_gradient(A,t,DeltaA,DeltaA_vars,fstar,h,Lambda,X,Z_next_candidate);
		obj_next := norm(residuals_next,2);
		
		#return obj_next, obj;
	
		if obj_next <= obj  then #valid step 
		#printf("Valid step. Proceeding to lower mu.\n");
		    mu_k :=max( mu_k*(lm_amp_factor),safety_factor); #take a smaller step next iteration
		   
		   
		    #this is based on a criteria from a paper by Wright.
		    #|| residuals||_2 approximates the solution, so the penalty term
		    #blows up at the same rate as the solution, giving asymptotic quadratic convergence.
		    if  norm(dx,2) <=  obj and obj <= 1.5*norm(dx,2) then 
			mu_k := obj_next;
		    fi:
		    
		    x_next := x_cur + dx;
		    obj := obj_next;
		    printf("k = %d, mu=%e, ||x_k - x_k+1|| = %e, ||f(x_k+1)||/||dx|| = %e, ||f(x_k)||=%e\n",i,mu_k, norm(dx,2), obj_next/norm(dx,2), obj_next);
		    
		    if norm(x_cur-x_next,2) < tol or obj < tol  then break fi:
		
		    x_cur := x_next;
		    
		    residuals := residuals_next; #recycle computation
		    
		    #J:= numeric_hessian22(my_func,my_args,x, x_cur);
		    Z_next := Z_next_candidate; #candidate was sucessful.
		    J:= fast_symbolic_hessian_full(A,t,DeltaA,DeltaA_vars,fstar,h,Lambda,X,Z_next);
		    JTJ := LinearAlgebra:-Transpose(J).J;
		else 
		    #printf("Invalid step. Decreasing mu\n");
		    mu_k := mu_k * (lm_damp_factor);  #take a smaller step and try again
		    
		    if norm(dx,2)<tol or  obj < tol  then break fi:
		
		fi:
		
		
		
	    
		
	    od:
	    
	    current_pt := subs(Equate(Vector(Z),x_cur));
	    t_end := time():
	    printf("Computation took %f seconds\n",t_end-t_start);
	    printf("||DA||=%e \n", poly_norm_opt(subs(current_pt,DeltaA),t));
	    return [obj ,current_pt, t_end-t_start];
	    
end proc:


#Proc: numeric_lm
#Input:  some other parameters
#Output: result of a Levenberg-Marquardt iteration
#Notes: I have no fucking clue if this will work.
#Notes: Merit function used is quadratic L2 norm of residuals and function is implicit as my_func

numeric_lm := proc(my_func,my_args,x,x_init,tol:=1e-16,max_iter := 100000,lm_param:= 1,lm_amp_factor:=1e-1,lm_damp_factor:=1e1)
      local i,
	    grad,
	    J,
	    JTJ,
	    rhs_k, #first argument to linear solve
	    lhs_k,#second rgument to linear solve
	    current_pt,
	    x_next,
	    Imat,
	    residuals,
	    obj,
	    j,
	    dx,
	    x_next_candidate,
	    x_cur,
	    mu_k,#the current levenberg-marquardt parameter
	    safety_factor #for LM
	    ;
	    
	    safety_factor:= 10^(-.5*Digits);
	    #default h for now
	    J:= numeric_hessian22(my_func,my_args,x,x_init);
	    JTJ := LinearAlgebra:-Transpose(J).J;

	    printf("Computed Numerical Hessian...\n");
	    residuals := numeric_gradient2(my_func,my_args,x,x_init);
	    obj := norm(residuals,2);
	    
	    
	    printf("Computed numerical gradient...\n");
	    
	    Imat:=Matrix(LinearAlgebra:-Dimension(JTJ));
	    for j from 1 to LinearAlgebra:-RowDimension(Imat) do 
		Imat[j,j]:=1:
	    od:
	    
#	    printf("Initial guess has dimension %d,%d and rank %d\n", LinearAlgebra:-Dimension(JTJ),LinearAlgebra:-Rank(subs(x_init,JTJ)));

	  
	    
	    x_cur :=x_init;
	    mu_k := lm_param;
	    
	    printf("Beginning iteration\n");

	    for i from 1 to max_iter do
		rhs_k := JTJ + mu_k * Imat;
		lhs_k := -1*LinearAlgebra:-Transpose(J).residuals;
		dx := LinearAlgebra:-LinearSolve(rhs_k,lhs_k);
		
		#printf("Computed dx\n");

		x_next_candidate := x_cur + dx;
		
		residuals_next := numeric_gradient2(my_func,my_args,x,x_next_candidate);
		obj_next := norm(residuals_next,2);
		
		#return obj_next, obj;
	
		if obj_next <= obj  then #valid step 
		#printf("Valid step. Proceeding to lower mu.\n");
		    mu_k :=max( mu_k*(lm_amp_factor),safety_factor); #take a smaller step next iteration
		   
		   
		    #this is based on a criteria from a paper by Wright.
		    #|| residuals||_2 approximates the solution, so the penalty term
		    #blows up at the same rate as the solution, giving asymptotic quadratic convergence.
		    if  norm(dx,2) <=  obj and obj <= (1.5)*norm(dx,2) then 
			mu_k := obj_next;
		    fi:
		    
		    x_next := x_cur + dx;
		    obj := obj_next;
		    printf("k = %d, mu=%e, ||x_k - x_k+1|| = %e, ||f(x_k+1)||/||dx|| = %e, ||f(x_k)||=%e\n",i,mu_k, norm(dx,2), obj_next/norm(dx,2), obj_next);
		    
		    if norm(x_cur-x_next,2) < tol or obj < tol then break fi:
		
		    x_cur := x_next;
		    
		    residuals := residuals_next; #recycle computation
		    
		    J:= numeric_hessian22(my_func,my_args,x, x_cur);
		    JTJ := LinearAlgebra:-Transpose(J).J;

		
		
		else 
		    #printf("Invalid step. Decreasing mu\n");
		    mu_k := mu_k * (lm_damp_factor);  #take a smaller step and try again
		fi:
		
		
		
	    
		
	    od:
	    
	    current_pt := subs(Equate(Vector(x),x_cur));
	    return [obj , Equate(x,x_cur)];
	    
end proc:


#Proc: numeric_gradient4
#Input: curent point to evaluate our (implicit) function my_func,
#Ouput: Gradient computed numerically of my_func at point x_init
#Notes: A is now a function argument, and not a function to approximate derivative.
numeric_gradient4 := proc(my_func,my_args,x,x_init,h:=sqrt(evalf(Digits))^(-1)) #possibly use a higher-order accuracy method here.
      local i,
	    Df,
	    x_h,
	    f1,
	    f2,
	    f3,
	    f4
	    ;

	    x_h := Vector(ArrayNumElems(x));
	    Df := Vector(ArrayNumElems(x));
	    
	    
	    
	    for i from 1 to ArrayNumElems(x) do 
		  x_h[i] := h; 
		  if i > 1 then x_h[i-1] :=0 fi: #re-use the same vector to reduce memory allocations 
		  
		  f1:= my_func( op(my_args), Equate(x,x_init+2*x_h));
		  f2:= my_func( op(my_args), Equate(x,x_init+x_h));
		  f3:= my_func( op(my_args), Equate(x,x_init-x_h));
		  f4:= my_func( op(my_args), Equate(x,x_init-2*x_h));
		
		  Df[i] :=  1./(12*h) * (-1*f1 +  8*f2 - 8*f3 + f4);

	    od:
	    
	    return Df; 
	    
	    
end proc:

#Proc: numeric_gradient2
#Input: curent point to evaluate our (implicit) function my_func,
#Ouput: Gradient computed numerically of my_func at point x_init
#Notes: A is now a function argument, and not a function to approximate derivative.
numeric_gradient2 := proc(my_func,my_args,x,x_init,h:=sqrt(evalf(Digits))^(-1)) #possibly use a higher-order accuracy method here.
      local i,
	    Df,
	    x_h,
	    f1,
	    f2
	    ;

	    x_h := Vector(ArrayNumElems(x));
	    Df := Vector(ArrayNumElems(x));
	    
	    
	    
	    for i from 1 to ArrayNumElems(x) do 
		  x_h[i] := h; 
		  if i > 1 then x_h[i-1] :=0 fi: #re-use the same vector to reduce memory allocations 
		  
		  f1:= my_func( op(my_args), Equate(x,x_init+x_h));
		  f2:= my_func( op(my_args), Equate(x,x_init-1*x_h));
		
		  Df[i] :=  1./(2*h) * (f1-f2);

	    od:
	    
	    return Df; 
	    
	    
end proc:

#Proc: numeric_jacobian4_constraint
#Input: A,A_init,DeltaA,DeltaA_vars #do not confuse h wih the gcd usage...
#Output: Jacobian of vec(adj(A)) at a point
#Notes: It is possible to get a closed-form for this using kronecker products
#	but is almost not worth implementing due to the complexity in sizing arrays properly

#Warning: The vars are implicitly defined by DeltaA, but the ordering of DeltaA_vars could 
#correspond to row/stacked vs column/stacked and would be a pain in the ass to debug.
numeric_jacobian4_adjoint := proc(A,t,A_init,DeltaA,DeltaA_vars,h:=sqrt(evalf(Digits))^(-1))
	local   i,
		J,
		d, n, degree_bound, #needed to vectorize the final results correctly.
		A_dim, A_h,
		DA_vars_dim, default_DA_pt,
		x1,x2,x3,x4, #points to evaluate at
		G1,G2,G3,G4,G #vector-derivative points
		;
		
		default_DA_pt := [seq(DeltaA_vars[i] = 0, i=1..nops(DeltaA_vars))]; #needed for some stuff
		d:=matrix_degree(A,t);
		n:=LinearAlgebra:-RowDimension(A);
		size_bound := (n-1)*d+1; #degree bound on the adjoint computation plus room for a constant.
		J:=Matrix(0,0);
		for i from 1 to nops(DeltaA_vars) do 
		      #get the right polynomial coefficient here.
		      A_h :=  subs(default_DA_pt,subs(DeltaA_vars[i]=h,DeltaA));
		   
		      
		      x1 :=  A+2*A_h;
		      x2 :=  A+  A_h;
		      x3 :=  A-  A_h;
		      x4 :=  A-2*A_h;
		      
		      G1 := symbolic_numeric_adjoint(x1);
		      G2 := symbolic_numeric_adjoint(x2);
		      G3 := symbolic_numeric_adjoint(x3);
		      G4 := symbolic_numeric_adjoint(x4);
		
		
		      G:= convert(1/(12. * h) *(-1*G1+8*G2-8*G3+G4),Vector); #vectorize
		    
		    G:=convert(`<,>`( seq(Vector(size_bound,PolynomialTools:-CoefficientVector(G[j],t)),j=1..ArrayNumElems(G))),Vector);
		      if ArrayNumElems(J) = 0 then 
			  J := G;
		      else
			  J:= `<|>`(J,G);
		      fi:
			  
		od:
		
		
		return J;
end proc:

#Note: there is a bug in the code using f_dim  for dimensions.
#and needs to be fixed...


#Proc: numeric_hessian44
#Input: Matrix poly A with initial point to compute hessian at and list of variables in x
#Output: 4th order truncation error  approximation to Hessian matrix at x_init of A
numeric_hessian44 := proc(my_func,my_args,x,x_init,h:=sqrt(evalf(Digits))^(-1)) #possibly use a higher-order accuracy method here.
	local   i,
		H,
		f_dim,
		x_dim,
		x_h,
		x1,x2,x3,x4, #points to evaluate at
		G1,G2,G3,G4 #vector-derivative points
		;
		
		H:=Matrix();
		x_h := Vector(ArrayNumElems(x));
		for i from 1 to ArrayNumElems(x) do
		     if i > 1 then 
			x_h[i-1] := 0;
		     fi:
		     x_h[i]:=h;
		     
		     G1 := numeric_gradient4(my_func,(my_args),x,x_init+2*x_h);
		     G2 := numeric_gradient4(my_func,(my_args),x,x_init+x_h);
		     G3 := numeric_gradient4(my_func,(my_args),x,x_init-x_h);
		     G4 := numeric_gradient4(my_func,(my_args),x,x_init-2*x_h);
		     
		     G:= 1/(12.*h) * (-1*G1+8*G2-8*G3+G4);
		     
		     if ArrayNumElems(H) =0 then 
			H:=G;
		     else
			H:=`<|>`(H,G);
		     fi:
		     
		od:
		
		return H;
end proc:




#Proc: numeric_hessian22
#Input: Matrix poly A with initial point to compute hessian at and list of variables in x
#Output: 2nd order truncation error  approximation to Hessian matrix at x_init of A
numeric_hessian22 := proc(my_func,my_args,x,x_init,h:=sqrt(evalf(Digits))^(-1)) #possibly use a higher-order accuracy method here.
	local   i,
		H,
		f_dim,
		x_dim,
		x_h,
		G1,G2 #vector-derivative points
		;
		
		H:=Matrix();
		x_h := Vector(ArrayNumElems(x));
		for i from 1 to ArrayNumElems(x) do
		     if i > 1 then 
			x_h[i-1] := 0;
		     fi:
		     x_h[i]:=h;
		     
		     G1 := numeric_gradient2(my_func,(my_args),x,x_init+x_h,h);
		     G2 := numeric_gradient2(my_func,(my_args),x,x_init-x_h,h);
		     
		     G:= 1/(2.*h) * (G1-G2);
		     if ArrayNumElems(H) =0 then 
			H:=G;
		     else
			H:=`<|>`(H,G);
		     fi:
		     
		od:
		
		return H;
end proc:


#Proc: numeric_hessian24
#Input: Matrix poly A with initial point to compute hessian at and list of variables in x
#Output: 2nd order truncation error  approximation to Hessian matrix at x_init of A
numeric_hessian24 := proc(my_func,my_args,x,x_init,h:=sqrt(evalf(Digits))^(-1)) #possibly use a higher-order accuracy method here.
	local   i,
		H,
		f_dim,
		x_dim,
		x_h,
		G1,G2 #vector-derivative points
		;
		
		H:=Matrix();
		x_h := Vector(ArrayNumElems(x));
		for i from 1 to ArrayNumElems(x) do
		     if i > 1 then 
			x_h[i-1] := 0;
		     fi:
		     x_h[i]:=h;
		     
		     G1 := numeric_gradient4(my_func,(my_args),x,x_init+x_h,h);
		     G2 := numeric_gradient4(my_func,(my_args),x,x_init-x_h,h);
		     
		     G:= 1/(2.*h) * (G1-G2);
		     if ArrayNumElems(H) =0 then 
			H:=G;
		     else
			H:=`<|>`(H,G);
		     fi:
		     
		od:
		
		return H;
end proc:
#Proc: extract_guess_information
#Input:	Matrix polynomial A in indeterminant t and the desired perturbation structure method
#Note:	This should be support  or entry almost always
#Purpose: This code computes information we need to pass into the Lagrangian function
#	and sets up an ANSATZ

extract_guess_information_original := proc(A,A_init,x:=t,max_gcd_degree:=2,method:="support",symbol:='a')
	local
	fstar_init,fstar_init_pt, fstar,fstar_vars,fstar_guesses,
	h_init,h_init_pt,h,h_vars,h_guesses,
	opt_index,
	Lambda,
	hfs_vec,J_fh,J_adj,J, #Jacobian constraint computation
	DA_vec,
	adjA,#the adjoint evaluated at a point.
	DeltaA,DeltaA_vars,DeltaA_init,DeltaA_init_pt,
	n,d,adjoint_degree_bound #needed to create bounds for number of Lagrange multipliers
	;
	
	DeltaA,DeltaA_vars:= generate_perturbation(A,x,method,symbol); #generate the perturbation structure
	DeltaA_init :=A_init-A;
	#DeltaA_init := Matrix(LinearAlgebra:-Dimension(A));
	
	
	if check_entry_degrees(A,A_init,x) = false then 
	    WARNING("Initial guess has invalid support.");
	    
	    return [];
	fi:
	#quite a reader here, just combine coefficient vectors and throw away equations not agreeing with the support.
	DeltaA_init_pt := [seq(op(Equate(Vector(1+max(degree(DeltaA(i),x), degree(DeltaA_init(i),x)),PolynomialTools:-CoefficientVector(DeltaA(i),x)),Vector(1+max(degree(DeltaA(i),x), degree(DeltaA_init(i),x) ),PolynomialTools:-CoefficientVector(DeltaA_init(i),x)))), i=1..ArrayNumElems(DeltaA))];

	#return DeltaA_init_pt;
	#1.
	d:= matrix_degree(A,x);
	n := min(LinearAlgebra:-Dimension(A)); 
	adjoint_degree_bound := (n-1)*d;
	
	
	adjA := convert(symbolic_numeric_adjoint(A_init),list); #this routine is located somewhere...
	
	#currently take the maximum degree of the co-factors for possible GCD's to try.
	#this is done because some examples could have co-factors nearby-zero
	#and would produce a small residual even with an $h$ that has no way to divide them.
	
	#This is subjective and can be changed. 
	adjA_degree := max(1,max(seq(degree(adjA[i],x),i=1..nops(adjA))));  
	
	#go compute an extract some guesses. 
	#The approximate GCD method employed can be changed to another suitable method.
	#the current implementation is a least-squares approximation to linear/quadratic factors
	#This method somewhat fails if the degree of the approximate GCD is guessed too low.
	#An alternative guess heuristic is SVD/QR-GCD of multiple polynomials, however this is too much work to code
	#and robustly debug for good performance, although possible. 
	
	#the method is very effective on small (absolute) error problems.
	h_guesses := [seq(generate_initial_agcd_degree(adjA,i,x), i=1..min(adjA_degree,max_gcd_degree) )];
	fstar_guesses := [seq(poly_least_squares_division_implicit(adjA, h_guesses[i]),i=1..nops(h_guesses))];
	
	opt_index:= min[index]([seq(fstar_guesses[i][1], i=1..nops(fstar_guesses))]);
	
	#some information describing the initial guess to the user. can tune this parameter later...
	printf("Extraced co-factor and gcd Error:  %e, Degree:  %d\n", fstar_guesses[opt_index][1],opt_index);
	
	
	#these are the evaluated points, not the ANSATZ. fstar_init will almost surely have leading zero terms. 
	h_init := h_guesses[opt_index];
	fstar_init := fstar_guesses[opt_index][2];
	
	
	#WARNING:
	#We are using the normalization that h is monic. 
	#The "funny" arithemetic here makes sure that this requirement is implicit.
	h := add(hh[i]*x^i,i=0..opt_index-1) + x^opt_index;
	
	h_vars := PolynomialTools:-CoefficientVector(h-x^opt_index,x);
	
	#this assumes the degree of h cannot change. This may not be entirely accurate.
	#a proper implementation would dynamically re-normalize the leading coefficient based on iterates nearby a solution.
	fstar := [seq(add(fs[j,i]*x^i,i=0..adjoint_degree_bound-opt_index),j=1..nops(fstar_init))]; 
	
	fstar_vars := `<,>`(seq(PolynomialTools:-CoefficientVector(fstar[i],x),i=1..nops(fstar)));
	
	
	
	#h and fstar are sort of initialized...


	
	#compute least squares division residual
	#loop back up and stuff
	
	#2. Create initial points and stuff 
	#Note: n=2 is a special case  and the code does not currently work on it for some reason.
	#obviously this should be debugged, but I cbf atm.
	
	
	#Compute h points and hvars
	#Compute fstar vars and plug in initial point, note higher-order terms that remain zero
	
		#this part is easy.
	h_init_pt := Equate(h_vars, Vector(PolynomialTools:-CoefficientVector(h_init,x)[1..opt_index]));
	#this is a bit trickier, requires nested seq's
	fstar_init_pt:= Equate(fstar_vars,`<,>`(seq(Vector(adjoint_degree_bound-opt_index+1, PolynomialTools:-CoefficientVector(fstar_init[i],x)), i=1..nops(fstar_init) )));
	
	#return (seq(degree(adjA[i],x),i=1..nops(adjA)));
	
	#return h_init;
	
	#return [h_init_pt, fstar_init_pt];
	
	#3. compute the number of Lagrange multipliers and ANSATZ for them
	Lambda := Vector(n^2*(adjoint_degree_bound+1),i->'lambda'[i]); #there are O(n^3d) constraint equations 
	
	X:= [op(DeltaA_vars), op(convert(`<,>`(h_vars,fstar_vars),list))];
	
	X_init := [op(DeltaA_init_pt), op(h_init_pt), op(fstar_init_pt)];
	
	#Step 4: Numerically compute the Jacoian of (Adj(A)-fstar*h) at $DeltaA=init_pt$
	
	
	#splice two Jacobians together.

	##this quantity should have the same number of elements as Lambda, by construction.
	
	hfs_vec := -1*convert(`<,>`( seq(PolynomialTools:-CoefficientVector(fstar[i]*h,x),i=1..nops(fstar))),Vector);

	
	

        #use the symbolic numeric algorithm in the other file for now.
	J:=fast_symbolic_jacobian_constraints(A,x,DeltaA,DeltaA_vars,fstar,h,Lambda,X,X_init);

	#Step 5: Compute initial guess for Lambda via linear least squares. 
	#the objective is ||DeltaA||^2
	
	DA_vec := convert(DeltaA,Vector);
	#this doesn't work due to sizing mismatching occuring except in the most dense case.
	#DA_vec := convert(`<,>`( seq( Vector(adjoint_degree_bound+1,PolynomialTools:-CoefficientVector(DA_vec[i],x)),i=1..ArrayNumElems(DA_vec))),Vector);
	DA_vec := convert(DeltaA_vars,Vector);
	DA_vec := subs(DeltaA_init_pt,DA_vec);
	DA_vec := convert(`<,>`(DA_vec, Vector(ArrayNumElems(fstar_vars) +ArrayNumElems(h_vars))),Vector);
	
	
	#the -2 scaling is because 
	#L = ||DeltaA||^2 + Lambda^T(...)
	#so Grad(L) = 0 => -2 DeltaA = Lambda^T. J(...)
	my_soln := Optimization:-LSSolve([-2*DA_vec,LinearAlgebra:-Transpose(J)]);
	
	Lambda_init := Equate(Lambda,my_soln[2]);
	
	printf("Initial Lagrange multiplier residual: %e \n", my_soln[1]);
	
	  
	
	return [DeltaA,fstar,h], [X,X_init],[convert(Lambda,list),Lambda_init], [DeltaA_vars,fstar_vars,h_vars],[J];
	
	
	
	#Note: Possible bug here could be due to ordering of terms and mis-sizing stuff from other functions.
	#at the moment this should be easy to detect.
end proc:


#Proc: extract_guess_information
#Input:	Matrix polynomial A in indeterminant t and the desired perturbation structure method
#Note:	This should be support  or entry almost always
#Purpose: This code computes information we need to pass into the Lagrangian function
#	and sets up an ANSATZ

extract_guess_information := proc(A,A_init,x:=t,min_gcd_degree:=1,max_gcd_degree:=2,method:="support",symbol:='a')
	local
	fstar_init,fstar_init_pt, fstar,fstar_vars,fstar_guesses,
	h_init,h_init_pt,h,h_vars,h_guesses,
	opt_index,
	Lambda,
	hfs_vec,J_fh,J_adj,J, #Jacobian constraint computation
	DA_vec,
	adjA,#the adjoint evaluated at a point.
	DeltaA,DeltaA_vars,DeltaA_init,DeltaA_init_pt,
	n,d,adjoint_degree_bound #needed to create bounds for number of Lagrange multipliers
	;
	
	DeltaA,DeltaA_vars:= generate_perturbation(A,x,method,symbol); #generate the perturbation structure
	DeltaA_init :=A_init-A;
	#DeltaA_init := Matrix(LinearAlgebra:-Dimension(A));
	
	
	if check_entry_degrees(A,A_init,x) = false then 
	    WARNING("Initial guess has invalid support.");
	    
	    return [];
	fi:
	
	if min_gcd_degree > max_gcd_degree then 
	    WARNING("Initial guess has invalid support.");
	    
	    return [];
	
	fi:
	
	#quite a reader here, just combine coefficient vectors and throw away equations not agreeing with the support.
	DeltaA_init_pt := [seq(op(Equate(Vector(1+max(degree(DeltaA(i),x), degree(DeltaA_init(i),x)),PolynomialTools:-CoefficientVector(DeltaA(i),x)),Vector(1+max(degree(DeltaA(i),x), degree(DeltaA_init(i),x) ),PolynomialTools:-CoefficientVector(DeltaA_init(i),x)))), i=1..ArrayNumElems(DeltaA))];

	#return DeltaA_init_pt;
	#1.
	d:= matrix_degree(A,x);
	n := min(LinearAlgebra:-Dimension(A)); 
	adjoint_degree_bound := (n-1)*d;
	
	
	adjA := convert(symbolic_numeric_adjoint(A_init),list); #this routine is located somewhere...
	
	#currently take the maximum degree of the co-factors for possible GCD's to try.
	#this is done because some examples could have co-factors nearby-zero
	#and would produce a small residual even with an $h$ that has no way to divide them.
	
	#This is subjective and can be changed. 
	adjA_degree := max(1,max(seq(degree(adjA[i],x),i=1..nops(adjA))));  
	
	#go compute an extract some guesses. 
	#The approximate GCD method employed can be changed to another suitable method.
	#the current implementation is a least-squares approximation to linear/quadratic factors
	#This method somewhat fails if the degree of the approximate GCD is guessed too low.
	#An alternative guess heuristic is SVD/QR-GCD of multiple polynomials, however this is too much work to code
	#and robustly debug for good performance, although possible. 
	
	#the method is very effective on small (absolute) error problems.
	h_guesses := [seq(generate_initial_agcd_degree(adjA,i,x), i=1..min(adjA_degree,max_gcd_degree) )];
	fstar_guesses := [seq(poly_least_squares_division_implicit(adjA, h_guesses[i]),i=1..nops(h_guesses))];
	
	opt_index:= min[index]([seq(fstar_guesses[i][1], i=1..nops(fstar_guesses))]);
	
	if min_gcd_degree <= nops(h_guesses) and opt_index < min_gcd_degree then 
                opt_index := min_gcd_degree;
	fi:
	
	#some information describing the initial guess to the user. can tune this parameter later...
	printf("Extraced co-factor and gcd Error:  %e, Degree:  %d\n", fstar_guesses[opt_index][1],opt_index);
	
	
	#these are the evaluated points, not the ANSATZ. fstar_init will almost surely have leading zero terms. 
	h_init := h_guesses[opt_index];
	fstar_init := fstar_guesses[opt_index][2];
	
	
	#WARNING:
	#We are using the normalization that h is monic. 
	#The "funny" arithemetic here makes sure that this requirement is implicit.
	h := add(hh[i]*x^i,i=0..opt_index-1) + x^opt_index;
	
	h_vars := PolynomialTools:-CoefficientVector(h-x^opt_index,x);
	
	#this assumes the degree of h cannot change. This may not be entirely accurate.
	#a proper implementation would dynamically re-normalize the leading coefficient based on iterates nearby a solution.
	fstar := [seq(add(fs[j,i]*x^i,i=0..adjoint_degree_bound-opt_index),j=1..nops(fstar_init))]; 
	
	fstar_vars := `<,>`(seq(PolynomialTools:-CoefficientVector(fstar[i],x),i=1..nops(fstar)));
	
	
	
	#h and fstar are sort of initialized...


	
	#compute least squares division residual
	#loop back up and stuff
	
	#2. Create initial points and stuff 
	#Note: n=2 is a special case  and the code does not currently work on it for some reason.
	#obviously this should be debugged, but I cbf atm.
	
	
	#Compute h points and hvars
	#Compute fstar vars and plug in initial point, note higher-order terms that remain zero
	
		#this part is easy.
	h_init_pt := Equate(h_vars, Vector(PolynomialTools:-CoefficientVector(h_init,x)[1..opt_index]));
	#this is a bit trickier, requires nested seq's
	fstar_init_pt:= Equate(fstar_vars,`<,>`(seq(Vector(adjoint_degree_bound-opt_index+1, PolynomialTools:-CoefficientVector(fstar_init[i],x)), i=1..nops(fstar_init) )));
	
	#return (seq(degree(adjA[i],x),i=1..nops(adjA)));
	
	#return h_init;
	
	#return [h_init_pt, fstar_init_pt];
	
	#3. compute the number of Lagrange multipliers and ANSATZ for them
	Lambda := Vector(n^2*(adjoint_degree_bound+1),i->'lambda'[i]); #there are O(n^3d) constraint equations 
	
	X:= [op(DeltaA_vars), op(convert(`<,>`(h_vars,fstar_vars),list))];
	
	X_init := [op(DeltaA_init_pt), op(h_init_pt), op(fstar_init_pt)];
	
	#Step 4: Numerically compute the Jacoian of (Adj(A)-fstar*h) at $DeltaA=init_pt$
	
	
	#splice two Jacobians together.

	##this quantity should have the same number of elements as Lambda, by construction.
	
	hfs_vec := -1*convert(`<,>`( seq(PolynomialTools:-CoefficientVector(fstar[i]*h,x),i=1..nops(fstar))),Vector);

	
	

        #use the symbolic numeric algorithm in the other file for now.
	J:=fast_symbolic_jacobian_constraints(A,x,DeltaA,DeltaA_vars,fstar,h,Lambda,X,X_init);

	#Step 5: Compute initial guess for Lambda via linear least squares. 
	#the objective is ||DeltaA||^2
	
	DA_vec := convert(DeltaA,Vector);
	#this doesn't work due to sizing mismatching occuring except in the most dense case.
	#DA_vec := convert(`<,>`( seq( Vector(adjoint_degree_bound+1,PolynomialTools:-CoefficientVector(DA_vec[i],x)),i=1..ArrayNumElems(DA_vec))),Vector);
	DA_vec := convert(DeltaA_vars,Vector);
	DA_vec := subs(DeltaA_init_pt,DA_vec);
	DA_vec := convert(`<,>`(DA_vec, Vector(ArrayNumElems(fstar_vars) +ArrayNumElems(h_vars))),Vector);
	
	
	#the -2 scaling is because 
	#L = ||DeltaA||^2 + Lambda^T(...)
	#so Grad(L) = 0 => -2 DeltaA = Lambda^T. J(...)
	my_soln := Optimization:-LSSolve([-2*DA_vec,LinearAlgebra:-Transpose(J)]);
	
	Lambda_init := Equate(Lambda,my_soln[2]);
	
	printf("Initial Lagrange multiplier residual: %e \n", my_soln[1]);
	
	  
	
	return [DeltaA,fstar,h], [X,X_init],[convert(Lambda,list),Lambda_init], [DeltaA_vars,fstar_vars,h_vars],[J];
	
	
	
	#Note: Possible bug here could be due to ordering of terms and mis-sizing stuff from other functions.
	#at the moment this should be easy to detect.
end proc:




#Proc: numeric_gradient2_turbo
#Input: same as numeric gradient2
#Output: Gradient of the Lagrangian
#Notes: Optimized for the current Lagrangian function.
numeric_gradient2_turbo := proc()

end proc:

#Proc: numeric_hessian22_turbo
#Input: Same as numeric_hessian22
#output: Hessian of the Lagrangian
#Notes: Optimized for the current Lagrangian function.
#	This function exploits symmetry and should be approximately 8 times faster.

numeric_hessian22_turbo := proc()

end proc:

#Change my_func based on what problem we are solving... Should be Lagrangian in later implementations.
#Proc: my_func
#Input: Matrix poly A
#Output: Approximation to the Adjoint matrix of A
#Notes: This uses the built in maple routine and may not be optimal 
my_func := proc(A)
    return evalf(LinearAlgebra:-Adjoint(convert(A,rational)));
end proc:


#Proc: numeric_lagrangian
#Input : A is user input, t is indeterminant
#	DeltaA is ANSATZ form
#	fstar,h are ANSATZ form
#	fstar is a vector of polynomials and h is a polynomial scalar argument
#	we assume that the leading coefficient of h is implicitly normalalized in this computation 
#	Lambda is vector of Lagrange multipliers
#	Z_pt is the point to eveluate the Lagrangian at
#Output: Lagrangian function eveluated at the point Z_pt.

numeric_lagrangian := proc(A,x,DeltaA,fstar,h,Lambda,Z_pt)
	  local DA_pt,
		fstar_pt,
		h_pt,
		adj_vec,
		coeff_vec,
		Lambda_pt,
		L,
		n,
		d
		;
    
		DA_pt := subs(Z_pt,DeltaA);
		fstar_pt := convert(subs(Z_pt,fstar),Vector);
		h_pt :=  subs(Z_pt,h);
		
		Lambda_pt := convert(subs(Z_pt,Lambda),Vector);
		
		#vector over R[t]
		adj_vec := convert(symbolic_numeric_adjoint(A+DA_pt),Vector);
		
		coeff_vec := vectorize_adjoint(A,adj_vec-fstar_pt * h_pt,x); #dimensions should be the same.
		
		L:= poly_norm_opt(DA_pt,x)^2 + (LinearAlgebra:-Transpose(Lambda_pt).coeff_vec)[1]; 
		
		return L;
		
end proc:



#Proc: symbolic_numeric_adjoint
#Input: Matrix (polynomial) A
#Output: Adjoint of A computed numerically
#Notes:  Uses a symbolic-numeric algorithm that converts A to rationals and returns a numeric approximation
symbolic_numeric_adjoint := proc (A)
	local adj_A
	      ;
	      #return evalf(LinearAlgebra:-Adjoint(convert(A,rational)));
	      return LinearAlgebra:-Adjoint(evalf(A));
end proc:


#Proc: vectorize_adjoint
#Input: Adjoint matrix Adj(A) with indeterminant $t$
#Output: Vectorized adjoint matrix
#Notes: This is needed because the Adjoint operator lives in a O(n^3d) space, as each degree in $t$ is bounded above in degree by $(n-1)d$
#	Dimension Mis-matching could happen in derivative computation
vectorize_adjoint := proc(A,Adj,t)
      local 
	    adj_vec,
	    n,
	    d,
	    size_bound
	    ;
	    
	    #adj := convert(symbolic_numeric_adjoint(A),Vector); #get the Adjoint matrix via computation.
	    n := LinearAlgebra:-RowDimension(A); #Assume square matrix for now.
	    d := matrix_degree(A,t);
	    size_bound := (n-1)*d+1;
	    
	    
	    adj_vec:= `<,>` (seq((  Vector( PolynomialTools:-CoefficientVector(Adj(i),t),size_bound)), i=1..ArrayNumElems(Adj)));
	   
	    
	    return convert(adj_vec,Vector); 
end proc:


#Proc: check_entry_degrees 
#Input: Matrix polynomials A and A_init
#Output: True if deg(A(i)) >= degree(A_init(i))
check_entry_degrees := proc(A,A_init,x)
    local i;
	  for i from 1 to ArrayNumElems(A) do 
	      if degree (A(i),x) < degree (A_init(i),x) then return false fi:
	  od:
	  return true;
end proc:


#Proc: fast_symbolic_gradient
#Input: Lagrange multipliers and Jacobian
#Output: Output: Lambda^T * J + 2*DeltaA, the gradient
#Purpose: Compute the gradient of the Lagrangian "fast"
fast_symbolic_gradient := proc(A,t,DeltaA,DeltaA_vars,fstar,h,Lambda,X,Z_init)
	  
	  local adjA_vec,
		hfs_vec,
		#fh_vars,
		J,
		G
		;
	  adjA_vec := vectorize_adjoint(A,symbolic_numeric_adjoint(A+subs(Z_init,DeltaA)),t);
	 # fh_vars := X[nops(DeltaA_vars)+1..nops(X)];
	  hfs_vec := subs(Z_init,compute_hfs_vec(fstar,h,t));
	  J:= fast_symbolic_jacobian_constraints(A,t,DeltaA,DeltaA_vars,fstar,h,Lambda,X,Z_init);
	  
	  G:= 2*Vector(convert(subs(Z_init,X[1..nops(DeltaA_vars)]),Vector),nops(X));
	  G:= G+ LinearAlgebra:-Transpose(J) . convert(subs(Z_init,Lambda),Vector);
	  
	  G:=convert(`<,>`(G,adjA_vec-hfs_vec),Vector);
	  return G;
end proc:


#Proc: compute_hfs_vec
#Input: fstar,h
#Ouput: convert fstar*h into a real vector from a polynomial one. 
compute_hfs_vec := proc(fstar,h,t)
    return     convert(`<,>`( seq(PolynomialTools:-CoefficientVector(fstar[i]*h,t),i=1..nops(fstar))),Vector);
end proc:

#Proc: fast_symbolic_hessian_full
#Input: 
#Output:
#Purpose: Compute the full hessian matrix
#Notes: Splice pieces together to generate the requisite block structure
fast_symbolic_hessian_full:= proc(A,t,DeltaA,DeltaA_vars,fstar,h,Lambda,X,Z_init)
	local H_A,
	      H_fh,
	      H,
	      J,
	      Lambda_pt,
	      fh_vars,
	      hfs_vec
	      ;
	      
	      fh_vars := X[nops(DeltaA_vars)+1..nops(X)];
	      Lambda_pt := subs(Z_init,convert(Lambda,Vector[row]));
	      hfs_vec := -1*compute_hfs_vec(fstar,h,t);
	      
	      H_A := fast_symbolic_hessianXX(A,t,DeltaA,DeltaA_vars,fstar,h,Lambda,X,Z_init);
	      
	      #Note that this step can be done once since the Hessian w.r.t. fstar and h is constant, as the function in question is bi-linear
	      #code can be modified to receive this as an argument instead.
	      H_fh := VectorCalculus:-Hessian(Lambda_pt.hfs_vec,fh_vars);
	      
	      J := fast_symbolic_jacobian_constraints(A,t,DeltaA,DeltaA_vars,fstar,h,Lambda,X,Z_init);
	      
	      
	      H:= LinearAlgebra:-DiagonalMatrix([H_A,H_fh]);
	      H:= `<,>`(H,J);
	      H:=`<|>`(H,`<,>`(LinearAlgebra:-Transpose(J),Matrix(LinearAlgebra:-RowDimension(J))));
	      return H;
end proc:

#Proc: fast_symbolic_hessianXX
#Input: [A,t,DeltaA,fstar,h,Lambda,X,Z]
#Output: Hessian  w.r.t. X evaluated at the point Z
#Notes: This algorithm tries to do a fast symbolic-numeric hessian computation that exploits structure.
fast_symbolic_hessianXX := proc(A,t,DeltaA,DeltaA_vars,fstar,h,Lambda,X,Z_init)
     local i, j,k,ell,
	    hfs_vec,
	    fh,
	    fh_vars,
	    n,
	    Adj,
	    J,J_temp, #the jacobian 
	    active_vars, active_vars_ij,#needed for book keeping
	    active_pt, #needed for book keeping
	    A_copy,
	    d, #degree bound
	    H, #hessian
	    Lambda_vec,
	    adjA_vec, #this is used for some black magic. 
	    size_bound
	    ;
	    
	    d:=matrix_degree(A,t);
	    n := min(LinearAlgebra:-Dimension(A));
	    size_bound := (n-1)*d+1;
	    fh := fstar * h;
	    fh_vars := X[nops(DeltaA_vars)+1..nops(X)];
	    J:=Matrix(0,0):
	    #needed for later.
	    hfs_vec := -1*convert(`<,>`( seq(PolynomialTools:-CoefficientVector(fstar[i]*h,t),i=1..nops(fstar))),Vector);

	    #we need to compute the Jacobian in two variables now for Hessian information.
	    #this is a bit trickier.
	    #adjA_vec := Vector(n^2 * size_bound);
	    
	    #Let's try the block method here.
	    for i from 1 to LinearAlgebra:-RowDimension(DeltaA) do 
		for j from 1 to LinearAlgebra:-ColumnDimension(DeltaA) do
		    active_vars_ij := convert(PolynomialTools:-CoefficientVector(DeltaA[i,j],t),set) minus {0};
		    adjA_vec:=Vector(n^2*size_bound); # allocate space
	   
		    for k from i+1 to LinearAlgebra:-RowDimension(DeltaA) do
			for ell from 1 to LinearAlgebra:-ColumnDimension(DeltaA) do 
			    #Due to symmetry in the underlying problem we need
			    #only N^2 choose 2 adjoint calls instead of N^2
			    if ell <> j then
			    
				active_vars := convert(PolynomialTools:-CoefficientVector(DeltaA[k,ell],t),set) minus {0};
				active_vars := active_vars_ij union active_vars;
				#we will use a block-variant of the algorithm. It should be faster. 
			    
				#this is basically the variables we are ignoring for differentiation.
				
				inactive_pt := convert(convert(DeltaA_vars,set) minus active_vars,Vector);   
				inactive_pt := Equate(inactive_pt, subs(Z_init,inactive_pt));
				
				#these are the variables we need for differentiation.
				active_pt := convert(active_vars,Vector);
				active_pt :=  Equate(active_pt, subs(Z_init,active_pt));
				
				
				active_vars := convert(active_vars,list);
				
			    
				A_copy := A+subs(inactive_pt,DeltaA); #substitute all points not used in diff
			    
				#convert into a vectorized form.
				#note that this routine uses only one Adjoint call.
				adjA_vec:=adjA_vec+ vectorize_adjoint(A,symbolic_numeric_adjoint(A_copy),t)-hfs_vec;
			    fi:
			od:
		    od:
		    
		    active_vars_ij := convert(active_vars_ij,list);
		    if ArrayNumElems(J) =0 then 
			J:= VectorCalculus:-Jacobian(adjA_vec,active_vars_ij);
		    else 
			J:= `<|>`(J,VectorCalculus:-Jacobian(adjA_vec,active_vars_ij));
			#J:= J+VectorCalculus:-Jacobian(adjA_vec,DeltaA_vars);
		    fi:
		    
		od:
	   od:
	    
	 #   J:= `<|>`(J,VectorCalculus:-Jacobian(hfs_vec,fh_vars)); 
	    
	    Lambda_vec := convert(subs(Z_init,Lambda),Vector[row]): 
	    
	    H:=VectorCalculus:-Jacobian( Lambda_vec.J, DeltaA_vars);
	   # H:=VectorCalculus:-Hessian(Lambda_vec.adjA_vec,DeltaA_vars);
	   H:=H+LinearAlgebra:-Transpose(H)+2*LinearAlgebra:-IdentityMatrix(LinearAlgebra:-Dimension(H)); #account for ${DeltaA}
	   
	    return subs(Z_init,H);
end proc:
#Proc: fast_symbolic_jacobian_constraints
#Input: [A,t,DeltaA,DeltaA_vars,fstar,h,Lambda,X,Z]
#Output: Jacobian of constraint  w.r.t. X evaluated at the point Z
#Notes: This algorithm tries to do a fast symbolic-numeric Jacobian computation that exploits structure.
#The algorithm appears to be quite fast. despite making multiple Adjoint calls (it is faster than doing a symbolic Adjoint and substituting).
fast_symbolic_jacobian_constraints := proc(A,t,DeltaA,DeltaA_vars,fstar,h,Lambda,X,Z_init)
      local i,
	    j,
	    hfs_vec,
	    fh,
	    fh_vars,
	   
	    n,
	    Adj,
	    J, #the jacobian 
	    active_vars, #needed for book keeping
	    active_pt, #needed for book keeping
	    A_copy,
	    d,
	    adjA_vec, #this is used for some black magic. 
	    size_bound
	    ;
	    
	    d:=matrix_degree(A,t);
	    n := min(LinearAlgebra:-Dimension(A));
	    size_bound := (n-1)*d+1;
	    fh := fstar * h;
	    fh_vars := X[nops(DeltaA_vars)+1..nops(X)];
	    J:=Matrix(0,0):
	    #needed for later.
	    hfs_vec := -1*convert(`<,>`( seq(PolynomialTools:-CoefficientVector(fstar[i]*h,t),i=1..nops(fstar))),Vector);

	    adjA_vec := Vector(n^2 * size_bound);
	    #Let's try the block method here.
	    for i from 1 to LinearAlgebra:-RowDimension(DeltaA) do 
		for j from 1 to LinearAlgebra:-ColumnDimension(DeltaA) do
		    active_vars := convert(PolynomialTools:-CoefficientVector(DeltaA[i,j],t),set) minus {0};
		    #we will use a block-variant of the algorithm. It should be faster. 
		    
		    
		    #this is basically the variables we are ignoring for differentiation.
		    
		    inactive_pt := convert(convert(DeltaA_vars,set) minus active_vars,Vector);   
		    inactive_pt := Equate(inactive_pt, subs(Z_init,inactive_pt));
		    
		    #these are the variables we need for differentiation.
		    active_pt := convert(active_vars,Vector);
		    active_pt :=  Equate(active_pt, subs(Z_init,active_pt));
		    
		    
		    active_vars := convert(active_vars,list);
		    
		 
		    A_copy := A+subs(inactive_pt,DeltaA); #substitute all points not used in diff
		 
		    #convert into a vectorized form.
		    #note that this routine uses only one Adjoint call.
		    adjA_vec:= vectorize_adjoint(A,symbolic_numeric_adjoint(A_copy),t)-hfs_vec;
		    
		    #this can be optimized by differentiatiing wrt to active vars and chaining
		    #rows together.
		    if ArrayNumElems(J) =0 then 
			J:= VectorCalculus:-Jacobian(adjA_vec,active_vars);
		    else 
			J:= `<|>`(J,VectorCalculus:-Jacobian(adjA_vec,active_vars));
			#J:= J+VectorCalculus:-Jacobian(adjA_vec,DeltaA_vars);
		    fi:
		od:
	   od:
	    
	    J:= `<|>`(J,VectorCalculus:-Jacobian(hfs_vec,fh_vars)); 
	    
	    return subs(Z_init,J);
end proc:


#proc: linearize_mccoy
#Purpose: produces the Lagrangian system to pass to LM
#Input: A, t, perturbation structure, desired McCoy rank approximation
#Output: gradient of the Lagrandian to pass to LM
linearize_mccoy:=proc(A,t,DeltaA,DA_vars,s_rank)
        local A_lin1,A_lin0, #leading and trailing coeff matrices
              DA_lin1,DA_lin0, #perturbation structure
              omega, r_omega,i_omega, #real + imaginary part of root
              m,n, #dimensions
              A_0,A_1,DA_0,DA_1,
              K, K_vars, #kernel
              #different constraints we need to piece to gether
              real_rank_constraint,
              imaginary_rank_constraint,
              rank_constraint,
              real_singularity_constraint,
              imaginary_singularity_constraint,
              singularity_constraint,
              constraints, #all of the constraints into a Vector
              Lambda,
              omega_vars,
              DA_lin,
              A_lin,
              L,
              vars,
              residuals,
              X
              ;

              A_lin1,A_lin0 := LinearAlgebra:-CompanionMatrix(A,t);
            DA_lin1,DA_lin0 := LinearAlgebra:-CompanionMatrix(A+DeltaA,t);
              A_lin := A_lin1-t*A_lin0;
              DA_lin := simplify(DA_lin1-t*DA_lin0-A_lin); #get rid of A_lin part; we want the coefficients only
              
              #now we need to figure out how to set up the problem with complex variables.

            omega := r_omega + I*i_omega;

            m,n := LinearAlgebra:-Dimension(A_lin):
            
            K:=Matrix( m,s_rank, (i,j)->(rk[i,j] + I*ik[i,j]) ):
            K_vars := convert(Re(K),list) assuming real: 
            K_vars := [op(K_vars), op(convert(Im(K),list)) ] assuming real:

            real_rank_constraint := convert(Re(evalc(LinearAlgebra:-HermitianTranspose(K).K-LinearAlgebra:-IdentityMatrix(s_rank))),Vector) assuming real:



            imaginary_rank_constraint := convert(Im(evalc(LinearAlgebra:-HermitianTranspose(K).K-LinearAlgebra:-IdentityMatrix(s_rank))),Vector) assuming real:

            rank_constraint := `<,>`(real_rank_constraint,imaginary_rank_constraint):



            A_0 := subs(t=0,A_lin):
            A_1 := (A_lin-A_0)/t: #use differentiation for larger dimensions...

            DA_0 := subs(t=0,DA_lin):
            DA_1 := (DA_lin-DA_0)/t: #again, use differentiation w.r.t. t for larger dimensions

            real_singularity_constraint := convert(Re( (A_0 + DA_0).K + evalc(omega * (A_1+DA_1).K)),Vector) assuming real:
            imaginary_singularity_constraint :=  convert(Im( (A_0 + DA_0).K + evalc(omega * (A_1+DA_1).K)),Vector) assuming real:


            singularity_constraint:= `<,>`(real_singularity_constraint,imaginary_singularity_constraint):


            constraints := `<,>`((singularity_constraint),(rank_constraint)):

            Lambda := Vector[row](ArrayNumElems(constraints), i->lambda[i]):

            omega_vars := [r_omega,i_omega]:
            
            X:= [op(DA_vars),op(omega_vars), op(K_vars)];
            vars := [op(X), op(convert(Lambda,list))];;
                


            L:= evalf(poly_norm_opt(DeltaA,t)^2 + subs([],(Lambda.constraints))[1]):


            residuals := VectorCalculus:-Gradient(L,vars):
            return [residuals,[A_lin,DA_lin,K,omega], [X,Lambda]];

end proc:


#Proc: linearize_mccoy_x_init
#Input: Objective function (gradient of Lagrangian), initial guess for omega and list of vars.
#Output: Initial values of X and Lambda
#Notes: Assumes DA_init = 0 for now; can be changed later.
linearize_mccoy_x_init := proc(A_lin,t,s_rank,DA_lin,omega_init,K,omega,A,A_init,DeltaA)
            
            local omega_init_pt,
            my_U,my_Vt, #kernel vars
            K_init,
            approx_kernel,
            DA_init,
            DeltaA_init
            
            ;
            
            omega_init_pt := [Re(omega)=Re(omega_init),Im(omega) = Im(omega_init)] assuming real;
            my_U,my_Vt:= LinearAlgebra:-SingularValues(subs(t=omega_init,A_lin),output=['U','Vt']);

            
       
            approx_kernel := LinearAlgebra:-HermitianTranspose(my_Vt[LinearAlgebra:-RowDimension(my_Vt)-s_rank+1..LinearAlgebra:-RowDimension(my_Vt)]);

            
            K_init :=[op( Equate(Re(K), Re(approx_kernel))),op( Equate(Im(K), Im(approx_kernel)))]  assuming real;


            #Okay, now we have the optimization problem sat up.  Get initial quantities


            #DA_init := [seq(DA_vars[i]=0,i=1..nops(DA_vars))]: #assume zero for now.

            DeltaA_init :=A_init-A;
            DA_init := [seq(op(Equate(Vector(1+mat(degree(DeltaA(i),t), degree(DeltaA_init(i),t)),PolynomialTools:-CoefficientVector(DeltaA(i),t)),Vector(1+max(degree(DeltaA(i),t), degree(DeltaA_init(i),t) ),PolynomialTools:-CoefficientVector(DeltaA_init(i),t)))), i=1..ArrayNumElems(DeltaA))];
            
            return [op(DA_init), op(K_init), op(omega_init_pt)];

end proc:


linearize_mccoy_lambda_init := proc(residuals,X_init,Lambda)
        local   AA,bb, #needed for LS system
                Lambda_init
                ;

                #get lagrange multipliers...
                AA,bb := LinearAlgebra:-GenerateMatrix(Equate(Vector(subs(X_init,residuals)),Vector(ArrayNumElems(residuals))),convert(Lambda,list));

                
                
                print(LinearAlgebra:-LeastSquares(AA,bb));
               Lambda_init :=  Equate(convert(Lambda,Vector), LinearAlgebra:-LeastSquares(AA,bb));


        return Lambda_init;

end proc:


#Proc: lower_mccoy
#Purpose: wrapper function to keep our sanity.
lower_mccoy := proc(A,t,A_init,omega_init,s_rank)
            local   my_args,
                    my_residuals,
                    A_lin,DA_lin,K,omega,
                    X,Lambda,X_init,
                    Lambda_init,
                    Z_init,
                    Z;
                    
                    my_args := linearize_mccoy(A,t,DeltaA,DA_vars,s_rank):
                    my_residuals := my_args[1]: #objective
                    A_lin,DA_lin,K,omega := op(my_args[2]):
                    X,Lambda := op(my_args[3]):
                    X_init := linearize_mccoy_x_init(A_lin,t,s_rank,DA_lin,omega_init,K,omega,A,A_init,DeltaA):
                    lambda_init := linearize_mccoy_lambda_init(my_residuals,X_init,Lambda);
                    Z_init := [op(X_init), op(lambda_init)];
                    Z:= [op(X),op(convert(Lambda,list))];
                    
                    return [my_residuals,Z,Z_init,omega];

end proc: