#read("../lib/alpha_optimization.mpl"):
#with(AlphaOptimization);
#read("../lib/nearest_singular_poly_functions.mpl"):
#read("../lib/nearest_smith_functions.mpl"):
#with(NearestPoly);
#with(NearestSmith);



PolyRankFactor := module()
option package; 
export nearest_singular_poly_rf; #do not give the user access to the other functions for now. 





#Proc: levenberg_marquardt_implicit_merit_rf
#Input: objective function and residuals, some other parameters
#Output: result of a Levenberg-Marquardt iteration
#Notes: implicit merit function being norm of residuals.
#Notes: this methd will be modified  for the rank-factor case to make it fast.

#DO NOT RUN
#uses some function pointers...
levenberg_marquardt_implicit_merit_rf := proc(residuals,J,vars,x_init,tol:=1e-16,max_iter := 100000,lm_param:= 1,lm_amp_factor:=1e-1,lm_damp_factor:=1e1)
      local i,
	    grad,
	    Ident,
	    j,
	    J_cur,
	    dx,
	    x_next_candidate,

	    JTJ,
	    rhs_k, #first argument to linear solve
	    lhs_k,#second rgument to linear solve
	    current_pt,
	    x_next,
	    I,
	    x_cur,
	    mu_k#the current levenberg-marquardt parameter
	    ;
	   
           # J := VectorCalculus:-Jacobian(residuals,vars);
	    #JTJ := LinearAlgebra:-Transpose(J).J;
            
            Ident:= LinearAlgebra:-IdentityMatrix(LinearAlgebra:-ColumnDimension(J));
	   # Ident:=Matrix(LinearAlgebra:-ColumnDimension(J));
	   # for j from 1 to LinearAlgebra:-RowDimension(Ident) do 
	#	Ident[j,j]:=1:
	   # od:
	    
	  #  printf("Initial guess has dimension %d,%d and rank %d\n", LinearAlgebra:-Dimension(JTJ),LinearAlgebra:-Rank(subs(x_init,JTJ)));

	  
	    
	    x_cur := subs(x_init, Vector(vars));
	    
	    mu_k := lm_param;
	    #printf("Beginning LM Iteration\n");
	    for i from 1 to max_iter do
		
		current_pt := subs(Equate(Vector(vars),x_cur));
		
		#mu_k := max(norm(subs(current_pt,residuals),1),lm_param);
		
		J_cur := subs(current_pt,J);
		JTJ := LinearAlgebra:-Transpose(J_cur).J_cur; 
		
		rhs_k := subs(current_pt,JTJ + mu_k * Ident);
		
		lhs_k := -1*LinearAlgebra:-Transpose(J_cur).subs(current_pt,residuals);
		
		dx := LinearAlgebra:-LinearSolve(rhs_k,lhs_k);
		
		x_next_candidate := Equate(Vector(vars),x_cur + dx);
		
		if norm(subs(x_next_candidate, residuals),2) <= norm(subs(current_pt,residuals),2)  then #valid step 
		    #mu_k := max(mu_k*(lm_amp_factor), 10^(-.5*Digits)); #take a smaller step next iteration
		   mu_k := max(mu_k*(lm_amp_factor), 0); #take a smaller step next iteration
		   
		   
		    #this is based on a criteria from a paper by Wright.
		    #|| residuals||_1 approximates the solution, so the penalty term
		    #blows up at the same rate as the solution, giving asymptotic quadratic convergence.
		    if norm(dx,2)<  norm(subs(x_next_candidate,residuals),2) and  norm(subs(x_next_candidate,residuals),2) <= 1.5 * norm(dx,2)   then 
			mu_k := norm(subs(x_next_candidate,residuals),2);
		    fi:
		    
		    x_next := x_cur + dx;
		    printf("k = %d, mu=%e, ||x_k - x_k+1|| = %e, ||dx|| = %e, ||f(x_k)||=%e\n",i,mu_k,norm(x_cur - x_next,2), norm(dx,2), norm(subs(current_pt,residuals),2));
		
		    if norm(x_cur-x_next,2) < tol or norm(subs(current_pt,residuals),2) < tol then break fi:
		
		    x_cur := x_next;
		
		
		else 
		    mu_k := mu_k * (lm_damp_factor);  #take a smaller step and try again
		fi:
		if mu_k >= 10^Digits then
                    WARNING("Unable to find a suitable step for improvement. Terminating.");
		break fi:
		
		
		
	    od:
	    
	    current_pt := subs(Equate(Vector(vars),x_cur));
	    return [norm(subs(Equate(vars,x_cur),residuals),2) , Equate(vars,x_cur)];
	    
end proc:





#Proc: fast_jacobian_rank_factor
#Input: "embedded A", perturbation vector, list of variables, U and V matrices for the rank factor
#Output:  Jacobian of the constraint vector $\nabla \vec(A+\Delta A - UV)$.
fast_jacobian_rank_factor := proc(A_lin,DA_vars,U,V)
    local JA,
          JU,
          JV,
          constraint_vector,
          J;
          
          
          
          constraint_vector := convert(A_lin,Vector);
          
          JA := VectorCalculus:-Jacobian(constraint_vector,DA_vars); #get the derivative of the perturbation structure. Do this with Maple because the closed-form is horrible.
          JU:=-1*LinearAlgebra:-KroneckerProduct(LinearAlgebra:-Transpose(V),LinearAlgebra:-IdentityMatrix(LinearAlgebra:-RowDimension(U),shape=identity));
          JV:=-1*LinearAlgebra:-KroneckerProduct(LinearAlgebra:-IdentityMatrix(LinearAlgebra:-ColumnDimension(V),shape=identity),U);
        
        
          J := <JA|JU|JV>;
          
          return J;
          
end proc:



#Proc: fast_hessian_rank_factor
#Input: "embedded A", perturbation vector, list of variables, vector of Lagrange multipliers, U and V matrices for the rank factor
#Output:  Jacobian of the constraint vector $\nabla \vec(A+\Delta A - UV)$.
#Notes: We assume $\lambda$ is a column vector

fast_hessian_rank_factor:= proc(A_lin,DA_vars,Lambda,U,V,X,X_init)
        local    Hxx, H,J,HU,
                m,r,n 
                ;
                
                m, r := LinearAlgebra:-Dimension(U);
                r,n := LinearAlgebra:-Dimension(V);
                J := fast_jacobian_rank_factor(A_lin,DA_vars,U,V); 
                
                #unfortunately this is faster than anything reasonable I could come up with.
                #note the scaling by -1 is because this is A_lin - U.V
                
                HU := LinearAlgebra:-KroneckerProduct(V,LinearAlgebra:-IdentityMatrix(m)).Lambda;
                HU := -1*VectorCalculus:-Jacobian(HU,convert(V,list));
                
                
                
                 H := <2*LinearAlgebra:-IdentityMatrix(nops(DA_vars)) | Matrix(nops(DA_vars), (m+n)* r,shape=zero)>; #first row blocks
              
                
                
                H := <H, <Matrix(m*r,nops(DA_vars),shape=zero) | Matrix(m*r,m*r,shape=zero) |HU > >;
                
                
               H := <H, <Matrix(n*r,nops(DA_vars),shape=zero) |LinearAlgebra:-Transpose(HU) | Matrix(n*r,n*r,shape=zero)>>;
                
                H:= <H| LinearAlgebra:-Transpose(J)>;
                
                
                H := < H, <J| Matrix(LinearAlgebra:-RowDimension(J), LinearAlgebra:-RowDimension(J) ,shape=zero )> >;
                
                return H;
                
        
                
end proc:

#Proc: fast_gradient_rf
#Input: 
#Output:
#Notes: Faster than maples gradient routine for some reason.

fast_gradient_rf := proc(A_lin,DA_vars,Lambda,U,V,X_init)
        local Gx,Gy;
        
        #easy
        Gy := convert(A_lin - U.V,Vector);
        
     
        
        Gx :=  convert(<2*convert(DA_vars,Vector), Vector(nops(X_init) - nops(DA_vars))> ,Vector)+
        LinearAlgebra:-Transpose(fast_jacobian_rank_factor(A_lin,DA_vars,U,V)) . Lambda;
        
        
        return convert(<Gx,Gy>,Vector);
                
        
        
        
end proc:



#Proc: slow_hessianXX_rank_factor
#Input: "embedded A", perturbation vector, list of variables, vector of Lagrange multipliers, U and V matrices for the rank factor
#Output:  Jacobian of the constraint vector $\nabla \vec(A+\Delta A - UV)$.
#Notes: We assume $\lambda$ is a column vector
#       This routine is pretty slow, but validates a theorem in the thesis. will not be used.
slow_hessianXX_rank_factor := proc(A_lin,DA_vars,Lambda,U,V)

        local HA,
          HU,
          H,
          Kom_MN, #commutation matrix
          Hxx,
          I_MR,I_RR,I_M,I_NR, I_R,I_N,I_MN, #some identity matrices
          m,n,r; #dimensions
          
          HA := 2*LinearAlgebra:-IdentityMatrix(nops(DA_vars),shape=diagonal); #the linear part in the constraints vanishes
          
          
          m, r := LinearAlgebra:-Dimension(U);
          r,n := LinearAlgebra:-Dimension(V);
          
          
          I_MR := LinearAlgebra:-IdentityMatrix(m*r,shape=identity);
          I_M := LinearAlgebra:-IdentityMatrix(m,shape=identity);
          I_NR := LinearAlgebra:-IdentityMatrix(n*r,shape=identity);
          I_RR := LinearAlgebra:-IdentityMatrix(r*r,shape=identity);
          I_R := LinearAlgebra:-IdentityMatrix(r,shape=identity);
          I_N := LinearAlgebra:-IdentityMatrix(n,shape=identity);
          I_MN := LinearAlgebra:-IdentityMatrix(m*n,shape=identity);
          I_MM := LinearAlgebra:-IdentityMatrix(m*m,shape=identity);
          
          Kom_MN := add( LinearAlgebra:-KroneckerProduct(LinearAlgebra:-KroneckerProduct(LinearAlgebra:-UnitVector[row](j,r,compact=false,storage=sparse),I_M,storage=sparse), LinearAlgebra:-UnitVector(j,r,compact=false,storage=sparse)), j = 1.. r) ; 
          #yes this is a sparse matrix and can be constructed more quickly,
          #furthermore,  this computation could even be done in place.
          
          
          print(LinearAlgebra:-Dimension(Lambda));
          
          print(m*n);
          
          print("computing 1\n");
          #HU :=  LinearAlgebra:-KroneckerProduct( LinearAlgebra:-Transpose(Lambda) , I_MR );
          
          print("computing 2\n");
          
          printf("Commutation matrix has dimension %d,%d\n", LinearAlgebra:-Dimension(Kom_MN) );
        
          
          #this needs to be sped up
          HU := #HU.
                 LinearAlgebra:-KroneckerProduct(LinearAlgebra:-KroneckerProduct(I_N, Kom_MN),I_M);
          print("computing 3\n");
          
            printf("HU matrix has dimension %d,%d\n", LinearAlgebra:-Dimension(HU) );
        
        
          
          HU :=-1* HU. 
                 LinearAlgebra:-KroneckerProduct(I_NR,convert(I_M,Vector));       
                 
            HU :=  LinearAlgebra:-KroneckerProduct( LinearAlgebra:-Transpose(Lambda) , I_MR ) . HU;
        
      #   return HU; #debugging for now.      
               
         H := <HA| Matrix(nops(DA_vars), (m+n)* r,shape=zero)>; #first row blocks
         H := <H, <Matrix(m*r,nops(DA_vars),shape=zero) | Matrix(m*r,m*r,shape=zero) |HU > >;
         H := <H, <Matrix(n*r,nops(DA_vars),shape=zero) |LinearAlgebra:-Transpose(HU) | Matrix(n*r,n*r,shape=zero)>>;
          
          
          #scaling is the final step
        return H;


end proc:


#Proc: nearest_singular_poly_rf
#Input: Matrix polynomial A, 
#       Initial guess of a singular matrix polynomial
#       Perturbation structure Delta A
#       list of variables for Delta A
#       indeterminant t
#       the desired rank for the approximation
#Output: A nearby rank desired_rank matrix polynomial with structure DeltaA or an indication of failure
#Notes: We will attempt to automate most of the procedure here. 
#       Several calls to quasi-Newton methods will be required in order to solve this problem.
#Computes a few initial guesses if needed...
#this calls a modified LM with some explicit function computation. 

#this function is mostly a wrapper...
nearest_singular_poly_rf :=proc(A,A_init,t,DeltaA,DA_vars, kernel_bound,R_bound,tol:=1e-10,project_feasible:=true,always_decrease:=false)
        
        #some local variables; the usual suspects. 
        local   X_init,X,
                UU,VV,A_lin,
                Lambda,Lambda_init,
                Z,Z_init,
                initial_guess,
                t_start,
                H,G # pre-compute these functions to save a lot of computational resources.
                ;
                
    
                
                initial_guess := compute_initial_guess_rf(A,A_init,t,DeltaA,DA_vars,kernel_bound,R_bound,tol,project_feasible,always_decrease): 
                
                X_init,X,UU,VV,A_lin := op(initial_guess); #get the arguments and stuff. 
                
                
                printf("Computing Lagrange multipliers...\n");
                t_start := time();
                #get the Lagrange multipliers. This is not super efficient, but who cares right now. This is a prototype in Maple... Code this in C if you want to be 1200X faster.
                Lambda,Lambda_init := op(initial_guess_lagrangian_rf(A_lin,DA_vars, UU,VV,X,X_init));
                
                printf("Lagrange multipliers computed in %f5 seconds.\n", time()-t_start);
                
                
                printf("Beginning Hessian and Gradient Computation...\n");
                
                t_start := time();
                H := fast_hessian_rank_factor(A_lin,DA_vars,Lambda, UU,VV,X,X_init);
                printf("Computed Hessian in %f5 seconds\n",time()-t_start);
                t_start := time();
                G:= fast_gradient_rf(A_lin,DA_vars,Lambda,UU,VV,X_init); #the residuals.
                
                printf("Computed Gradient in %f5 seconds\n",time()-t_start);
                
                
                Z:= [op(X), op(convert(Lambda,list))];
                
                
                
                Z_init := [op(X_init), op(Lambda_init)];
                
                printf("Beginning Quasi-Newton method.\n");
                
                t_start := time();
                my_soln := levenberg_marquardt_implicit_merit_rf(G,H,Z,Z_init,tol);
                printf("Stationary point computed in %f5 seconds\n",time()-t_start);
                
                #we will compute the same answer, but this takes much longer because of how derivatives are computed. The expensive part here is differentiation of the required quantities because Maple .
              #  t_start := time();
                
               # levenberg_marquardt_implicit_merit(G,Z,Z_init,1e-10);
               
               # printf("Second Stationary point computed in %f5 seconds\n",time()-t_start);
                
                return my_soln;
           
       
end proc:



#proc: compute_initial_guess_rf
#Input: A, 
#   initial guess for A
#   indeterminant in t
#   perturbation structure
#   variables for perturbation structure
#   bound on the degree of the kernel, note that kernel_bound -1 = deg(ker(A)), the -1 is for a constant term since...
#   bound on the rank of the block convolution matrix sought
compute_initial_guess_rf := proc(A,A_init,t,DeltaA,DA_vars, kernel_bound,R_bound,tol:=1e-10,project_feasible:=true,always_decrease := false)
        local   UU,U_init,U_init_pt,U_vars,
                VV,V_init,V_init_pt,V_vars,
                DA_init,
                X,X_init,
                m,n,d,M,N,R,
                A_rank, A_lin
                ;
        
    
            m,n := LinearAlgebra:-Dimension(A);
            d:=NearestPoly:-matrix_degree(A,t);
            

            
            
            A_lin := NearestPoly:-linearize_A_bound(A+DeltaA,t,kernel_bound)[1]:

            
            M,N := LinearAlgebra:-Dimension(A_lin);
            
            printf("Computing initial guesses\n");
            t_start := time();
            
            #call a routine from legacy functions
            DA_init := NearestPoly:-generate_initial_point(A,A_init,DeltaA,t):
            
            cf_vec,cf_mat := LinearAlgebra:-GenerateMatrix(DA_init, DA_vars);
            
            DA_init := Equate(convert(DA_vars,Vector),Optimization:-LSSolve([cf_mat,cf_vec])[2]);
            
            printf("Computed Initial guess for DeltaA\n");

            U_init,S_init,V_init := LinearAlgebra:-SingularValues(subs(DA_init,A_lin),output=['U','S','Vt']):

            #figure out rank here. Can do some form of compression if needed.
            Sigma := LinearAlgebra:-DiagonalMatrix(S_init[1..LinearAlgebra:-ColumnDimension(A_lin)], shape=diagonal):
            
            A_rank := LinearAlgebra:-Rank(Sigma); #lazy me!

            printf("A_lin has rank %d  and dimension %d,%d\n",A_rank,M,N );
            
            #take the smallest between the  compression
            #and bound provided by the user.
            R := max(A_rank, R_bound );
            
            if always_decrease = true then
                R := max(A_rank-1, R_bound );
            fi:
            Sigma := Sigma[1..R, 1..R];

            UU:= Matrix(LinearAlgebra:-RowDimension(A_lin), R, (i,j)->uu[i,j]):

            VV:= Matrix( R, LinearAlgebra:-ColumnDimension(A_lin) ,(i,j)->vv[i,j]): 
      
            U_vars := convert(UU,list):
            V_vars := convert(VV,list):
            
            

            U_init := (U_init)[1..LinearAlgebra:-RowDimension(UU),1..LinearAlgebra:-ColumnDimension(UU)];


            
            #need to re-scale by singular values for 
            V_init := Sigma.V_init[1..LinearAlgebra:-RowDimension(VV), 1 .. LinearAlgebra:-ColumnDimension(VV)] ; #because this is the transpose

        
        
            #V_init := LinearAlgebra:-Transpose(V_init);
            U_init_pt := Equate(UU,U_init):
            V_init_pt := Equate(VV,V_init):

        
            

            X:= [op(DA_vars),op(U_vars),op(V_vars)]:
            X_init:= [op(DA_init), op(U_init_pt), op(V_init_pt)]:
            
            
            constraints := convert(A_lin - UU.VV,Vector); #don't convert to set I guess..
        
            
            
            
            if project_feasible = true and norm(subs(X_init,constraints),2)  > 1e-8 then 
            printf("Projecting initial guess to feasible point...\n");
            
                #replace this function call with a faster one for this problem.     
                X_init := levenberg_marquardt_implicit_merit_rf(constraints,fast_jacobian_rank_factor(A_lin,DA_vars,UU,VV),X,X_init,tol)[2]:
                
                #might need to do another compression after this step!
                
                
                
            printf("Feasible point has rank %d  and dimension %d,%d\n",LinearAlgebra:-Rank(subs(X_init,A_lin)),M,N );
                
               # print("Beginning second projection.\n");
                #X_init := levenberg_marquardt_implicit_merit(constraints,X,X_init,1e-8)[2]:
                
                
                            U_init,S_init,V_init := LinearAlgebra:-SingularValues(subs(DA_init,A_lin),output=['U','S','Vt']):

            #figure out rank here. Can do some form of compression if needed.
            Sigma := LinearAlgebra:-DiagonalMatrix(S_init[1..LinearAlgebra:-ColumnDimension(A_lin)], shape=diagonal):
            
            A_rank := LinearAlgebra:-Rank(Sigma); #lazy me!

            printf("A_lin has rank %d  and dimension %d,%d\n",A_rank,M,N );
            
            #take the smallest between the  compression
            #and bound provided by the user.
            R := max(A_rank, R_bound );
            
            if always_decrease = true then
                R := max(A_rank-1, R_bound );
            fi:
            Sigma := Sigma[1..R, 1..R];

            UU:= Matrix(LinearAlgebra:-RowDimension(A_lin), R, (i,j)->uu[i,j]):

            VV:= Matrix( R, LinearAlgebra:-ColumnDimension(A_lin) ,(i,j)->vv[i,j]): 
      
            U_vars := convert(UU,list):
            V_vars := convert(VV,list):
            
            

            U_init := (U_init)[1..LinearAlgebra:-RowDimension(UU),1..LinearAlgebra:-ColumnDimension(UU)];


            
            #need to re-scale by singular values for 
            V_init := Sigma.V_init[1..LinearAlgebra:-RowDimension(VV), 1 .. LinearAlgebra:-ColumnDimension(VV)] ; #because this is the transpose

        
        
            #V_init := LinearAlgebra:-Transpose(V_init);
            U_init_pt := Equate(UU,U_init):
            V_init_pt := Equate(VV,V_init):

        
            

            X:= [op(DA_vars),op(U_vars),op(V_vars)]:
            X_init:= [op(DA_init), op(U_init_pt), op(V_init_pt)]:
            
            
         #   constraints := convert(A_lin - UU.VV,Vector); #don't convert to set I guess..
        
                
            fi:
            printf("Feasible point computed\n");
            
            t_end := time()-t_start; 
            
            
            printf("Computed initial guess in %5f seconds\n",t_end);
            
            
            #initial guess is sort of done at this point. we want to project to a feasible point first if the initial guess sort of sucks. 
            
            
            printf("Returning Initial Guess.\n");
            return [X_init,X,UU,VV,A_lin];
            
        
end proc:

#Proc: compute_lagrange_multipliers
#Proc: initial_guess_lagrangian
#Input: stuff we need to compute the Lagrangian (Gradient, initial guess, lambda variables)
#Output: initial guess for Lagrange multipliers (computed via linear least squares approximation)

initial_guess_lagrangian_rf := proc(A_lin,DA_vars,U,V,X,X_init)
	local lagrangian_system,
	      my_eqns,
	      cf_mat,
	      cf_vec,
	      Gx;
	      ineq_cons,
	      ls_soln,
	      Lambda_init, Lambda,
	      J
	      ; 

	      
    
        #solving a linear system for initial guess
	Lambda := Vector(ArrayNumElems(A_lin),j->lambda[j]);
	    Gx :=  convert(<2*convert(DA_vars,Vector), Vector(nops(X_init) - nops(DA_vars))> ,Vector)+
        LinearAlgebra:-Transpose(fast_jacobian_rank_factor(A_lin,DA_vars,subs(X_init,U),subs(X_init,V))) . Lambda;
        
	#lagrangian_system := convert(LinearAlgebra:-Transpose(J).Lambda,Vector);
	lagrangian_system := subs(convert(subs(X_init,Gx),Vector));
	
	
	
	my_eqns := Equate(lagrangian_system, Vector(ArrayNumElems(lagrangian_system)));
	cf_mat,cf_vec := LinearAlgebra:-GenerateMatrix(my_eqns,convert(Lambda,list));
	

	
	ls_soln := Optimization:-LSSolve([cf_vec,cf_mat])[2];
	Lambda_init := Equate(convert(Lambda,Vector), ls_soln);

	
	return [Lambda,Lambda_init];
    

end proc:



end module: